#include <gnome.h>
#include <glade/glade.h>
#include <mysql/mysql.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include <sys/utsname.h>
#include <ctype.h>

#define		DATABASE_VERSION 1


/* Structures - See below for comments. */
typedef struct _Mailbox Mailbox;
typedef struct _Message Message;
typedef struct _GmailApp _GmailApp_;
typedef struct _DTimer DTimer;
typedef struct mime_part Mime_Part;

/* There is one of these mailbox structures attached as row data in every
 * row of the clist in the leftpane.
 * The name isn't in here as it is in the clist.
 * 
 */
struct _Mailbox
{
	gboolean unreadmail; 	/* Is true if mailbox contains unread mail. */
	gchar *query; /* SQL Query. */
	gchar *type; /* Currently this only is 'vfolder' */
};

/* This is a common widget that we reuse around the place. */
struct _DTimer
  {
	GtkWidget *dialog;		/* The popup window */
	GtkWidget *label;
	GtkWidget *clist;			/* Contains dynamic download info. */
	GtkWidget *button;
	GtkWidget *progress; 	
	gint	value;
	gint	total;
	gint	timeout;
};


/* 
 * This struct used to store the information about a single mail message.
 * Not all of these fields are used in every use.
 */
struct _Message
  {
	gint  id;					/* sql id  */
	GString *headers;		/* The headers bit without the body. */
	GString *message;		/* The message = body. */
	gchar *date;			/* Date sent. */
	gchar *subject;		/* Subject header. */
	gchar *from;			/* Sender. */
	gchar *to;				/* Receiver. */
	gchar *cc;				/* cc field */
	gchar *bcc;				/* bcc field */
	gchar *readstatus; 		/* Gmail status. Read/unread etc etc.. */
	gchar *direction;	/* Either 'Incoming' or 'Outgoing'  */
	gchar *uid;				/* Pop 3 UID. */
};

/* 
 * GmailApp is one copy of all general global vars.
 * We only ever keep _one_ copy of this structure in use.
 * 
 * Alot of the variables here are just hacks and could be avoided.
 * 
 * This struct is always evolving.
 *
 * We also use globals in here to record status of certain widgets.
 * eg: progressbar values, radio button highlights etc.
 *
 */
struct _GmailApp
	{

	GnomeClient *client;	

	/* Global window widgets. */
	GtkWidget *app, *appbar, *hpaned, *leftpane;

	/* The right pane contains two possible panes. */
	GtkWidget *rightpane, *msglistpane, *msgpane;

	GtkWidget *mblist; /* clist in left pane. */
	GtkWidget *rightlist; /* clist in right pane. */

	/* Database info. */
	MYSQL mysql;
	MYSQL mysql2; /* sometimes we need two connections at once. */

	/* The DRUID! */
	GtkWidget *druid;
	GtkWidget *root_user, *root_pass; /* needed by druid */

	/* Our beloved fonts. */
	GtkStyle 	*boldstyle;
	GtkStyle 	*style;

	/* Global var hacks. These should be done away with. */
	gint	left_pos; /* Row selected in left clist. */
	gint	right_pos; /* Row selected in right clist. */
	gint	msg_id; /* database id of last opened message. */

	gchar *sortfield, *sortorder;

	/* The following mysql variables are kept globally to save us a gnome_config
	 * call every time we open a mysql connection. This is because we are 
	 * opening and closing connections ALOT.
	 */
	gchar *hostname, *password, *username, *dbname;

	/* These widgets are used in the preferences notebook. */
	GtkWidget *nbook;	/* Notebook for prefs */

	/* identity tab */
  GnomePropertyBox *pbox;
  GtkWidget *real_name, *email, *signature;

	/* Mysql Mail Store Tab. */
  GtkWidget *db_host_entry;
  GtkWidget *db_user_entry;
  GtkWidget *db_pass_entry;
  GtkWidget *db_name_entry;
  GtkWidget *db_table_entry;

		/* mailboxes tab */
    GtkWidget *pr_mboxes, *smtp_server, *mail_directory;
    GtkWidget *rb_local_mua, *rb_smtp_server;
    GtkWidget *quote_str;
		gint	pr_mbox_row; /* The currently selected row. */

		/* POP3 Servers Tab. */		
    GtkWidget *mbox_popup;
    GtkWidget *username_entry;
    GtkWidget *pop_server_entry;
    GtkWidget *pop_password_entry;
    GtkWidget *pop_leave_mail;
    GtkWidget *pop_threshold;
	
		/* Display tab. */
  	GtkWidget *defaultfont, *boldfont;
		GtkWidget *cache_vf_toggle, *open_first_msg;
    GtkWidget *rb_twopane, *rb_threepane;
  	GtkWidget *addressbook_entry;

	/* Add Mailbox Popup Window. */
  	GtkWidget *mbox_name_entry;
  	GtkWidget *vfolder_name_entry;
  	GtkWidget *sql_entry;

	/* Used in the download email popup. */	
	gint 	newmsgs;
	GtkWidget *rightlist_popup;
	gint updated_vfolder; 	/* Used when a vfolder has downloaded new messages. */
	};

/* Spruce stuff... */

/* define IPv4 dotted-decimal string length */
#ifndef INET_ADDRSTRLEN
#define INET_ADDRSTRLEN 16
#endif

/* define IPv6 hex string length */
#ifndef INET6_ADDRSTRLEN
#define INET6_ADDRSTRLEN 46
#endif

typedef unsigned short int port_t;

struct _server_
{
   char *hostname;
   char ip[INET_ADDRSTRLEN + 1];     /* IPv4 dotted decimal -> 16 chars + NULL */
   struct sockaddr_in sin;
   port_t port;
};

/* PLEASE use this typedef when creating a variable of this type */
typedef struct _server_ Server;

struct mime_part {
	gchar type[64];
	gchar subtype[64];
	gchar parameter[256];
  gchar disposition[256];
	gchar encoding[64];
	guint pos, len;
};

