/*
 * Gmail. A Gnome email client.
 * Copyright (C) 1999-2000 Wayne Schuller
 *
 * matching.c - matching is the gmail feature which keeps an indexed record of which vfolders a msg matches too.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "main.h"


extern gboolean	DestroyRightPane(void);
extern gboolean open_db(MYSQL *mysql);
extern DTimer * msg_match_popup (void);
extern gboolean vfolder_msg_unread(gint id);
extern void vfolder_contains_unread(gchar *vfolder, gboolean unread_status);

extern _GmailApp_	*GmailApp;

void rebuild_matched_values (GtkWidget *widget, void *data);
void recreate_matched_index (GtkWidget *widget, void *data);
void mark_as_matched(gint id, GString *vfolders, gboolean overwrite); 
void check_vfolder_matches(gint id, DTimer *timer);
static void notify_popup_match(gchar *name, DTimer *timer);


#define set_status_label(label, text) gtk_label_set_text((GtkLabel*)label, text); while (g_main_iteration(FALSE));
#define set_status(a) set_status_label(GmailApp->pop3_status_label, a)

/* recreate_matched_index
 *
 * Update the matched 'enum' design in the database according to the names
 * of the vfolders that the user has.
 *
 */
void
recreate_matched_index (GtkWidget *widget, void *data) 
{
	GtkWidget *dialog;
	GtkWidget *label;
	gint 	row;
	gint 	i;
	GString *query;
	gchar *msg;


	/* FIXME: Create popup dialog with a progress bar?. */
	dialog = gnome_dialog_new("Gmail Warning:",
                        GNOME_STOCK_BUTTON_OK,
                        GNOME_STOCK_BUTTON_CANCEL,
                        NULL);
 	if (GmailApp->app)
		gnome_dialog_set_parent(GNOME_DIALOG(dialog), GTK_WINDOW(GmailApp->app));

	label = gtk_label_new("Warning: Recreating the matched index involves rebuilding one of the gmail tables.\nIt is required for fast gmail caching, but may take a few minutes on large databases.\nClick OK to continue.");
	gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(dialog)->vbox),label, FALSE,FALSE,0);
	gtk_widget_show_all(dialog);
	/*set up an infinite loop*/
	for(;;) {
		i = gnome_dialog_run(GNOME_DIALOG(dialog));
		if	(i == 0) {
			gnome_dialog_close(GNOME_DIALOG(dialog));
			break;
		} else if(i == 1) {

			gnome_dialog_close(GNOME_DIALOG(dialog));
			return;
			break;
		} else if(i < 0) {

			/*the user closed the dialog from the window
       manager*/
			return;
			break;
		}
 	}


	/* We are going to modify the matched field according to vfolder design */
	query = g_string_new("ALTER TABLE display MODIFY matched SET(' '");


	/* Add everything from the prefs clist to the leftpane clist. */
	for (row = 0; row != GTK_CLIST (GmailApp->mblist)->rows; row++) {
		gchar *text = NULL;

		/* Grab the name of the vfolder */
		gtk_clist_get_text(GTK_CLIST(GmailApp->mblist), row, 0, &text);

		if (text != NULL)
			g_string_sprintfa(query, ",'%s'", text);
		}

	g_string_append(query, ") NOT NULL");

	/* Open Database. */
	if (!open_db(&GmailApp->mysql)) { return; };

	/* FIXME: use the appbar and be nicer about failure here. */

	if (mysql_query(&GmailApp->mysql, query->str) != 0) {
		msg = g_strdup_printf ("Query failed:\n%s\nSee the standard output to see the full query.\n", mysql_error(&GmailApp->mysql));
		dialog = gnome_message_box_new(msg, GNOME_MESSAGE_BOX_ERROR, "doh!", 0);
		gtk_window_set_modal(GTK_WINDOW (dialog), TRUE);
 		gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
 		gtk_widget_show(dialog);
		g_free(msg);	
		g_print("Failed query was: |%s|\n", query->str);
		mysql_close(&GmailApp->mysql); 
		g_string_free(query, TRUE);
		return;
		} 

	mysql_close(&GmailApp->mysql); 
	g_string_free(query, TRUE);

	/* Report to the user. */
	/* We were using a gnome_message_box here but it coredumped alot :( */
	dialog = gnome_dialog_new (
                 "Gmail Popup",
                 GNOME_STOCK_PIXMAP_EXEC,
                 GNOME_STOCK_BUTTON_OK,
                 NULL);
	if (GmailApp->app)
             gnome_dialog_set_parent (GNOME_DIALOG (dialog), GTK_WINDOW (GmailApp->app));
	/* gtk_widget_set_usize(GTK_WIDGET(dialog), 200, 125);	 */
	label = gtk_label_new ("Matched Index Updated\nYou should now rescan the database using \"Settings->Advanced->Rebuild MI Values\"\nTo do this immediately, click 'Exec', otherwise click 'OK' to go back to gmail.");
	gtk_box_pack_start (GTK_BOX (GNOME_DIALOG (dialog)->vbox), label, TRUE, TRUE, 0);

  /* gtk_signal_connect_object(GTK_OBJECT(dialog), "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dialog)); */

	gtk_widget_show_all(GNOME_DIALOG(dialog)->vbox);

	switch(gnome_dialog_run_and_close(GNOME_DIALOG (dialog))) {
		case 0: rebuild_matched_values(NULL, NULL);
						break;

		case 1: break;

		default: break;
	}
}

/* rebuild_matched_values is called from the menu when the user wants
 * to manually rescan the entire database for updated vfolder entries.
 */
void
rebuild_matched_values (GtkWidget *widget, void *data) {
	MYSQL_RES *res;
	MYSQL_ROW row;
	gchar *query;
	gint affected_rows;
	gchar text[256];
	GtkWidget *child;
	GtkWidget *dialog, *label;
	gint rw;
	DTimer *timer;
	gint i;

	/* Send warning dialog. */
	dialog = gnome_dialog_new("Gmail Warning:",
                        GNOME_STOCK_BUTTON_OK,
                        GNOME_STOCK_BUTTON_CANCEL,
                        NULL);
 	if (GmailApp->app)
		gnome_dialog_set_parent(GNOME_DIALOG(dialog), GTK_WINDOW(GmailApp->app));

	label = gtk_label_new("Rebuilding the MI values is important for fast gmail caching.\nDon't run this if you haven't updated your matched-index or\nif you haven't tested that all your vfolders work.\nYou should have an up to date matched-index (MI) before running this.\nClick OK to continue.");
	gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(dialog)->vbox),label, FALSE,FALSE,0);
	gtk_widget_show_all(dialog);
	/*set up an infinite loop*/
	for(;;) {
		i = gnome_dialog_run(GNOME_DIALOG(dialog));
		if	(i == 0) {
			gnome_dialog_close(GNOME_DIALOG(dialog));
			break;
		} else if(i == 1) {

			gnome_dialog_close(GNOME_DIALOG(dialog));
			return;
			break;
		} else if(i < 0) {

			/*the user closed the dialog from the window
       manager*/
			return;
			break;
		}
 	}


	/* Loop through each vfolder clearing them as containing no unread
	 * messages. The ones that really do have unread messages will get 
	 * marked as such when we process every message.
	 */
	for (rw = 0; rw != GTK_CLIST (GmailApp->mblist)->rows; rw++) {
		Mailbox	*mailbox;
		mailbox = gtk_clist_get_row_data(GTK_CLIST(GmailApp->mblist), rw);	
		mailbox->unreadmail = FALSE;
	}

	/* Open the database. 
	 * We will keep this one connection open for the entire process.
	 * Because it involves lots of queries it is fast to keep reusing the
	 * same connection.
	 * FIXME: Why don't we just do this all the time?
	 */
	if (!open_db(&GmailApp->mysql)) { return; };

	/* if (mysql_ping(&GmailApp->mysql) == 0) g_print("PING11\n"); */

	/* Select all from the database. */
	query = g_strdup_printf("SELECT (id) FROM display ORDER BY id DESC");

	/* Execute query.. */
	if (mysql_query(&GmailApp->mysql, query) != 0) {
		g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		mysql_close(&GmailApp->mysql);
		g_free(query);
		return;
	} 

	/* Check the result. */	
	res = mysql_store_result(&GmailApp->mysql);

	if (res == NULL) {
		g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		g_warning("vfolder match query returned NULL. :(");
		mysql_free_result(res); 
		mysql_close(&GmailApp->mysql);
		g_free(query);
		return;
		}

	affected_rows = mysql_num_rows(res);

	timer = msg_match_popup();
	timer->value = 0;
	timer->total = affected_rows;

	/* Loop through every row. Stop if the pop3_download_popup dissappears. */
	while ((row = mysql_fetch_row(res)) && 
				  GTK_IS_WIDGET(timer->dialog)) {

			check_vfolder_matches(atoi(row[0]), timer); /* Check matches for that msg. */
			timer->value++;										/* Increment count */

			sprintf(text, "Matching msg %i of %i\n", timer->value, timer->total);
			set_status_label(timer->label, text);
			}

	/* Close the database, free result. */
	mysql_free_result(res); 
	mysql_close(&GmailApp->mysql);  
	g_free(query);

	/* Update the button text. */
	child = GTK_BIN (timer->button)->child;
	if (GTK_IS_LABEL(child)) gtk_label_set(GTK_LABEL(child), "Close");

}

/* mark_as_matched - add a vfolder name to the msg's matched field.
 * Put the vfolder name in the matched field of msg 'id' in database.
 * When check_vfolder_matches finds that an id is matched in a vfolder
 * query, it calls this function to put the name of that vfolder in it's
 * matched field.
 * Overwrite whatever is in the matched field currently if they want to.
 * If overwrite is false, we simply append vfolders to the current matched
 * field (but we check that it isn't there first.)
 */
void 
mark_as_matched(gint id, GString *vfolders, gboolean overwrite) {
	MYSQL_RES *res;
	MYSQL_ROW row;
	gchar *query;
	gboolean finished = FALSE;

	/* Open the database. (use second connection)*/
	if (!open_db(&GmailApp->mysql2)) { return; };

	if (!overwrite) {
		/* We need to only add what is in the string. We assume there is
		 * only one vfolder being added.
		 * The first thing is to check that the name isn't already there.
		 */

		if (vfolders->len <= 1) {
			return; /* nothing to do */
			}


		/* FIXME: I think this query is flawed because it will get a match 
		 * if the vfolder name is a superset of the current one.
		 */
		query = g_strdup_printf("SELECT (matched) FROM display where id = %i", id);

		/* Execute query.. */
		if (mysql_query(&GmailApp->mysql2, query) != 0) {
			g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql2));
			mysql_close(&GmailApp->mysql2);
			g_free(query);
			return;
		} 

		res = mysql_use_result(&GmailApp->mysql2);

		if (res == NULL) {
			g_warning("Result failed: %s\n", mysql_error(&GmailApp->mysql2));
			mysql_close(&GmailApp->mysql2);
			g_free(query);
			return;
			}

		row = mysql_fetch_row(res);


		/* Check if vfolder name is in this row. */
		if (strstr(row[0], vfolders->str) == NULL) {
				g_string_append(vfolders, ",");
				g_string_append(vfolders, row[0]);
				/* g_print("msg id %d matched value will be:  %s\n", id, vfolders->str);*/
				finished = FALSE;
			} else finished = TRUE; /* Our vfolder already is in the field. */

		/* Close the database connection. */
		g_free(query);
		mysql_free_result(res); 
		if (finished) {
			mysql_close(&GmailApp->mysql2);
			return;
		}
	}

	if (vfolders->len == 0) g_string_assign(vfolders, "NULL");

	/* g_print("inserting: %s\n", vfolders->str); */

	query = g_strdup_printf("UPDATE display SET matched=\"%s\" WHERE id = %i", vfolders->str, id);

	/* Execute query.. */
	if (mysql_query(&GmailApp->mysql2, query) != 0) {
		g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql2));
		mysql_close(&GmailApp->mysql2);
		g_free(query);
		return;
	} 

	/* g_print("Successfully marked %d as: %s\n", id, vfolders->str);  */

	/* Close the database connection. */
	mysql_close(&GmailApp->mysql2);
	g_free(query);
}

/* check_vfolder_matches - check which vfolders this msg belongs to.
 *
 * Go through every vfolder and see if this message belongs to that query.
 * If so add the name of that vfolder to the msg's 'matched' field.
 * This function clears the current contents of the 'matched' field.
 *
 * This code is called once for every message on the server, and within it
 * there is a loop through every vfolder. Thus there are lots of iterations.
 * Any mem leaks or sub-optimal things will really drag out gmail.
 *
 * FIXME: Check for mem leaks here.
 */
void 
check_vfolder_matches(gint id, DTimer *timer) 
{
	MYSQL_RES *res;
	gchar *newquery;
	GString *matches; /* Names of vfolders that match. */
	gint affected_rows;
	gboolean found_match = FALSE;
	gboolean msg_unread = FALSE; 
	gint row;

	matches = g_string_new(NULL);

	msg_unread = vfolder_msg_unread(id);

	/* Create the database connection before going into the loop through 
	 * the vfolders. That we we get a speed up by not opening and closing
	 * it all the time.
	 * FIXME: open the connection at one higher level up.
	 * FIXME: Check we are freeing mysql data properly as we go.
	 */
	if (!open_db(&GmailApp->mysql)) { return; };

	/* Loop through each vfolder looking for matches.. */
	for (row = 0; row != GTK_CLIST (GmailApp->mblist)->rows; row++) {
		Mailbox	*mailbox;
		gchar *vfolder;

		mailbox = gtk_clist_get_row_data(GTK_CLIST(GmailApp->mblist), row);	
		gtk_clist_get_text(GTK_CLIST(GmailApp->mblist), row, 0, &vfolder);

		/* Don't match Inbox or Unread.
		 * Anything that doesn't get matched in this loop is matched 
		 * against the Inbox. 
		 *
		 * We handle Unread manually because if we matched it in this
		 * loop it would get excluded from Inbox. We only want to exclude
		 * from the Inbox anything which matches the users custom vfolders.
		 * 
		 * By the same logic we shouldn't be matching against Outbox in here
		 * either, but we do because Inbox excludes outgoing messages
		 * by definition, so it doesn't matter if this loop does it for us.
		 *
		 * Focus your concentration on the above comments and they will make sense.
		 */	
		if ((strcmp(vfolder, "Inbox") != 0) &&
			(strcmp(vfolder, "Unread") != 0) &&
			(mailbox->query != NULL))	{

			/* g_print("is %d in %s? (mailbox->query: %s)\n", id, vfolder, mailbox->query); */

			newquery = g_strdup_printf("SELECT display.id FROM display,details WHERE display.id=details.id AND (%s) AND display.id = \"%d\"", mailbox->query, id); 

			/* g_print("oldquery: %s\n", mailbox->query);
			g_print("newquery: %s\n", newquery);  */


			/* Execute query.. */
			if (mysql_query(&GmailApp->mysql, newquery) != 0) {
				g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql));
				g_print("match query was: %s\n", newquery);

 				/* Abort this vfolder but keep going on the matching. */
				g_free(newquery);
				continue; 
			} 
	
			/* Check the result. */	
			res = mysql_store_result(&GmailApp->mysql);
		
			if (res == NULL) {
				g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql));
				g_warning("vfolder match query returned NULL. :(");
 				/* Abort this vfolder but keep going on the matching. */
				g_free(newquery);
				continue;
				}

			affected_rows = mysql_num_rows(res);

			/* We're finished with the results now. */
			mysql_free_result(res); 
			g_free(newquery);

			switch (affected_rows) {
				case 0:
				/* 0 matches - just ignore and keep going.*/
				/* g_print("id %d didn't match %s\n", id, vfolder);  */
				break;

				case 1:
				/* We have a match! */
				g_string_append(matches, vfolder);
				g_string_append_c(matches, ','); 

				/* If the message is unread, then mark the vfolder as
				 * having an unread message at this point.
				 */
				if (msg_unread) {
					vfolder_contains_unread(vfolder, TRUE);
					}

				/* Check if the match is in the vfolder currently displayed.
				 * If so set the global saying which row it is.
				 * This way at the end of a pop3 download the code will
				 * update that row for the user.
				 */
				if (row == GmailApp->left_pos) GmailApp->updated_vfolder = row;

				/* Send output to user.
				 */
				notify_popup_match(vfolder, timer);

				found_match = TRUE; /* remember this */
				break;

				case -1:
				/* error */
				g_print("Affected rows error: %s\n", mysql_error(&GmailApp->mysql));
				break;

				default:
				/* If it matched more than one it's probably due to the
				 * AND clause we add confuses some other logical operator.
         * We need people to use brackets around logical clauses.
				 */
				g_print("vfolder match (%d on %s) returned weird result: %d rows. You probably need to put round brackets around your vfolder queries and try again.\n", id, vfolder, affected_rows);
				break;
				}

			}
	}

	mysql_close(&GmailApp->mysql);

	/* If we haven't found a match we consider it part of the Inbox. 
	 */
	if (!found_match) {

		/* Leftovers match the Inbox vfolder. (assumed to always exist) */
		g_string_append(matches, "Inbox");
		g_string_append_c(matches, ','); 

		/* Add it to the download report. */
		notify_popup_match("Inbox", timer);

		/* Bold the vfolder name if the message is unread. */
		if (msg_unread) 
			vfolder_contains_unread("Inbox", TRUE);
		}

	if (msg_unread) {

		/* It matches the unread vfolder. (assumed to always exist)*/
		g_string_append(matches, "Unread");
		g_string_append_c(matches, ',');

		/* We don't add it to the download report because if it is a 
		 * download it is obvious to the user that the message would be
		 * unread.
		 * FIXME: But it might be useful to see the number of unread
		 * messages if they are rebuilding their matched fields.
		 */

		/* Bold the vfolder name. */
		vfolder_contains_unread("Unread", TRUE);

		/* See comments for the similar line to this above. */
		if (GmailApp->left_pos == 0) GmailApp->updated_vfolder = 0;
		}

	/* Update the matched index for this message.
	 * 
	 */
	mark_as_matched(id, matches, TRUE); 

	g_string_free(matches, TRUE);
}

/* notify_popup_match - Add another matched msg to the popup menu. */
static void 
notify_popup_match(gchar *name, DTimer *timer) 
{
	gint row;

	/* FIXME: It is still crashing here on send sometimes. */
	/* Need to check there is a popup. */
	if (timer == NULL || 
		!GTK_IS_WIDGET(timer->dialog) || !GTK_IS_WIDGET(timer->clist)) 
			return;

	row = gtk_clist_find_row_from_data(GTK_CLIST(timer->clist), name);
	/* g_print("match %i (%s)\n", row, name); */

	if (row == -1) {
		gchar *newtext[2];

		/* Create a new row. */

		newtext[0] = g_strdup(name);
		newtext[1] = g_strdup("1");
		gtk_clist_append (GTK_CLIST (timer->clist), newtext);
		gtk_clist_set_row_data(GTK_CLIST (timer->clist), GTK_CLIST (timer->clist)->rows-1, name);
		} else {
		gint val;
		gchar *text;

		/* Increment row by one. */

		gtk_clist_get_text(GTK_CLIST(timer->clist), row, 1, &text);
		val = atoi(text); val++;
		text = g_strdup_printf("%i", val);	
		gtk_clist_set_text(GTK_CLIST(timer->clist), row, 1, text);


		}

}
