/*
 * Gmail. A Gnome email client.
 * Copyright (C) 1999 Wayne Schuller
 *
 * mblist.c - contains functions for displaying vfolders + messages.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "main.h"

void 	mboxclicked_cb (GtkWidget *mblist, gint row, gint column, GdkEventButton *event, gpointer data);
void	columnclicked_cb (GtkCList * clist, gint column);
void	delete_msg (GtkWidget *widget, gpointer data);
gchar * 	message_content(Mailbox *, gint);
void	nextmsg_cb (GtkWidget *widget, gint row, gint column, GdkEventButton *event, gpointer data);
void	prevmsg_cb (GtkWidget *widget, gint row, gint column, GdkEventButton *event, gpointer data);
void	display_message(Message *msg, gboolean forced);
GtkWidget * create_msg_pane(void);
void attachment_button_press_event (GtkWidget *widget, GdkEventButton *event, gpointer data);
void save_attach (GtkWidget *widget, gint row);

extern 	_GmailApp_	*GmailApp;
extern  void vfolder_display_rightpane(gint row);
extern 	Message * load_vfolder_msg(gint id);
extern  gboolean open_db(MYSQL *mysql);
extern	gchar * strip_name(gchar *to);
extern  gboolean				TwoPaneView();
extern	gboolean				DestroyRightPane(void);
extern 	gint mime_parse (gchar *message);
extern GList *mime_parts;
extern void broken_cb (GtkWidget *widget, void *data);
extern void set_ok (GtkWidget *widget, gboolean *dialog_result);
extern gchar *mime_get_part (gchar *message, gint num, gint *len);
extern void free_message(Message *message);
extern void mark_readstatus(gint msg_id, gchar *val);
extern void address_drag_data_get (GtkWidget *widget, GdkDragContext *context, GtkSelectionData *selection_data, guint info, guint32 time);

static GnomeUIInfo mime_popup_menu [] = {
  { GNOME_APP_UI_ITEM, N_("Save Attachment"), N_("_Save attachment"), save_attach, NULL, NULL, GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_SAVE, 0, 0, NULL },
  GNOMEUIINFO_END
};

/* Standard DnD types */
enum { TARGET_EMAIL, };
extern GtkTargetEntry *email_field_drop_types;
#define ELEMENTS(x) (sizeof (x) / sizeof (x[0]))

/* 
 * Called when a user clicks on one of the mailbox names in the left pane.
 * 
 */
void
mboxclicked_cb (GtkWidget *mblist, gint row, gint column, GdkEventButton *event, gpointer data)
{

	/* Check for double click. */	
	if (event != NULL && event->button == 1 && event->type == GDK_2BUTTON_PRESS) {
		static const char *soundtrigger[] = {"gtk-events", "activate", NULL};
	
		gnome_triggers_vdo("", NULL, soundtrigger); /* Play sound */

		GmailApp->left_pos = row;

		vfolder_display_rightpane(row);
		}		
}


/* This code actually displays a mail message in the relevent pane.
 * It display the msg that is passed to it.
 * It also processes MIME attachments.
 * 
 * It is up to the calling function to free the memory used by the msg.
 *
 * At the moment we have a global variable called GmailApp->msgpane.
 * We assume the message always goes in here. If it doesn't exist we 
 * call create_msg_pane.
 * We rely on libglade to give us the widgets we need, so this code
 * will generalise easy to have multiple message windows.
 *
 * If forced is true it means the message is being displayed without the user
 * manually doing it. If so we don't mark the message as read or unbold it.
 */
void 
display_message(Message *msg, gboolean forced) 
{
	gint parts = 0;
	gchar *title;
	GList *mime;
	GladeXML *xml; 
	GtkWidget *wid;
		
	if (TwoPaneView()) {

		if (!GmailApp->msgpane) {
			DestroyRightPane();
			GmailApp->msgpane = create_msg_pane();
			gtk_paned_add2 (GTK_PANED(GmailApp->hpaned), GmailApp->msgpane); 
			}

	} else {
		/* We are in the infamous THREE PANE VIEW */
		/* Remove the current message that is in there. */

		if (!GmailApp->msgpane) {
			GmailApp->msgpane = create_msg_pane();
			gtk_paned_add2 (GTK_PANED(GmailApp->rightpane), GmailApp->msgpane); 
			}
	}

	xml = glade_get_widget_tree(GTK_WIDGET (GmailApp->msgpane));

	wid = glade_xml_get_widget(xml, "from"); 
	gtk_label_set_text(GTK_LABEL(wid), msg->from);
	gtk_widget_set_style (wid, GmailApp->boldstyle); 

	/* FIXME: Set up dnd source. 
  gtk_drag_source_set (wid, GDK_BUTTON1_MASK,
		     email_field_drop_types, ELEMENTS (email_field_drop_types),
		     GDK_ACTION_COPY); 
  gtk_signal_connect( GTK_OBJECT(wid->parent), "drag_data_get",
		      GTK_SIGNAL_FUNC(address_drag_data_get), NULL);  */

	wid = glade_xml_get_widget(xml, "to"); 
	gtk_label_set_text(GTK_LABEL(wid), msg->to);
	gtk_widget_set_style (wid, GmailApp->style); 

	wid = glade_xml_get_widget(xml, "subject"); 
	gtk_label_set_text(GTK_LABEL(wid), msg->subject);
	gtk_widget_set_style (wid, GmailApp->boldstyle); 

	wid = glade_xml_get_widget(xml, "date"); 
	gtk_label_set_text(GTK_LABEL(wid), msg->date);
	gtk_widget_set_style (wid, GmailApp->style); 

	/* Freeze, clear and insert the text message. */
	wid = glade_xml_get_widget(xml, "textmsg"); 
	gtk_text_freeze(GTK_TEXT (wid));
	gtk_widget_set_style (wid, GmailApp->style); 
	gtk_text_backward_delete(GTK_TEXT (wid), gtk_text_get_length(GTK_TEXT (wid)));
	/* Insert message into the textmsg box. */
  gtk_text_insert (GTK_TEXT (wid), NULL, NULL, NULL, msg->message->str, -1);
	gtk_text_thaw(GTK_TEXT (wid));

	/* Freeze, clear and insert the text headers. */
	wid = glade_xml_get_widget(xml, "textheaders"); 
	gtk_text_freeze(GTK_TEXT (wid));
	gtk_widget_set_style (wid, GmailApp->style); 
	gtk_text_backward_delete(GTK_TEXT (wid), gtk_text_get_length(GTK_TEXT (wid)));
	/* Insert message into the textmsg box. */
  gtk_text_insert (GTK_TEXT (wid), NULL, NULL, NULL, msg->headers->str, -1);
	gtk_text_thaw(GTK_TEXT (wid));


	/* Unbold the message in the vfolder list if it is unforced. */
	if (!forced) {
	gtk_clist_set_cell_style(GTK_CLIST (GmailApp->rightlist), GmailApp->right_pos, 0, GmailApp->style);
	gtk_clist_set_cell_style(GTK_CLIST (GmailApp->rightlist), GmailApp->right_pos, 1, GmailApp->style);
	gtk_clist_set_cell_style(GTK_CLIST (GmailApp->rightlist), GmailApp->right_pos, 2, GmailApp->style);
	gtk_clist_set_cell_style(GTK_CLIST (GmailApp->rightlist), GmailApp->right_pos, 3, GmailApp->style);

	/* For incoming msg's, set the clist read status to 'read'. */
	if (strcmp(msg->direction, "Incoming") == 0)  {
	gtk_clist_set_text(GTK_CLIST (GmailApp->rightlist), 
													GmailApp->right_pos, 0, "Read"); }

		}

	/* Parse MIME stuff. EXPERIMENTAL!. 
	 * For now only parse incoming messages.
   */
	if (strcmp(msg->direction, "Incoming") == 0) {
	msg->headers = g_string_append(msg->headers, msg->message->str);
	mime_parse(msg->headers->str);


	/* Loop the mime_parts. */
	wid = glade_xml_get_widget(xml, "mimeclist"); 
	gtk_clist_clear(GTK_CLIST(wid));
	mime = mime_parts;
	for (mime = mime_parts; mime != NULL; mime = mime->next) {
		gchar *datum[5];		/* temporary storage area for clist */
		Mime_Part *mpart;

		mpart = mime->data;
		parts++;

		datum[0] = g_strdup_printf("%i", parts);
		datum[1] = g_strdup_printf("%i", mpart->len); 
		datum[2] = g_strdup_printf("%s/%s", mpart->type, mpart->subtype);
		datum[3] = g_strdup_printf("%s",  mpart->encoding);
		datum[4] = g_strdup_printf("%s",  mpart->disposition);
	
   	gtk_clist_append (GTK_CLIST (wid), datum);

		g_free(datum[0]);
		g_free(datum[1]);
		g_free(datum[2]);
		g_free(datum[3]);
		g_free(datum[4]);
		}

	}

	gtk_widget_show_all (GmailApp->msgpane);

	/*
	 * Set the status bar message with the msg database id.
	 */
	title = g_strdup_printf("Msg id %i", msg->id);
	gnome_appbar_set_status (GNOME_APPBAR (GmailApp->appbar), title);
	g_free(title);

	/* Mark the message as read, now that we have displayed it. */
	/* We don't mark outgoing messages as read.. they are 'sent'. */
	if (!forced) 
 	if (strcmp(msg->direction, "Incoming") == 0) { mark_readstatus(msg->id, "Read"); }

	return;
}

/* Create the widgets for the msg_pane. 
 * This function is called by display_message.
 * We now don't rely on any global variables, but purely on libglade.
 */
GtkWidget * create_msg_pane(void)
{
	GtkWidget *label;
	GladeXML *xml;
	GtkWidget *glademsg;
	gchar *glade_src;

	/* Load the glade file. */
	glade_src = g_strdup_printf("%s/message.glade", GMAIL_GLADEDIR);
	xml = glade_xml_new(glade_src, "frame1");
	g_free(glade_src);

	if (!xml) {
		g_warning("Unable to load xml interface. Are the gmail glade files installed onto your system?\n");
		exit(1);
		}

	glade_xml_signal_autoconnect(xml);

	/* Grab our glade widgets. */
	glademsg = glade_xml_get_widget(xml, "frame1");

	/* FIXME: life would be easier if I knew how to set styles globally. */
	label = glade_xml_get_widget(xml, "fromlabel");
	gtk_widget_set_style (label, GmailApp->style); 

	label = glade_xml_get_widget(xml, "tolabel");
	gtk_widget_set_style (label, GmailApp->style); 

	label = glade_xml_get_widget(xml, "subjectlabel");
	gtk_widget_set_style (label, GmailApp->style); 

	label = glade_xml_get_widget(xml, "datelabel");
	gtk_widget_set_style (label, GmailApp->style); 


	/* FIXME: Do we need to free the xml reference here?. 
	 */

	return(glademsg); /* Return the msg frame. */
}


/* This function called when a user clicks on the 'Delete' toolbar button. 
 * Currently msg deletion is only supported for the mysql messages.
 * FIXME: Should we give them a warning before deleting a message?
 * The warning could contain some headers of the email, just to be sure.
 */
void			
delete_msg (GtkWidget *widget, gpointer data)
{
	gchar *query;
	gint id;
	static const char *soundtrigger[] = {"gmail", "delete", NULL};
	Message *mesg = NULL;

	if (GTK_CLIST(GmailApp->rightlist)->rows == 0) return;

	if (GmailApp->right_pos < 0) {
		/* Nothing to delete. */
		return;
		}

  id = GmailApp->msg_id;

	/* Open Database. */
	if (!open_db(&GmailApp->mysql)) { return; };

	query = g_strdup_printf("DELETE FROM display WHERE id = %i", id);

	if (mysql_query(&GmailApp->mysql, query) != 0) {
		/* FIXME: send a dialog if the delete failed. */
		g_print("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		return;
	} 

	g_free(query);

	query = g_strdup_printf("DELETE FROM details WHERE id = %i", id);
	if (mysql_query(&GmailApp->mysql, query) != 0) {
		/* FIXME: send a dialog if the delete failed. */
		g_print("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		return;
	} 

	g_free(query);

	/* delete the row from the clist. */
	if (GTK_IS_CLIST(GmailApp->rightlist))
	gtk_clist_remove(GTK_CLIST(GmailApp->rightlist), GmailApp->right_pos);	

	gnome_triggers_vdo("", NULL, soundtrigger); /* Play sound */

	/* If there are no more messages, return. */
	if (GTK_CLIST(GmailApp->rightlist)->rows == 0){
		GmailApp->right_pos = -1; /* Nothing selected */
		return;
		}

	/* select the next msg. then we can delete stacks in a row. */
	gtk_clist_select_row(GTK_CLIST (GmailApp->rightlist), GmailApp->right_pos, 0);

	/* If we are in three pane mode display the email. */
	if (!TwoPaneView()) {
		mesg = load_vfolder_msg(GmailApp->msg_id);
		if (mesg != NULL) { display_message(mesg, TRUE); }
		free_message(mesg); /* Free the memory */
	}

}

/* nextmsg_cb
 * If a message is opened, this will open the next message in that mailbox. 
 *
 * FIXME: need to keep updating the rightlist if we are in 3 pane mode.
 */
void			
nextmsg_cb (GtkWidget *widget, gint row, gint column, GdkEventButton *event, gpointer data) 
{
	gint id;
	Mailbox *mailbox;
	GtkWidget *clist;
	Message *msg = NULL;

	if (GTK_CLIST(GmailApp->rightlist)->rows == 0) return;

 	mailbox = gtk_clist_get_row_data(GTK_CLIST (GmailApp->mblist), GmailApp->left_pos);

	clist = GmailApp->rightlist;

	g_return_if_fail(GTK_IS_CLIST(clist));

	/* Check boundaries. */
	g_return_if_fail(GmailApp->right_pos >= 0);
	g_return_if_fail(GmailApp->right_pos < GTK_CLIST(clist)->rows - 1);


	GmailApp->right_pos++;
  id = (gint) gtk_clist_get_row_data(GTK_CLIST (clist), GmailApp->right_pos);

	/* change the selection as well. */
	gtk_clist_select_row(GTK_CLIST (clist), GmailApp->right_pos, 0);
	gtk_clist_moveto(GTK_CLIST (clist), GmailApp->right_pos , 0, 0.5, 0);

	msg = load_vfolder_msg(id);
	if (msg == NULL) return;
	display_message(msg, FALSE); 
	free_message(msg); /* Free the memory */
}

/* prevmsg_cb
 * If a message is opened, this will open the previous message in that mailbox. 
 */
void			
prevmsg_cb (GtkWidget *widget, gint row, gint column, GdkEventButton *event, gpointer data) 
{
	gint id;
	Mailbox *mailbox;
	GtkWidget *clist;
	Message *msg = NULL;

	if (GTK_CLIST(GmailApp->rightlist)->rows == 0) return;

 	mailbox = gtk_clist_get_row_data(GTK_CLIST (GmailApp->mblist), GmailApp->left_pos);

	clist = GmailApp->rightlist;

	g_return_if_fail(GTK_IS_CLIST(clist));

	/* Check boundaries. */
	g_return_if_fail(GmailApp->right_pos > 0);
	g_return_if_fail(GmailApp->right_pos < GTK_CLIST(clist)->rows);

	GmailApp->right_pos--;
  id = (gint) gtk_clist_get_row_data(GTK_CLIST (clist), GmailApp->right_pos);
	
	/* change the selection as well. */
	gtk_clist_select_row(GTK_CLIST (clist), GmailApp->right_pos, 0);
	gtk_clist_moveto(GTK_CLIST (clist), GmailApp->right_pos , 0, 0.5, 0);

	msg = load_vfolder_msg(id);
	if (msg == NULL) return;

	display_message(msg, FALSE); 
	free_message(msg); /* Free the memory */
}

/* Sort a column.
 * This sorts the column by modifying the SQL query to order by the field
 * that the user is clicking on.
 * It's a hack.
 * Nevertheless there are two main advantages:
 * 1. Order is remembered between sessions.
 * 2. The database does the job of ordering fields and not us.
 * 
 */
void
columnclicked_cb (GtkCList * clist, gint column) 
{
  gchar *field = NULL;
	Mailbox *mailbox;
	gint selected = -1;
	gpointer msg_data;

  mailbox = gtk_clist_get_row_data(GTK_CLIST (GmailApp->mblist), GmailApp->left_pos);
  msg_data = gtk_clist_get_row_data(GTK_CLIST (GmailApp->rightlist), GmailApp->right_pos);

	/* Remember the msg id of their selection. */
	if (GmailApp->right_pos != -1)
		selected = (gint) gtk_clist_get_row_data(GTK_CLIST(GmailApp->rightlist), GmailApp->right_pos);	

	/* Find which column was clicked. */
  /* FIXME: these won't be hardcoded forever. */
	switch (column) {
		case 0:
			field = g_strdup("readstatus");
			break;
		case 1:
			field = g_strdup("fromfield");
			break;
		case 2:
			field = g_strdup("subject");
			break;
		case 3:
			field = g_strdup("date");
			break;
		default:
			return;
		}

	if (strcmp(GmailApp->sortfield, field) == 0) {
		if (strcmp(GmailApp->sortorder, "ASC") == 0) {
			g_free(GmailApp->sortorder);
			GmailApp->sortorder = g_strdup("DESC");
		} else {
			g_free(GmailApp->sortorder);
			GmailApp->sortorder = g_strdup("ASC");
		}
	} else {
		g_free(GmailApp->sortfield);
		GmailApp->sortfield = g_strdup(field); 
		GmailApp->sortorder = g_strdup("DESC");
	}

	vfolder_display_rightpane(GmailApp->left_pos); 

	/* FIXME Reselect the row that was previously highlighted, if any. 
   * This code works, but the behaviour doesn't seem intuitive.
   * Think about this more.
   *
	GmailApp->right_pos = gtk_clist_find_row_from_data(GTK_CLIST(GmailApp->rightlist), msg_data);
	gtk_clist_select_row(GTK_CLIST(GmailApp->rightlist), GmailApp->right_pos, 0);
	gtk_clist_moveto(GTK_CLIST (GmailApp->rightlist), GmailApp->right_pos, 0, 0.5, 0); */

	g_free(field);
}

/* Handle rightclick popup menus on mime attachments. */
void
attachment_button_press_event (GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	GtkWidget *menu;
	gint result;
	gint row, col;

	if (event->button != 3) return;

	gtk_clist_get_selection_info(GTK_CLIST(widget), event->x, event->y, &row, &col);
	gtk_clist_select_row(GTK_CLIST(widget), row, col);

  menu = gnome_popup_menu_new (mime_popup_menu);
	result = gnome_popup_menu_do_popup_modal (menu, NULL, NULL, event, GINT_TO_POINTER(row));
  gtk_widget_destroy (menu);
}

/* Save attachment to a file.
 * 
 */
void
save_attach (GtkWidget *widget, gint row)
{
	FILE *fd;
	GtkFileSelection *fsel;
	gboolean accepted = FALSE;
	gint id;
	gchar *title;

  id = GmailApp->msg_id;

	if (id < 0) return;

	title = g_strdup_printf("Save attachment as");
	fsel = GTK_FILE_SELECTION (gtk_file_selection_new (title));
	g_free(title);

  gtk_window_set_modal (GTK_WINDOW (fsel), TRUE);

	/* Connect the signals for Ok and Cancel */
  gtk_signal_connect (GTK_OBJECT (fsel->ok_button), "clicked",
          GTK_SIGNAL_FUNC (set_ok), &accepted);
  gtk_signal_connect(GTK_OBJECT (fsel->cancel_button), "clicked",
          GTK_SIGNAL_FUNC (gtk_main_quit), NULL);

	gtk_widget_show(GTK_WIDGET(fsel));
	gtk_grab_add (GTK_WIDGET (fsel));
	gtk_main(); 

	if (accepted) {
		char *name = NULL;
		Message *msg = NULL;
		gchar *attach;
		gint len;

		/* Load the message. */
		msg = load_vfolder_msg(id); if (msg == NULL) return;

		/* Put the whole message back together in the headers gstring. */
		msg->headers = g_string_append(msg->headers, msg->message->str);

		mime_parse(msg->headers->str); /* mime parse the whole message. */

		/* Get the attachment we want (and remember the length). */
		attach = mime_get_part((gchar *) msg->headers->str, (gint) row+1, &len);

		name = gtk_file_selection_get_filename (fsel);
		g_print("Saving msg to file %s\n", name);

		/* Open and write the file. */
		if ((fd = fopen(name, "wb"))) {
			fwrite(attach, 1, len, fd);
			fclose(fd);
		} else {
			/* FIXME: Give GUI response. */
			fprintf(stderr, _("Couldn't open %s\n"), name);
		}

		/* Free memory. */
		g_free(attach);
		free_message(msg); /* Free the memory */
		gtk_widget_destroy (GTK_WIDGET (fsel));

	}


	return;
}

