/*
 * Gmail. A Gnome email client.
 * Copyright (C) 1999 Wayne Schuller
 *
 * mysql.c - functions which involve talking to mysql server.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "main.h"

extern _GmailApp_	*GmailApp;		/* Global vars live here. */

extern void		columnclicked_cb (GtkCList *clist, gint column);
extern void broken_cb (GtkWidget *widget, void *data);
extern void	delete_msg (GtkWidget *widget, gpointer data);
extern void mark_as_matched(gint id, GString *vfolders, gboolean overwrite); 
extern void check_vfolder_matches(gint id, DTimer *timer);
extern gboolean				TwoPaneView();
extern	gchar * strip_name(gchar *to);
extern	gboolean				DestroyRightPane(void);
extern void address_drag_data_get (GtkWidget *widget, GdkDragContext *context, GtkSelectionData *selection_data, guint info, guint32 time);


void vfolder_display_rightpane(gint row);
Message * load_vfolder_msg(gint id);
gboolean add_msg_to_db(Message *message, DTimer *timer);
void open_vfolder_msg (GtkWidget *clist, gint row, gint column, GdkEventButton *event, gpointer data);
gboolean parse_pop3_msg(GString *msg, gchar *uid, DTimer *timer);
GString * setup_db (gchar *ruser, gchar *rpass, gchar *user, gchar *pass, gchar *hostname, gchar *db);
gboolean uid_present(gchar *uid);
DTimer * msg_match_popup (void);
gint progress_timeout(DTimer *timer);
void destroy_dtimeout(GtkWidget *widget, DTimer *timer);
gboolean vfolder_msg_unread(gint id);
void vfolder_contains_unread(gchar *vfolder, gboolean unread_status);
void columnresize_cb(GtkCList *clist, gint column, gint width, gpointer user_data);
void rightlist_popup (GtkWidget *widget, GdkEventButton *event, gpointer data);
void tag_msg (GtkWidget *widget, gpointer data);
void ctx_open_msg (GtkWidget *widget, gpointer data);
void ctx_open_msg_adv (GtkWidget *widget, gpointer data);
gchar * date822_to_iso(gchar *date);
gint message_count (void);
void free_message(Message *message);
void mark_readstatus(gint msg_id, gchar *val);
static void rightlist_clist_create(void);

gboolean open_db(MYSQL *mysql);

/* main.c */
extern void save_msg (GtkWidget *widget, void *data);

extern	void	display_message(Message *msg, gboolean forced);

/* Standard DnD types */
enum { TARGET_EMAIL, };
extern GtkTargetEntry *email_field_drop_types;
#define ELEMENTS(x) (sizeof (x) / sizeof (x[0]))


static GnomeUIInfo rightlist_popup_menu [] = {
  { GNOME_APP_UI_ITEM, N_("Message options:"), N_("Message options"), NULL, NULL, NULL, GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_MAIL, 0, 0, NULL },
  GNOMEUIINFO_SEPARATOR,
  { GNOME_APP_UI_ITEM, N_("_Open"), N_("_Open message."), ctx_open_msg, NULL, NULL, GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_OPEN, 0, 0, NULL },
  { GNOME_APP_UI_ITEM, N_("_Save"), N_("_Save message."), save_msg, NULL, NULL, GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_SAVE, 0, 0, NULL },
  { GNOME_APP_UI_ITEM, N_("_Delete"), N_("_Delete message."), delete_msg, NULL, NULL, GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_TRASH, 0, 0, NULL },
  GNOMEUIINFO_END
};

/* open_db - Does the common code for opening the database. 
 * Returns TRUE for success, and FALSE for failure.
 *
 */
gboolean
open_db (MYSQL *mysql)
{

	/* Call mysql_init */
	if (mysql_init(mysql) == NULL) {
		GtkWidget *dialog;
		gchar *msg;
		msg = g_strdup_printf ("MYSQL Init failed:\n%s\nYou cannot continue until this error is fixed.\n", mysql_error(mysql));
		dialog = gnome_message_box_new(msg, GNOME_MESSAGE_BOX_ERROR, "doh!", 0);
		gtk_window_set_modal(GTK_WINDOW (dialog), TRUE);
 		gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
 		gtk_widget_show(dialog);
		g_free(msg);	
		return(FALSE);
		}

	if (!mysql_real_connect(mysql, GmailApp->hostname, GmailApp->username, GmailApp->password, GmailApp->dbname, 0, NULL, 0)) {
		GtkWidget *dialog;
		gchar *msg;
		msg = g_strdup_printf ("MYSQL Connection Failed:\n%s\nYou cannot continue until this error is fixed.\n", mysql_error(mysql));
		dialog = gnome_message_box_new(msg, GNOME_MESSAGE_BOX_ERROR, "doh!", 0);
		gtk_window_set_modal(GTK_WINDOW (dialog), TRUE);
 		gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
 		gtk_widget_show(dialog);
		g_free(msg);	
		mysql_close(mysql);
		return(FALSE);
		} 

	return(TRUE);
}

/* vfolder_display_rightpane - show vfolder contents in rightpane.
 * This function is passed an sql query and will display the results
 * of that query in the right pane.
 *
 * As of 0.4.8 we now use GmailApp->rightlist to always hold this list.
 * We tried using a seperate clist for every vfolder, but it wasn't worth
 * the effort.
 */
void 
vfolder_display_rightpane(gint row)
{
	MYSQL_RES *res;
	MYSQL_ROW mysqlrow;
	gchar *datum[4];		/* temporary storage area. */
	gint	*id;						/* Message id */
	gboolean unread_found = FALSE, cache_vfolders = FALSE;
	gchar *name, *query, *newquery;
	Mailbox	*mailbox;
	gchar *msg, *cachevf;
	Message *mesg = NULL;

	/* Get the vfolder details we are opening.*/
	/* FIXME: If we were passed a bad row, simply display the first vfolder. */
	gtk_clist_get_text(GTK_CLIST(GmailApp->mblist), row, 0, &name);
	mailbox = gtk_clist_get_row_data(GTK_CLIST(GmailApp->mblist), row);	
	query = mailbox->query;

	gnome_appbar_set_progress (GNOME_APPBAR(GmailApp->appbar), 0.1);

	/* Open Database. */
	if (!open_db(&GmailApp->mysql)) { return; };

	/* Find out if we are caching vfolders. */
	cachevf = gnome_config_get_string("/gmail/UserInfo/CacheVfolders");
	if (cachevf != NULL && (strcmp(cachevf, "TRUE") == 0)) 
		cache_vfolders = TRUE; 
	else
		cache_vfolders = FALSE; 

	g_free(cachevf);

	/* Check that query is valid. */
	if (query == NULL || strlen(query) < 2) {
		GtkWidget *dialog;
		gchar *msg;
		msg = g_strdup_printf ("Vfolder contains empty query.\n");
		dialog = gnome_message_box_new(msg, GNOME_MESSAGE_BOX_ERROR, "doh!", 0);
		gtk_window_set_modal(GTK_WINDOW (dialog), TRUE);
 		gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
 		gtk_widget_show(dialog);
		g_free(msg);	
		g_warning("vfolder_display_rightpane passed bad data.\n");
		mysql_close(&GmailApp->mysql);
		return;
		}

	/* 
	 * Create the query to call to generate the vfolder.
	 * If we're not caching, run the true vfolder query.
	 * If we are caching, just search on the matched index.
	 *
	 * Exceptions:
	 * 	- Unread should always be done live.
	 *
	 * NOTE: if we're not caching we update the matched index on the fly.
	 */	
	if ((strcmp(name, "Unread") == 0) || !cache_vfolders) 
	newquery = g_strdup_printf("SELECT readstatus,fromfield,subject,date,display.id FROM display,details WHERE display.id = details.id AND (%s) ORDER BY %s %s", query, GmailApp->sortfield, GmailApp->sortorder);
	else
	newquery = g_strdup_printf("SELECT readstatus,fromfield,subject,date,id FROM display WHERE matched LIKE '%%%s%%' ORDER BY %s %s", name, GmailApp->sortfield, GmailApp->sortorder);


	if (mysql_query(&GmailApp->mysql, newquery) != 0) {
		GtkWidget *dialog;
		gchar *msg;
		msg = g_strdup_printf ("MYSQL Query failed:\n%s\n", mysql_error(&GmailApp->mysql));
		dialog = gnome_message_box_new(msg, GNOME_MESSAGE_BOX_ERROR, "doh!", 0);
		gtk_window_set_modal(GTK_WINDOW (dialog), TRUE);
 		gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
 		gtk_widget_show(dialog);
		g_free(msg);	
		g_print("Failed query was: %s\n", newquery);
		g_free(newquery);
		mysql_close(&GmailApp->mysql);
		return;
		} 

	g_free(newquery);

	res = mysql_use_result(&GmailApp->mysql);

	gnome_appbar_set_progress (GNOME_APPBAR(GmailApp->appbar), 0.2);
     

	/* If we already have the clist , just clear it. 
	 * Reuse this widget to avoid flicker.
	 */
	if (GTK_IS_CLIST(GmailApp->rightlist)) {
		gtk_clist_clear(GTK_CLIST(GmailApp->rightlist));
		} else {
		rightlist_clist_create(); /* Create the rightlist. */
		}

	/* Destroy whatever is in the rightpane. Even if it is already a vfolder */
	if (TwoPaneView()) { DestroyRightPane(); }

	/* Setup scrolled window for the first time. */
	if (!GTK_IS_WIDGET(GmailApp->msglistpane)) {

	  /* Create a scrolled window to hold the clist. */
  	GmailApp->msglistpane = gtk_scrolled_window_new (NULL, NULL);
		gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (GmailApp->msglistpane),
									GTK_POLICY_AUTOMATIC, 
									GTK_POLICY_AUTOMATIC);    

		gtk_container_add(GTK_CONTAINER(GmailApp->msglistpane), GmailApp->rightlist);
		if (TwoPaneView()) {
			gtk_paned_add2 (GTK_PANED(GmailApp->hpaned), GmailApp->msglistpane);
			} else {
			gtk_paned_add1 (GTK_PANED(GmailApp->rightpane), GmailApp->msglistpane);
			} 
		}	


		gtk_widget_show_all (GmailApp->msglistpane);

	gnome_appbar_set_status (GNOME_APPBAR (GmailApp->appbar), "Updating clist.");
	gnome_appbar_refresh(GNOME_APPBAR (GmailApp->appbar));

	/* Add each row in the result as a clist item. */
	gtk_clist_freeze(GTK_CLIST(GmailApp->rightlist));
	while ((mysqlrow = mysql_fetch_row(res))) {

		datum[0] = mysqlrow[0];
		datum[1] = mysqlrow[1];
		datum[2] = mysqlrow[2];
		datum[3] = mysqlrow[3]; 

		/* We use prepend as it is apparently O(1). */
   	gtk_clist_prepend (GTK_CLIST (GmailApp->rightlist), datum);

		/* Set unread messages to look a bit different. */
		if (strncmp(datum[0], "Unread", 6) == 0) {	
			gtk_clist_set_cell_style(GTK_CLIST (GmailApp->rightlist), 0, 0, GmailApp->boldstyle);
			gtk_clist_set_cell_style(GTK_CLIST (GmailApp->rightlist), 0, 1, GmailApp->boldstyle);
			gtk_clist_set_cell_style(GTK_CLIST (GmailApp->rightlist), 0, 2, GmailApp->boldstyle);
			gtk_clist_set_cell_style(GTK_CLIST (GmailApp->rightlist), 0, 3, GmailApp->boldstyle);
			unread_found = TRUE;
			} 

		id = g_malloc(sizeof(gint)); 
		id = (gint *) atoi(mysqlrow[4]);

		/* Store the mysql id of the email as the row data. */
    gtk_clist_set_row_data(GTK_CLIST (GmailApp->rightlist), 0, (gpointer) id);

		/* When we are not caching, we also update the matched field
		 * on the run. This gives a speed hit, but if they want speed
		 * they should be caching.
		 */
		if (!cache_vfolders) {
			GString *vname = g_string_new(name);	
			if (vname != NULL && atoi(mysqlrow[4]) > 0)
			 	mark_as_matched(atoi(mysqlrow[4]), vname, FALSE);  
			g_string_free(vname, FALSE);
			}

		}
	gtk_clist_thaw(GTK_CLIST(GmailApp->rightlist));


	mysql_free_result(res); 
	mysql_close(&GmailApp->mysql); 
	gnome_appbar_set_progress (GNOME_APPBAR(GmailApp->appbar), 0.5);

	/* Update vfolders unread status. */
	vfolder_contains_unread(name, unread_found);

	GmailApp->right_pos = -1; /* Nothing in new clist selected yet. */

	/* Go to the row that was last selected. */
	if (GTK_CLIST(GmailApp->rightlist)->rows >= 0)
		gtk_clist_moveto(GTK_CLIST (GmailApp->rightlist), GmailApp->right_pos, 0, 0.5, 0);

	/* Let do some redrawing on the screen now */
	gnome_appbar_set_progress (GNOME_APPBAR(GmailApp->appbar), 0.6);

	gtk_window_set_title(GTK_WINDOW (GmailApp->app), "Gmail - Mysql vfolder");
	gnome_appbar_set_progress (GNOME_APPBAR(GmailApp->appbar), 1);

	/* Highlight the first message. */
	gtk_clist_select_row(GTK_CLIST(GmailApp->rightlist), 0, 0); 

	/* If we are in three pane mode select the first email. */
	if (!TwoPaneView() && gnome_config_get_bool_with_default("/gmail/UserInfo/display_first_message=false", NULL)) {
		mesg = load_vfolder_msg(GmailApp->msg_id);
		if (mesg != NULL) { 
			display_message(mesg, TRUE); 
			free_message(mesg); /* Free the memory */
			}
		GmailApp->right_pos = 0; /* Nothing in new clist selected yet. */
	}

	/* Display info about the vfolder in the appbar. */
	msg = g_strdup_printf("Vfolder '%s' contains %i messages.", name, GTK_CLIST(GmailApp->rightlist)->rows);
	gnome_appbar_set_status (GNOME_APPBAR (GmailApp->appbar), msg);
	g_free(msg);


}


/* add_msg_to_db - Add a message to the mysql database. 
 * Is passed a single Message type.
 * Returns TRUE for success, FALSE for failure.
 * Everything should have funny \r\n's filtered out but we need to
 * escape most of the fields cos nearly everything can have a ' or " in it.
 * Some people have their names bracketed with one or the other... annoying.
 *
 * It is up to the calling function to free the memory in the message.
 * This function doesn't free the memory used by the message.
 * It will free only the memory it has allocated for use.
 *
 * FIXME: do we need error checking on the message?
 */
gboolean
add_msg_to_db(Message *message, DTimer *timer)
{
	gint	tmp, id, len;
	gchar *query, *headers, *msg, *subject, *to, *from, *cc, *uid;

	/* Open Database. */
	if (!open_db(&GmailApp->mysql)) { return FALSE; };

	/* Escape the content of the message so newlines and ' symbols don't
   * confuse our query.
	 * Mysql docs say we need length*2+1 reserved for an escaped string.
	 */
	msg = g_malloc(sizeof(gchar) * ((2 * message->message->len) + 1));
	mysql_escape_string(msg, message->message->str, message->message->len);
	headers = g_malloc(sizeof(gchar) * ((2 * message->headers->len) + 1));
	mysql_escape_string(headers, message->headers->str, message->headers->len);
	len = strlen(message->subject);
	subject = g_malloc(sizeof(gchar) * ((2 * len) + 1));
	mysql_escape_string(subject, message->subject, len);
	len = strlen(message->from);
	from = g_malloc(sizeof(gchar) * ((2 * len) + 1));
	mysql_escape_string(from, message->from, len);
	len = strlen(message->to);
	to = g_malloc(sizeof(gchar) * ((2 * len) + 1));
	mysql_escape_string(to, message->to, len);
	len = strlen(message->cc);
	cc = g_malloc(sizeof(gchar) * ((2 * len) + 1));
	mysql_escape_string(cc, message->cc, len);
	len = strlen(message->uid);
	uid = g_malloc(sizeof(gchar) * ((2 * len) + 1));
	mysql_escape_string(uid, message->uid, len);

	if (headers == NULL) headers = g_strdup("Parse Error");
	if (msg == NULL) msg = g_strdup("Parse Error");

	/* If we have the date put quotes around it, else use NOW() */
	if (message->date != NULL) {
		gchar *tmp;
		tmp = g_strdup_printf("'%s'", message->date);
		g_free(message->date);
		message->date = tmp;
	} else
		message->date = g_strdup("NOW()");

	/* Insert data into 'display' table. */
	query = g_strdup_printf("INSERT INTO display (date, subject, fromfield, readstatus, direction, uid) VALUES (%s, '%s' , '%s' , '%s', '%s', '%s')", 
								message->date,
								subject,	
								from,	
								message->readstatus,	
								message->direction,
								uid);


	if (mysql_real_query(&GmailApp->mysql, query, strlen(query)) != 0) {
		g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		g_print("query was: |%s|\n", query);
		mysql_close(&GmailApp->mysql);
		return(FALSE);
		} 


	/* It _should_ have only affected one row. */	
	tmp = mysql_affected_rows(&GmailApp->mysql);
	if (tmp == 0) {
			/* Affecting zero rows is considered failure. */
			g_warning("add_msg_to_db affect %d rows...", tmp);
			g_free(query);
			mysql_close(&GmailApp->mysql);
			return FALSE;
		} else if (tmp > 1) {
			/* Affecting more than one row on an INSERT is just plain weird.
       * Better to return success to make the user think that open source
			 * software is reliable. 
       */
			g_warning("add_msg_to_db affected %d rows. (It should have only effected one.", tmp);
	}	

	g_free(query);

	/* Find the database id of the message we just put in. */
	id = mysql_insert_id(&GmailApp->mysql);

	/* Insert data into 'details' table. */
	query = g_strdup_printf("INSERT INTO details (id, headers, message, tofield, ccfield) VALUES ('%d', '%s' , '%s' , '%s', '%s')", 
							id,
							headers,
							msg,
							to,
							cc);

	if (mysql_real_query(&GmailApp->mysql, query, strlen(query)) != 0) {
		g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		g_print("query was: |%s|\n", query);
		mysql_close(&GmailApp->mysql);
		return(FALSE);
		} 

	/* It _should_ have only affected one row. */	
	tmp = mysql_affected_rows(&GmailApp->mysql);
	if (tmp == 0) {
			/* Affecting zero rows is considered failure. */
			g_warning("add_msg_to_db affect %d rows...", tmp);
			g_free(query);
			mysql_close(&GmailApp->mysql);
			return FALSE;
		} else if (tmp > 1) {
			/* Affecting more than one row on an INSERT is just plain weird.
			 * Better to return success to continue the illusion that 
			 * open source software is reliable. 
			 */
			g_warning("add_msg_to_db affected %d rows. (It should have only effected one.", tmp);
	}	

	mysql_close(&GmailApp->mysql);
	g_free(msg);
	g_free(headers);
	g_free(subject);
	g_free(from);
	g_free(to);
	g_free(cc);
	g_free(uid);
	g_free(query);


	/* Update the matched index for this message. */
	check_vfolder_matches(id, timer); 

	return(TRUE);
}

/* open_vfolder_msg - Open a vfolder message in the right pane.
 * This function is called when the user double clicks on a message
 * in a vfolder message list in the right pane.
 * It gets the message id from the clist row data.
 *
 * If we are in three pane view we display even if they don't double click.
 * FIXME: above functionality should be an option.
 */
void
open_vfolder_msg (GtkWidget *clist, gint row, gint column, GdkEventButton *event, gpointer data)
{
	gint id;
	Mailbox	*mailbox;
	gchar *title, *strip;
	Message *msg = NULL;
	
	mailbox = gtk_clist_get_row_data(GTK_CLIST(GmailApp->mblist), GmailApp->left_pos);	

	GmailApp->right_pos = row;

  id = (gint) gtk_clist_get_row_data(GTK_CLIST (clist), row);

	GmailApp->msg_id = id;

	/* Load the msg details so we can set the title, display it etc. */
	msg = load_vfolder_msg(id);

	/*
	 * Set the status bar message with the msg database id.
	 */
	title = g_strdup_printf("Msg id %i", msg->id);
	gnome_appbar_set_status (GNOME_APPBAR (GmailApp->appbar), title);
	g_free(title);

	/*
	 * Set the window title with from/subject. 
	 */
	strip = strip_name(msg->from);
	title = g_strdup_printf("Gmail - From: %s, Subject: %s", strip, msg->subject);
	gtk_window_set_title(GTK_WINDOW (GmailApp->app), title);
	g_free(title);
	g_free(strip);


	/* Display message */
	if (msg != NULL && event != NULL && event->button == 1) {
		if (event->type == GDK_2BUTTON_PRESS || !TwoPaneView()) {
				display_message(msg, FALSE);
				}
		}

	free_message(msg); /* Free the memory */
}

/* We need two versions of the next function because gnome_popup menu's
 * don't pass on the data. It's a gnome-libs thing.
 */
void ctx_open_msg (GtkWidget *widget, gpointer data)
{
	Message *msg = NULL;

	msg = load_vfolder_msg(GmailApp->msg_id);
	if (msg != NULL) {
		display_message(msg, FALSE);
	}
	free_message(msg); /* Free the memory */
}

void ctx_open_msg_adv (GtkWidget *widget, gpointer data)
{
	Message *msg = NULL;

	msg = load_vfolder_msg(GmailApp->msg_id);
	if (msg != NULL) {
		display_message(msg, FALSE);
	}
	free_message(msg); /* Free the memory */
}

/* load_vfolder_msg - Load a message based on a mysql id.
 * This doesn't display the message. You have to call display_message to do
 * that.
 * It returns the message, or NULL if there is a problem. 
 */
Message *
load_vfolder_msg(gint id)
{
	MYSQL_RES *res;
	MYSQL_ROW mysqlrow;
	gchar *query;
	gboolean outgoing = FALSE;
	Message *msg = NULL;

	/* Open Database. */
	if (!open_db(&GmailApp->mysql)) { return(NULL); };

	/* Initialise the message.*/
	msg = g_malloc(sizeof(Message));
	msg->id = id;
	msg->headers = NULL;
	msg->message = NULL;
	msg->date = NULL;
	msg->subject = NULL;
	msg->from = NULL;
	msg->to = NULL;
	msg->cc = NULL;
	msg->bcc = NULL;
	msg->uid = NULL;
	msg->readstatus = NULL;
	msg->direction = NULL;

	/* Get info from display table. */
	query = g_strdup_printf("SELECT fromfield,subject,date,direction FROM display WHERE id = %i", id);

	if (mysql_query(&GmailApp->mysql, query) != 0) {
		g_print("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		g_print("Query was: %s\n", query);
		/* Close the database connection. */
		mysql_close(&GmailApp->mysql);
		return(NULL);
	} 

	res = mysql_use_result(&GmailApp->mysql);

	if ((mysqlrow = mysql_fetch_row(res))) {

		/* FIXME: Do error checking on the result.. bad data causes core-dumps. */

		msg->from = g_strdup(mysqlrow[0]);		
		msg->subject = g_strdup(mysqlrow[1]);		
		msg->date = g_strdup(mysqlrow[2]);		
		msg->direction = g_strdup(mysqlrow[3]);		
		if (strcmp(mysqlrow[3], "Outgoing") == 0) outgoing = TRUE; 
	} else { 
			/* The query didn't have any results. The msg_id isn't in the database */
			return(NULL);
	}

	mysql_free_result(res); 

	g_free(query);

	/* Get info from details table. */
	query = g_strdup_printf("SELECT tofield,ccfield,message,headers FROM details WHERE id = %i", id);

	/* Query the database for the email. */
	if (mysql_query(&GmailApp->mysql, query) != 0) {
		g_print("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		g_print("Query was: %s\n", query);
		/* Close the database connection. */
		mysql_close(&GmailApp->mysql);
		return(NULL);
	} 

	res = mysql_use_result(&GmailApp->mysql);

	if ((mysqlrow = mysql_fetch_row(res))) {

		msg->to = g_strdup(mysqlrow[0]);		

		if ((mysqlrow[1] != NULL) && (strlen(mysqlrow[1]) > 2)) {
			msg->cc = g_strdup(mysqlrow[1]);		
			} else msg->cc = NULL;

		msg->message = g_string_new(mysqlrow[2]);
		msg->headers = g_string_new(mysqlrow[3]);

	} else { 
		/* The query didn't have any results. The msg_id isn't in the database. */	
		return(NULL);
	}

	g_free(query);
	mysql_free_result(res); 

	/* Close the database connection. */
	mysql_close(&GmailApp->mysql);

	return(msg);
}


/* parse_pop3_msg - Parse a new msg and insert into database.
 * This function is called by the pop3 code as it downloads messages.
 * It does very poor parsing of the message. The root of all download
 * problems are usually in here. Boy if we could get this function tidied
 * up we would have one sweet email program.
 * Note: we don't want the specially parsed fields to have newlines in them.
 * These are: to, subject, from, etc.
 * Only headers, message and attachments have newlines.
 */

gboolean
parse_pop3_msg(GString *msg, gchar *uid, DTimer *timer)
{
	Message *message;
	gchar 	*line;
	gchar   *ptr;
	gboolean ret = TRUE;

	/* We're going to parse everything into a Message structure.
     * Let just initialise it with some dummy values. 
     */
	message = g_malloc(sizeof(Message));
	message->headers = g_string_new(NULL);
	message->message = g_string_new(NULL);
	message->date = NULL;
	message->subject = NULL;
	message->from = NULL;
	message->to = NULL;
	message->cc = NULL;
	message->bcc = NULL;
	message->readstatus = NULL;
	message->direction = NULL;

	/* Lets parse grabbing only the special fields. */
	line = strdup(msg->str);

	/* FIXME: strok considered harmful. */
	line = strtok(line, "\r\n");
	while (line) {

		if (message->from == NULL && strncmp(line, "From:", 5) == 0) {
			message->from = g_malloc(strlen(line) * sizeof(gchar));
			strcpy((char *) message->from, (const char *) line+(6*sizeof(gchar)));
			g_strstrip(message->from);

			} else if (message->date == NULL && strncmp(line, "Date:", 5) == 0) {
			gchar *date;

			date = g_malloc(strlen(line) * sizeof(gchar));
			strcpy((char *) date, (const char *) line+(6*sizeof(gchar)));
			g_strstrip(date);

			message->date = date822_to_iso(date);

			g_free(date);

			} else if (message->subject == NULL && strncmp(line, "Subject:",8) == 0) {

			/* Allow for case with subject line but blank value. */
			if (strlen(line) < 9) 
				message->subject = g_strdup("(no subject)");
			else {
				message->subject = g_malloc(strlen(line) * sizeof(gchar));
				strcpy((char *) message->subject, (const char *) line+(9*sizeof(gchar)));
				g_strstrip(message->subject);
				}

			} else if (message->cc == NULL && (strncmp(line, "Cc:", 3) == 0 || strncmp(line, "cc:", 3) == 0 || strncmp(line, "CC:", 3) == 0)) {

			message->cc = g_malloc(strlen(line) * sizeof(gchar));
			strcpy((char *) message->cc, (const char *) line+(3*sizeof(gchar)));
			g_strstrip(message->cc);

			} else if (message->to == NULL && (strncmp(line, "To:",3) == 0)) {

			message->to = g_malloc(strlen(line) * sizeof(gchar));
			strcpy((char *) message->to, (const char *) line+(3*sizeof(gchar)));
			g_strstrip(message->to);

			} 
		line = strtok(NULL, "\r\n");
		} 

	/* Now lets parse again putting the grabbing the headers of the text. 
	 * This code + algorithm taken from Spruce! 
   */ 
	for (ptr = msg->str; *ptr != '\0'; ptr++) { 
		if (*ptr == '\n')
    	if ((*(ptr - 1) == '\r' && *(ptr - 2) == '\n') || *(ptr - 1) == '\n')
            break;	
		/* Don't save the \r's. */
		if (*ptr != '\r') g_string_append_c(message->headers, *ptr); 
		}  

  /* advance to the actual beginning of the message (gets rid of any
   * extra line feeds that might be between the header and the actual
   * message). 
	 */
  for ( ; *ptr != '\0'; ptr++)
    if (*ptr != '\r' && *ptr != '\n')
        break; 

	g_string_append_c(message->message, '\n'); /* I like one newline. */

	/* FIXME: strip out \r's here as well? */
  	for (; *ptr != '\0'; ptr++)
		g_string_append_c(message->message, *ptr); 

	/* Construct the message and add it to our database. */
	if (message->headers->str == NULL) message->headers = g_string_append(message->message, "Parse Error");
	if (message->message->str == NULL) message->message = g_string_append(message->message, "Parse Error");
	if (message->subject == NULL) message->subject = g_strdup("(no subject given)");
	if (message->from == NULL) message->from = g_strdup("Parse Error");
	if (message->to == NULL) message->to = g_strdup("Parse Error");
	if (message->cc == NULL) message->cc = g_strdup(" ");
	message->readstatus = g_strdup("Unread");
	message->direction = g_strdup("Incoming");
	message->uid = g_strdup(uid);

	if (!add_msg_to_db(message, timer)) ret = FALSE; 

	/* Free memory and stuff here. */
	g_free(line);
	free_message(message); /* Free the memory */

	return(ret);
}


/* setup_db - Setup the database for the user. 
 * It even puts in a welcome message.
 */

GString *
setup_db (gchar *ruser, gchar *rpass, gchar *user, gchar *pass, gchar *hostname, gchar *db)
{
	gchar *query;
	GString *result; /* Status info returned by the function. */

	result = g_string_new(NULL);

	mysql_init(&GmailApp->mysql);

	if (!mysql_real_connect(&GmailApp->mysql, hostname, ruser, rpass, NULL, 0, NULL, 0)) {
			g_string_append(result, "Unable to open the database.\nIs the mysql root username and password correct?");	
			return(result); }

	query = g_strdup_printf("create database %s", db);
	if (mysql_query(&GmailApp->mysql, query) != 0) {
		g_string_sprintfa(result, "Query failed: %s\n", mysql_error(&GmailApp->mysql));
		g_string_sprintfa(result, "Gmail was unable to create the database on the mysql server.\nYou may wish to try doing this yourself using the mysqladmin program.\n");

		/* Close the database connection. */
		mysql_close(&GmailApp->mysql);
		return result;
		}

	/* Attempt to connect to the database. */
	if (mysql_select_db(&GmailApp->mysql, db) != 0) {
		g_string_sprintfa(result, "Open database (%s) failed: %s\n", db, mysql_error(&GmailApp->mysql));
		g_string_sprintfa(result, "Unable to select the database '%s' on the server.\n", db);
		/* Close the database connection. */
		mysql_close(&GmailApp->mysql);
		return result;
		}

		/* IMPORTANT: Below is what gmail currently thinks the table design is. */

	/* 'display' table */
	query = g_strdup_printf("CREATE TABLE display (id MEDIUMINT(11) UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, date DATETIME NOT NULL, index date (date), subject CHAR(255) NOT NULL, index subject (subject), fromfield CHAR(255) NOT NULL, index fromfield (fromfield), readstatus ENUM('Read', 'Unread', 'Sent', 'Queued') NOT NULL, index readstatus (readstatus), direction ENUM('Incoming', 'Outgoing') NOT NULL, index direction (direction), uid CHAR(255) NOT NULL, index uid (uid), matched SET(' ','Inbox','Outbox','Unread') NOT NULL, index matched (matched))");

	if (mysql_query(&GmailApp->mysql, query) != 0) {
			g_string_sprintfa(result, "Query failed: %s\n", mysql_error(&GmailApp->mysql));
			g_string_sprintfa(result, "Unable to create the table 'display' in database.\n");
			mysql_close(&GmailApp->mysql); /* Close the connection. */
			g_free(query);
			return result;
		}

	g_free(query);
	/* 'details' table */
	query = g_strdup_printf("CREATE TABLE details (id MEDIUMINT(11) UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, headers TEXT NOT NULL, message TEXT NOT NULL, tofield TEXT NOT NULL, ccfield TEXT, bccfield TEXT, attachments TEXT)");
	if (mysql_query(&GmailApp->mysql, query) != 0) {
			g_string_sprintfa(result, "Query failed: %s\n", mysql_error(&GmailApp->mysql));
			g_string_sprintfa(result, "Unable to create the table 'details' in database.\n");
			mysql_close(&GmailApp->mysql); /* Close the connection. */
			g_free(query);
			return result;
		}

	g_free(query);

	/* Create the user account */
	query = g_strdup_printf("GRANT ALL PRIVILEGES ON %s.* TO %s@localhost IDENTIFIED BY '%s' WITH GRANT OPTION", db, user, pass);
	if (mysql_query(&GmailApp->mysql, query) != 0) {
			g_string_sprintfa(result, "Query failed: %s\n", mysql_error(&GmailApp->mysql));
			g_string_sprintfa(result, "Database was created but unable to create a user for this account.\n");

			mysql_close(&GmailApp->mysql); /* Close the connection. */
			g_free(query);
			return(result);
			}

	/* FIXME: Check the result of this query. Should have modified 1 row. */


	g_free(query);

	query = g_strdup_printf("INSERT INTO display (date, subject, fromfield, readstatus, direction, uid, matched) VALUES (NOW(), 'Welcome to Gmail!', 'k_wayne@linuxpower.org', 'Unread', 'Incoming', '0', 'Inbox')"); 

	if (mysql_query(&GmailApp->mysql, query) != 0) {
		g_print("Couldn't provide welcome email: %s\n", mysql_error(&GmailApp->mysql));
		g_print("query was: %s\n", query);
		}

	g_free(query);

	query = g_strdup_printf("INSERT INTO details (tofield, message) VALUES ('%s@localhost', '\nDear Valued Customer,\n\nCongratulations on choosing Gmail.\nMake sure you read the online help. It will explain the basics of vfolders, they are like mail filters but more powerful.\n\nThe online help will explain clearly how to write vfolder queries. They are SQL based, and are quite simple.\n\nOnce you get a large number of vfolders with lots of messages, you will want to use the vfolder caching system. It allows for really fast speeds on huge mailboxes, but make sure you read the online help first.\n\nIf you are not satisfied with this software, please return it for a full refund.\n\nPlease reply to this message to send an email the author of gmail, me! I would be grateful for your comments, good or bad.\n\nthanks,\nWayne Schuller.')", user);

	if (mysql_query(&GmailApp->mysql, query) != 0) {
		g_print("Couldn't provide welcome email: %s\n", mysql_error(&GmailApp->mysql));
		g_print("query was: %s\n", query);
		}

	g_free(query);


	mysql_close(&GmailApp->mysql); /* Close the connection. */
	return(result);
}

/* uid_present - check for existing msg UID in database. 
 * Returns FALSE if UID not found in the database.
 * If the uid message is found we return TRUE.
 * If there is an error we return FALSE. 
 */
gboolean uid_present(gchar *uid)
{
	gint rows;
	MYSQL_RES *res;
	gchar *query, *new_uid;

	g_return_val_if_fail(uid != NULL, FALSE);
		
	/* Open Database. */
	if (!open_db(&GmailApp->mysql)) { return FALSE; };

	new_uid = g_malloc(sizeof(gchar) * ((2 * strlen(uid)) + 1));
	mysql_escape_string(new_uid, uid, strlen(uid));

	query = g_strdup_printf("select id from display where uid = \"%s\"", new_uid);
	g_free(new_uid);

	/* Query the database for the email. */
	if (mysql_query(&GmailApp->mysql, query) != 0) {
		g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		g_print("query was: |%s|\n", query);
		/* Close the database connection. */
		mysql_close(&GmailApp->mysql);
		g_free(query);
		return FALSE;
	} 

	g_free(query);

	res = mysql_store_result(&GmailApp->mysql);

	if (res == NULL) {
		g_warning("Store result failed: %s\n", mysql_error(&GmailApp->mysql));
		/* Close the database connection. */
		mysql_close(&GmailApp->mysql);
		return FALSE;
	}

	rows = mysql_num_rows(res);
	mysql_free_result(res); 

	/* Did we not find it? */
	if (rows == 0) {
			mysql_close(&GmailApp->mysql);
			return FALSE;
			} 

	/* Close the database connection. Don't forget this! DUH!*/
	mysql_close(&GmailApp->mysql);

	return TRUE;	
}



/* msg_match_popup - Create popup window for matching and pop3 downloads. 
 * We are being tricky and using this popup window for two things:
 * a) pop3 downloads
 * b) manual matching runs. From Advanced menu "Rebuild Matched Values"
 *
 */
DTimer *
msg_match_popup (void) {
 	GtkWidget *vbox;
 	GtkWidget *sw, *label;
	DTimer *timer;
	gchar *titles[2] = {"name", "matches"};
	
	timer = g_malloc(sizeof(DTimer));
	timer->value = 0;	
	timer->total = 1;
	timer->dialog = gnome_dialog_new (
                 "Gmail Popup",
                 NULL);
	if (GmailApp->app)
		gnome_dialog_set_parent (GNOME_DIALOG (timer->dialog), GTK_WINDOW (GmailApp->app));

	gtk_widget_set_usize(GTK_WIDGET(timer->dialog), 240, 400);	
  /* gtk_window_set_modal(GTK_WINDOW (timer->dialog), TRUE); */
	gtk_signal_connect(GTK_OBJECT (timer->dialog), "destroy", GTK_SIGNAL_FUNC (destroy_dtimeout), NULL);
	gtk_widget_show_all(timer->dialog);

	vbox = GNOME_DIALOG(timer->dialog)->vbox;

	timer->label = gtk_label_new("Looking up host...");
	gtk_box_pack_start (GTK_BOX (vbox), timer->label, FALSE, FALSE, 5);
	gtk_widget_show(timer->label);


	timer->progress = gtk_progress_bar_new ();
	timer->timeout = gtk_timeout_add (10, (GtkFunction) progress_timeout, timer);
	gtk_progress_set_show_text (GTK_PROGRESS (timer->progress), TRUE);
	gtk_box_pack_start (GTK_BOX (vbox), timer->progress, FALSE, FALSE, 5);
  gtk_widget_show (timer->progress);

	timer->button = gtk_button_new_with_label("Abort");
	gtk_box_pack_start (GTK_BOX (vbox), timer->button, FALSE, FALSE, 5);
	gtk_signal_connect_object (GTK_OBJECT (timer->button), "clicked",
                                   GTK_SIGNAL_FUNC (gnome_dialog_close),
                                   GTK_OBJECT (timer->dialog));
  gtk_widget_show(timer->button);

	label = gtk_label_new("Vfolder matches:");
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 5);
	gtk_widget_show(label);

  /* Create a scrolled window to hold the clist. */
  sw = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
									GTK_POLICY_AUTOMATIC, 
									GTK_POLICY_AUTOMATIC);    
	gtk_box_pack_start (GTK_BOX (vbox), sw, TRUE, TRUE, 5);
  timer->clist = gtk_clist_new_with_titles (2, titles);
  gtk_clist_set_column_title(GTK_CLIST(timer->clist), 0, _("mailbox name"));
  gtk_clist_column_titles_show(GTK_CLIST(timer->clist));
  gtk_clist_column_titles_passive(GTK_CLIST(timer->clist));
	gtk_clist_set_column_width(GTK_CLIST(timer->clist), 0, 120);
  gtk_container_add (GTK_CONTAINER (sw), timer->clist);

	gtk_widget_show_all(timer->dialog);
	
	while (g_main_iteration(FALSE));
	return(timer);
}


/* Update the value of the progress bar so that we get
 * some movement */
gint progress_timeout(DTimer *timer)
{
	gfloat percent;

	if (timer == NULL || !GTK_IS_WIDGET(timer->dialog) || !GTK_IS_WIDGET(timer->progress)) {
		return(FALSE);
		}

	percent = (gfloat) timer->value / timer->total;

	/* g_print("updating %%%f (%i / %i) %f\n", percent, timer->value, timer->total, (gfloat) timer->value/timer->total);   */

 	gtk_progress_bar_update(GTK_PROGRESS_BAR(timer->progress), percent);

  /* As this is a timeout function, return TRUE so that it
   * continues to get called */
  return(TRUE);
}


/* Clean up allocated memory and remove the timer */
void destroy_dtimeout(GtkWidget *widget, DTimer *timer)
{
	if (timer == NULL) return;
	gtk_timeout_remove(timer->timeout);

	/* We need to free the space we allocated sometime, but
	 * doing it here seems to cause coredumps???
	 */
	g_free(timer);
}



/* Check if vfolder msg is unread or not. */
gboolean 
vfolder_msg_unread(gint id) {
	MYSQL_RES *res;
	gchar *query;
	gint affected_rows;

	/* Open the database. */
	if (!open_db(&GmailApp->mysql)) { return FALSE; };

	/* Select all from the database. */
	query = g_strdup_printf("SELECT id FROM display WHERE (id = %d AND readstatus=\"Unread\")", id);

	/* Execute query.. */
	if (mysql_query(&GmailApp->mysql, query) != 0) {
		g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		g_print("query was: |%s|\n", query);
		mysql_close(&GmailApp->mysql);
		return FALSE;
	} 

	/* Check the result. */	
	res = mysql_store_result(&GmailApp->mysql);

	if (res == NULL) {
		g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		g_warning("vfolder match query returned NULL. :(");
		g_print("query was: |%s|\n", query);
		mysql_close(&GmailApp->mysql);
		return FALSE;
		}

	affected_rows = mysql_num_rows(res);

	mysql_free_result(res); 
	mysql_close(&GmailApp->mysql);
	g_free(query);

	if (affected_rows == 1)
		return(TRUE); 
	else 
		return(FALSE); 
}

/* 
 * Mark a vfolder as being read or unread and set it's font accordingly.
 *
 * FIXME: the real problem is undoing this mark when an unread msg is read.
 * How do we know when the vfolder should no longer be marked as containing
 * unread msg's?
 */
void vfolder_contains_unread(gchar *vfolder, gboolean unread_status) {
	gint row;

	/* Loop through each vfolder looking for it. */
	for (row = 0; row != GTK_CLIST (GmailApp->mblist)->rows; row++) {
		Mailbox	*mailbox;
		gchar *name;

		mailbox = gtk_clist_get_row_data(GTK_CLIST(GmailApp->mblist), row);	
		gtk_clist_get_text(GTK_CLIST(GmailApp->mblist), row, 0, &name);

		if (strcmp(name, vfolder) == 0) {
			if (mailbox->unreadmail != unread_status) {
				mailbox->unreadmail = unread_status;
				if (unread_status) 
					gtk_clist_set_cell_style(GTK_CLIST (GmailApp->mblist), row, 0, GmailApp->boldstyle);
					else gtk_clist_set_cell_style(GTK_CLIST (GmailApp->mblist), row, 0, GmailApp->style);
				}
			break;
		}
	}
}


/* Remeber the column resizes by setting gnome_config. */
void
columnresize_cb  (GtkCList *clist, gint column, gint width, gpointer user_data) {
	gchar *config = NULL;

if (GmailApp->rightlist != NULL && clist == GTK_CLIST(GmailApp->rightlist)) {
		config = g_strdup_printf("/gmail/UserInfo/rightlist%i", column);
		gnome_config_set_int(config, width);
	} 
	/* g_print("Resize %s by %i\n", config, width); */
	gnome_config_sync();	/* Write info to config file */
	g_free(config);
}

void
rightlist_popup (GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	gint result;
	gint row, col;

	if (event->button != 3 || event->type != GDK_BUTTON_PRESS) return;

	if (GmailApp->right_pos < 0) return;
	gtk_clist_get_selection_info(GTK_CLIST(widget), event->x, event->y, &row, &col);
	gtk_clist_select_row(GTK_CLIST(widget), row, col);

  GmailApp->rightlist_popup = gnome_popup_menu_new (rightlist_popup_menu);
	result = gnome_popup_menu_do_popup_modal (GmailApp->rightlist_popup, NULL, NULL, event, widget);
  gtk_widget_destroy (GmailApp->rightlist_popup);
	/* g_print("User chose %i\n", result); */
}


/* tag_msg - Tag a message as being special. EXPERIMENTAL CODE
 * FIXME: set in database as well.
 * FIXME: use italics not bold.
 */
void			
tag_msg (GtkWidget *widget, gpointer data)
{
	gint clistrow;

	if (GmailApp->right_pos < 0) {
		/* Nothing to delete. */
		return;
		}

	clistrow = GmailApp->right_pos;

	gtk_clist_set_cell_style(GTK_CLIST (GmailApp->rightlist), clistrow, 0, GmailApp->boldstyle);
	gtk_clist_set_cell_style(GTK_CLIST (GmailApp->rightlist), clistrow, 1, GmailApp->boldstyle);
	gtk_clist_set_cell_style(GTK_CLIST (GmailApp->rightlist), clistrow, 2, GmailApp->boldstyle);
	gtk_clist_set_cell_style(GTK_CLIST (GmailApp->rightlist), clistrow, 3, GmailApp->boldstyle);

}

/* rfc822_to_iso - Convert 822 header date to ISO format date.
 *
 * The date is of the following format:
 *
 * Date: Fri, 22 Oct 1999 17:13:04 +1000 or
 * Date: Sun, 31 Oct 1999 01:15:29 -0500 or
 * Date: 31 Oct 1999 10:45:43 +0100
 *
 * We need to convert this string to:
 *
 * 1999-10-22 17:13:04
 *
 * (the above format is ISO 8601 time)
 *
 * Reasons for using mysql date fields are:
 * 	- It's clean.
 * 	- Working sorts.
 *  - Smaller storage type than a string.
 *
 * FIXME: Handle all RFC 822 compliant variations?
 * 13/6/2000 - Ove Kaaven (ovehk@ping.uio.no) submitted patch to do variation
 * without the day of week in the envelope.
 */
gchar * 
date822_to_iso(gchar *date)
{
	gchar *dow, *day, *time, *year, *month, **ptr;
	gchar *parsed;


	ptr = g_strsplit (date, " ", 255);

	/* skip day of week */
	dow = ptr[0];
	if (dow[0] && dow[strlen(dow)-1]==',') ptr++;
	else dow=NULL;

	/* Is there a standard library to do this? I couldn't find one. */
	if (strcmp(ptr[1], "Jan") == 0) {
		month = g_strdup("1");
	} else if (strcmp(ptr[1], "Feb") == 0) {
		month = g_strdup("2");
	} else if (strcmp(ptr[1], "Mar") == 0) {
		month = g_strdup("3");
	} else if (strcmp(ptr[1], "Apr") == 0) {
		month = g_strdup("4");
	} else if (strcmp(ptr[1], "May") == 0) {
		month = g_strdup("5");
	} else if (strcmp(ptr[1], "Jun") == 0) {
		month = g_strdup("6");
	} else if (strcmp(ptr[1], "Jul") == 0) {
		month = g_strdup("7");
	} else if (strcmp(ptr[1], "Aug") == 0) {
		month = g_strdup("8");
	} else if (strcmp(ptr[1], "Sep") == 0) {
		month = g_strdup("9");
	} else if (strcmp(ptr[1], "Oct") == 0) {
		month = g_strdup("10");
	} else if (strcmp(ptr[1], "Nov") == 0) {
		month = g_strdup("11");
	} else if (strcmp(ptr[1], "Dec") == 0) {
		month = g_strdup("12");
	} else 
		month = g_strdup("parse error"); /* will cause 0-0-0 date */

	day = g_strdup(ptr[0]);
	year = g_strdup(ptr[2]);
	time = g_strdup(ptr[3]);

	parsed = g_strdup_printf("%s-%s-%s %s\n", year, month, day, time);

	g_free(time);
	g_free(year);
	g_free(month);
	g_free(day);

	return(parsed);
}

gint message_count (void) {
	MYSQL_RES *res;
	gchar *query;
	gint affected_rows;

	/* Open the database. */
	if (!open_db(&GmailApp->mysql)) { return FALSE; };

	/* Select all from the database. */
	query = g_strdup_printf("SELECT id FROM display;");

	/* Execute query.. */
	if (mysql_query(&GmailApp->mysql, query) != 0) {
		g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		g_print("query was: |%s|\n", query);
		mysql_close(&GmailApp->mysql);
		return FALSE;
	} 

	/* Check the result. */	
	res = mysql_store_result(&GmailApp->mysql);

	if (res == NULL) {
		g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		g_warning("vfolder match query returned NULL. :(");
		g_print("query was: |%s|\n", query);
		mysql_close(&GmailApp->mysql);
		return FALSE;
		}

	affected_rows = mysql_num_rows(res);

	mysql_free_result(res); 
	mysql_close(&GmailApp->mysql);
	g_free(query);

	return(affected_rows);
}

/* Free up the entire contents of a message structure. */
void free_message(Message *message)
{
	g_return_if_fail(message != NULL);
	
	if (message->headers != NULL) g_string_free(message->headers, TRUE);
	if (message->message != NULL) g_string_free(message->message, TRUE);
	if (message->date != NULL) g_free(message->date);
	if (message->subject != NULL) g_free(message->subject);
	if (message->from != NULL) g_free(message->from);
	if (message->to != NULL) g_free(message->to);
	if (message->cc != NULL) g_free(message->cc);
	if (message->bcc != NULL) g_free(message->bcc);
	if (message->readstatus != NULL) g_free(message->readstatus);
	if (message->direction != NULL) g_free(message->direction);
	if (message->uid != NULL) g_free(message->uid);
	g_free(message);
}

/* Mark the email with id msg_id as read in the database. */
void
mark_readstatus(gint msg_id, gchar *val)
{
	gchar *query;

	/* Open Database. */
	if (!open_db(&GmailApp->mysql)) { return; };

	query = g_strdup_printf("UPDATE display SET readstatus=\"%s\" WHERE id = %i", val, msg_id);

	/* Execute query.. */
	if (mysql_query(&GmailApp->mysql, query) != 0) {
		g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		g_print("Query was: %s\n", query);
		mysql_close(&GmailApp->mysql);
		return;
		} 

	g_free(query);

	/* Close the database connection. */
	mysql_close(&GmailApp->mysql);
}

static void rightlist_clist_create(void)
{
	gint width;
	gchar *titles[4] = { "Status", "From", "Subject", "Date"};
	GtkWidget *gnomepixmap = NULL;
	GtkWidget *pixtitle;

	GmailApp->rightlist = gtk_clist_new(4);

	gtk_clist_set_shadow_type(GTK_CLIST(GmailApp->rightlist), GTK_SHADOW_ETCHED_IN); 

	/* Get a stock icon in the clist titles.   */
	gnomepixmap = gnome_stock_new_with_icon(GNOME_STOCK_MENU_EXEC); 
	pixtitle = gnome_pixmap_button((GtkWidget *) gnomepixmap, titles[0]); 
	gtk_button_set_relief(GTK_BUTTON(pixtitle), GTK_RELIEF_NONE);
	gtk_clist_set_column_widget(GTK_CLIST (GmailApp->rightlist), 0, pixtitle);

	/* Callback for clicking on a column title. */
	gtk_signal_connect(GTK_OBJECT(pixtitle), "clicked",
                       GTK_SIGNAL_FUNC(columnclicked_cb), GINT_TO_POINTER(0));

	gnomepixmap = gnome_stock_new_with_icon(GNOME_STOCK_MENU_HOME); 
	pixtitle = gnome_pixmap_button((GtkWidget *) gnomepixmap, titles[1]); 
	gtk_button_set_relief(GTK_BUTTON(pixtitle), GTK_RELIEF_NONE);
	gtk_clist_set_column_widget(GTK_CLIST (GmailApp->rightlist), 1, pixtitle);

	/* Callback for clicking on a column title. */
	gtk_signal_connect(GTK_OBJECT(pixtitle), "clicked",
                       GTK_SIGNAL_FUNC(columnclicked_cb), GINT_TO_POINTER(1));

	gnomepixmap = gnome_stock_new_with_icon(GNOME_STOCK_MENU_BOOK_OPEN); 
	pixtitle = gnome_pixmap_button((GtkWidget *) gnomepixmap, titles[2]); 
	gtk_button_set_relief(GTK_BUTTON(pixtitle), GTK_RELIEF_NONE);
	gtk_clist_set_column_widget(GTK_CLIST (GmailApp->rightlist), 2, pixtitle);

	/* Callback for clicking on a column title. */
	gtk_signal_connect(GTK_OBJECT(pixtitle), "clicked",
                       GTK_SIGNAL_FUNC(columnclicked_cb), GINT_TO_POINTER(2));

	gnomepixmap = gnome_stock_new_with_icon(GNOME_STOCK_MENU_TIMER); 
	pixtitle = gnome_pixmap_button((GtkWidget *) gnomepixmap, titles[3]); 
	gtk_button_set_relief(GTK_BUTTON(pixtitle), GTK_RELIEF_NONE);
	gtk_clist_set_column_widget(GTK_CLIST (GmailApp->rightlist), 3, pixtitle);

	/* Callback for clicking on a column title. */
	gtk_signal_connect(GTK_OBJECT(pixtitle), "clicked",
                       GTK_SIGNAL_FUNC(columnclicked_cb), GINT_TO_POINTER(3));

	gtk_clist_column_titles_show(GTK_CLIST (GmailApp->rightlist));

	width = gnome_config_get_int("/gmail/UserInfo/rightlist0"); 
	if (width <= 0) width = 50;
	gtk_clist_set_column_width (GTK_CLIST (GmailApp->rightlist), 0, width);
	width = gnome_config_get_int("/gmail/UserInfo/rightlist1"); 
	if (width <= 0) width = 150;
	gtk_clist_set_column_width (GTK_CLIST (GmailApp->rightlist), 1, width);
	width = gnome_config_get_int("/gmail/UserInfo/rightlist2"); 
	if (width <= 0) width = 150;
	gtk_clist_set_column_width (GTK_CLIST (GmailApp->rightlist), 2, width);
	width = gnome_config_get_int("/gmail/UserInfo/rightlist3"); 
	if (width <= 0) width = 50;
	gtk_clist_set_column_width (GTK_CLIST (GmailApp->rightlist), 3, width);
	gtk_clist_set_selection_mode (GTK_CLIST (GmailApp->rightlist), GTK_SELECTION_SINGLE);
	gtk_clist_column_titles_active (GTK_CLIST (GmailApp->rightlist));
	gtk_widget_set_style (GmailApp->rightlist, GmailApp->style); 
	gtk_widget_ref(GmailApp->rightlist);
	gnome_appbar_set_progress (GNOME_APPBAR(GmailApp->appbar), 0.3);

	/* Callback for clicking on a row. */
	gtk_signal_connect(GTK_OBJECT(GmailApp->rightlist), "select_row",
                       GTK_SIGNAL_FUNC(open_vfolder_msg),
                       NULL);

	/* Callback for clicking on a column title. */
	gtk_signal_connect(GTK_OBJECT(GmailApp->rightlist), "click_column",
                       GTK_SIGNAL_FUNC(columnclicked_cb),
                       NULL);

	/* Callback for resizing a column. */
	gtk_signal_connect(GTK_OBJECT(GmailApp->rightlist), "resize_column",
                       GTK_SIGNAL_FUNC(columnresize_cb),
                       NULL);

	/* callback for right click menu. */
  gtk_signal_connect (GTK_OBJECT (GmailApp->rightlist), "button_press_event",
          GTK_SIGNAL_FUNC (rightlist_popup), NULL);
} /* Creation of clist */

