/*
 * Gmail. A Gnome email client.
 * Copyright (C) 1999 Wayne Schuller
 *
 * outgoing.c - stuff for composing new emails and sending them out. 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "main.h"
#include "smtp.h"


extern gboolean	DestroyRightPane(void);
extern gboolean add_msg_to_db(Message *message, DTimer *timer);
extern ssize_t recvline(int sock, void *vptr, size_t maxlen);
extern gint progress_timeout(DTimer *timerinfo);
extern void destroy_dtimeout(GtkWidget *widget, DTimer *timer);
extern void mark_as_matched(gint id, GString *vfolders, gboolean overwrite); 
extern 	Message * load_vfolder_msg(gint id);
extern void broken_cb (GtkWidget *widget, void *data);
extern void about_cb (GtkWidget *widget, void *data);
extern void free_message(Message *message);
extern gchar *mime_get_boundary (void);
extern gint mime_insert_part (gchar **message, gint index, gchar *boundary, gchar *file);
extern void  entry_drag_data_received(GtkWidget *widget, GdkDragContext *context, gint x, gint y, GtkSelectionData *data, guint info, guint time);
extern  gboolean open_db(MYSQL *mysql);
extern void mark_readstatus(gint msg_id, gchar *val);
extern  void vfolder_display_rightpane(gint row);

extern _GmailApp_	*GmailApp;


void  compose_cb (GtkWidget *widget, gchar *type);
void  send_cb (GtkWidget *widget, gboolean queue);
void	cancel_send_cb (GtkWidget *widget, gpointer *data);
void	IndentMsg (Message *msg);
gchar * strip_email(gchar *to);
GList * split_tofield(gchar *tofield);
gboolean smtp_rcpt(gchar *to, gint sock, GtkWidget *label);
gchar * strip_name(gchar *to);
gboolean smtp_send(Message *msg);
void addressbook_cb (GtkWidget *widget, gpointer *data);
void queue_cb (GtkWidget *widget, gpointer *data);
void send_queued_cb (GtkWidget *widget, gpointer *data);

#define set_status_label(label, text) gtk_label_set_text((GtkLabel*)label, text); while (g_main_iteration(FALSE));


/* Standard DnD types */
enum
  {
    TARGET_EMAIL,
  };

GtkTargetEntry email_field_drop_types[] =
{
  {"x-application/x-email", 0, TARGET_EMAIL }
};

#define ELEMENTS(x) (sizeof (x) / sizeof (x[0]))

/* 
 * Routine to compose mail.
 * Is called by the compose/reply/reply to all/forward buttons.
 * type can be: "new", "reply", "replytoall", "forward".
 */
void
compose_cb (GtkWidget *widget, gchar *type)
{
	GtkWidget	*window, *to, *from, *subject, *cc, *bcc, *text;
	gchar			*str;
	Message *msg = NULL;
	FILE	*sig;
	GladeXML *xml;
	gchar *glade_src;


	/* Load the glade file. */
	glade_src = g_strdup_printf("%s/compose.glade", GMAIL_GLADEDIR);
	xml = glade_xml_new(glade_src, "compose");
	g_free(glade_src);

	if (!xml) {
		g_warning("Unable to load xml interface. Are the gmail glade files installed onto your system?\n");
		exit(1);
		}

	glade_xml_signal_autoconnect(xml);

	/* Grab our glade widgets. */
	window = glade_xml_get_widget(xml, "compose");
	to = glade_xml_get_widget(xml, "to");
	from = glade_xml_get_widget(xml, "from");
	subject = glade_xml_get_widget(xml, "subject");
	cc = glade_xml_get_widget(xml, "cc");
	bcc = glade_xml_get_widget(xml, "bcc");
	text = glade_xml_get_widget(xml, "textmsg");


	gtk_text_set_editable (GTK_TEXT (text), TRUE);
	gtk_text_set_word_wrap (GTK_TEXT (text), TRUE);
	gtk_widget_set_style (text, GmailApp->style);

	/* Set up drag and drop targets. */
  gtk_drag_dest_set (to, GTK_DEST_DEFAULT_ALL,
		     email_field_drop_types, ELEMENTS (email_field_drop_types),
		     GDK_ACTION_COPY); 
  gtk_signal_connect( GTK_OBJECT(to), "drag_data_received",
		      GTK_SIGNAL_FUNC( entry_drag_data_received), NULL); 
  gtk_drag_dest_set (cc, GTK_DEST_DEFAULT_ALL,
		     email_field_drop_types, ELEMENTS (email_field_drop_types),
		     GDK_ACTION_COPY); 
  gtk_signal_connect( GTK_OBJECT(cc), "drag_data_received",
		      GTK_SIGNAL_FUNC( entry_drag_data_received), NULL); 
  gtk_drag_dest_set (bcc, GTK_DEST_DEFAULT_ALL,
		     email_field_drop_types, ELEMENTS (email_field_drop_types),
		     GDK_ACTION_COPY); 
  gtk_signal_connect( GTK_OBJECT(bcc), "drag_data_received",
		      GTK_SIGNAL_FUNC( entry_drag_data_received), NULL); 

	/* Put in values to the fields if we need too.  */

	/* Put in the From field from user prefs. */
	/* FIXME: Ideally we will put the 'from' address to match the 'to' address
   * of the email we are replying to. This allows gmail to easily extend
   * to multiple accounts.
	 * This behaviour should be prefs controllable.
   * No this logic is flawed, as it would then put mailing list address
	 * as your from address....
   */
		
	str = gnome_config_get_string("/gmail/UserInfo/EmailAddress"); 

	while (str == NULL) {
		gnome_config_set_string("/gmail/UserInfo/EmailAddress", "compose_cb this now has a value");	
		gnome_config_sync();	/* Write info to config file */
		g_free(str);
		str = gnome_config_get_string("/gmail/UserInfo/EmailAddress"); 
		}

	gtk_entry_set_text(GTK_ENTRY (from), str); 
	g_free(str);

	/* Load the details of the message we might be replying to. */	
	msg = load_vfolder_msg(GmailApp->msg_id);

	/* FIXME: the below code should be re-written logically. */
	if (!strcmp(type, "new") == 0 && msg != NULL) {

		/* If we're forwarding we do nothing. Otherwise we have more work to do. */
		if (!strcmp(type, "forward")==0) {

			 if (msg->from)
					gtk_entry_set_text(GTK_ENTRY (to), msg->from);	

				/* If there were cc fields put them in also. */
				if (strcmp(type, "replytoall")==0) {
					gchar *current = gtk_entry_get_text(GTK_ENTRY(from));


					/* If the msg to field is different to the users default address,
					 * (eg: when it is to a mailing list)
					 * then put that address into the cc list.
					 */
					if (strstr(msg->to, current) == NULL) {
			 			gtk_entry_append_text(GTK_ENTRY (cc), msg->to);	
						if (msg->cc) {
							gtk_entry_append_text(GTK_ENTRY (cc), ", ");
							gtk_entry_append_text(GTK_ENTRY (cc), msg->cc);	
							}
						} else  if (msg->cc) gtk_entry_append_text(GTK_ENTRY (cc), msg->cc);

					}
				}

		/* If we have a subject header insert it. */
		if (msg->subject) {
			if (strncasecmp(msg->subject, "Re:", 3) != 0)
				gtk_entry_set_text(GTK_ENTRY (subject), "Re: ");	
			gtk_entry_append_text(GTK_ENTRY (subject), msg->subject);	
			} 

		IndentMsg(msg);	 /*Indent the old message */

		gtk_text_insert (GTK_TEXT (text), NULL, NULL, NULL, msg->message->str, -1);
		free_message(msg); /* Free the memory */
		}	

	str = gnome_config_get_string("/gmail/UserInfo/Signature");

	if (str != NULL && strlen(str) > 0) {
		sig = fopen(str, "r");
		if (sig == NULL)
			g_print("Cant open signature file: %s\n", str);
		else {
			/* Lets insert signature file now */
			gchar *txt;
			gchar line[80];

			/* Insert two newlines. */
			gtk_text_insert (GTK_TEXT (text), NULL, NULL, NULL, "\n\n", -1);
			do {
				txt = fgets(line, 80, sig);
				gtk_text_insert (GTK_TEXT (text), NULL, NULL, NULL, txt, -1);
			} while(txt!=NULL);
		fclose(sig);
	}}
	g_free(str);

	/* Go to the top of the message. */
	gtk_text_set_point(GTK_TEXT (text), 0);

	gtk_widget_set_usize(GTK_WIDGET(window), 650, 500);
	gtk_widget_show_all(window);

	return;
}

/* 
 * Callback for when they click on the 'send' button in the compose window. 
 * Will parse the compose window, and send the message.
 * FIXME: avoid global vars. See cancel_send_cb for better code.
 */
void
send_cb (GtkWidget *widget, gboolean queue)
{
	Message	*msg;
	gchar *attach_file;	/* File for the attachment. */
	GtkWidget *dialog; 
	gboolean result = TRUE; /* Did we successfully send? */ 
	GtkWidget *window, *to, *from, *sub, *cc, *bcc, *text, *attach;
	GladeXML *xml = glade_get_widget_tree(GTK_WIDGET (widget));
	gboolean mime = FALSE;
	gchar *boundary = NULL; /* MIME boundary */

	/* Get a pointer to the window the button was in. */
	window = glade_xml_get_widget(xml, "compose");
	to = glade_xml_get_widget(xml, "to");
	from = glade_xml_get_widget(xml, "from");
	sub = glade_xml_get_widget(xml, "subject");
	cc = glade_xml_get_widget(xml, "cc");
	bcc = glade_xml_get_widget(xml, "bcc");
	text = glade_xml_get_widget(xml, "textmsg");
	attach = glade_xml_get_widget(xml, "fileentry1");


	/* Create the message to be sent. 
	 * We want all the data in our own memory space, so we g_strdup the info
	 * from the gtk_entry widgets.
	 */
	msg = g_malloc(sizeof(Message));
	msg->headers = g_string_new(NULL); /* We build headers later. */
	msg->message = g_string_new(NULL); /* We build the body later. */
	msg->date = NULL;
	msg->subject = g_strdup(gtk_entry_get_text(GTK_ENTRY(sub)));
	msg->from = g_strdup(gtk_entry_get_text(GTK_ENTRY(from)));
	msg->to = g_strdup(gtk_entry_get_text(GTK_ENTRY(to)));
	msg->cc = g_strdup(gtk_entry_get_text(GTK_ENTRY(cc)));
	msg->bcc = g_strdup(gtk_entry_get_text(GTK_ENTRY(bcc)));
	if (queue)
		msg->readstatus = g_strdup("Queued");
	else
		msg->readstatus = g_strdup("Sent");
	msg->direction = g_strdup("Outgoing");
	msg->uid = g_strdup("Outgoing messages have no UID.");

	/* g_print("To: %s\nFrom: %s\nSubject: %s\nCc: %s\n\n%s\n", to, from, subject, cc, body); */

	/* Lets do some primitive error checking. */

	/* Missing "To:" field. */
	if (strlen(msg->to) < 2) {  
		dialog = gnome_message_box_new("please fill out the \"To:\" field", GNOME_MESSAGE_BOX_ERROR, "ok", 0);
    	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
			gtk_window_set_modal(GTK_WINDOW (dialog), TRUE);
    	gtk_widget_show(dialog);
			return;
		}

	/* Missing "Subject:" field. */
	if (strlen(msg->subject) < 2) {
		dialog = gnome_message_box_new("subject field missing\n(don't you realise how incredibly annoying emails without subjects are?)", GNOME_MESSAGE_BOX_WARNING, "ok", 0);
    	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
			gtk_window_set_modal(GTK_WINDOW (dialog), TRUE);
    	gtk_widget_show(dialog);
	    return;
		}

	/* Encode the attachment into the message here. */ 

	/* If they have given us a valid attachment we go into MIME mode. */
	attach_file = gnome_file_entry_get_full_path(GNOME_FILE_ENTRY(attach), TRUE);	
	if (attach_file == NULL) mime = FALSE; else mime = TRUE; 

	/* If we have an attachment send a MIME message. 
	 * If there are no attachments we just send a plain txt email.
	 *
	 * This little section of code pretty much follows spruce exactly.
	 * It is fairly unreadable, it would be good to rewrite it using
	 * g_strings and g_string_append, etc. 
	 */
	if (mime) {
		gchar *body, *buffer = NULL;
		gint index;

		boundary = g_strdup(mime_get_boundary());
		body = g_malloc0(210 + strlen(boundary) + strlen(gtk_editable_get_chars(GTK_EDITABLE(text), 0, -1)));
		index = 0;
		index += sprintf(body, "This is a multi-part message in MIME format.\n\n");
		buffer = g_strdup(gtk_editable_get_chars(GTK_EDITABLE(text), 0, -1));
		index += sprintf(body+index, "--%s\n", boundary);
		index += sprintf(body+index, "Content-Type: text/plain\nContent-Transfer-Type: 8bit\n\n%s\n", buffer);
		g_free(buffer);
		/* Insert a single attachment. */
		index += mime_insert_part (&body, index, boundary, attach_file);
		body = g_realloc(body, strlen(body) + strlen(boundary) + 7);
		/* insert the end marking boundary */
		sprintf(body + index, "--%s--\n", boundary);
		body[index + 2 + strlen(boundary) + 3] = '\0';
		msg->message = g_string_append(msg->message, body);
		g_free(body);
		mime = TRUE;
	} else {
		/* Just a plain old text message. */
		msg->message = g_string_append(msg->message, gtk_editable_get_chars(GTK_EDITABLE (text), 0, -1));
		mime = FALSE;
	}
	

	/* Create the headers for the outgoing message. */	
	g_string_sprintf(msg->headers, "To: %s\r\n", msg->to);
	if (msg->cc != NULL) g_string_sprintfa(msg->headers, "Cc: %s\r\n", msg->cc); 

	/* If we are using MIME then we add some extra stuff to the headers. */
	if (mime) {
		g_string_sprintfa(msg->headers, "From: %s\r\nSubject: %s\r\nMIME-Version: 1.0\r\nContent-Type: multipart/mixed; boundary=\"%s\"\r\nX-Mailer: Gmail %s (http://gmail.linuxpower.org)\r\n", msg->from, msg->subject, boundary, VERSION); 
		} else {
		g_string_sprintfa(msg->headers, "From: %s\r\nSubject: %s\r\nX-Mailer: Gmail %s (http://gmail.linuxpower.org)\r\n", msg->from, msg->subject, VERSION); 
		} 

	if (queue)
		result = TRUE;
	else 
		result = smtp_send(msg);

	/* If we have a result add the msg to the database and close.
	 * Otherwise do nothing.
	 */
	if (result) {
	add_msg_to_db(msg, NULL);

	/* We can close the compose window now. */
	if (window != NULL) gtk_widget_destroy(GTK_WIDGET (window));

	/* Free the xml reference. */
	gtk_object_unref(GTK_OBJECT(xml));

	}

	/* Free the memory from above. */
	free_message(msg); /* Free the memory */
}


/* Callback for when they click on the 'cancel' button in a compose window.
 * Need to close the relevant compose window.
 * 
 * FIXME: We need to check and remind them if they have a half composed
 * message open. 
 */
void
cancel_send_cb (GtkWidget *widget, gpointer *data)
{
	GtkWidget *window;
	GladeXML *xml = glade_get_widget_tree(GTK_WIDGET (widget));

	/* Get a pointer to the window the button was in. */
	window = glade_xml_get_widget(xml, "compose");

	if (window != NULL) gtk_widget_destroy(GTK_WIDGET (window));

	/* Free the xml reference. */
	gtk_object_unref(GTK_OBJECT(xml));
}

/* This routine is called to put indents at the start of each line of the
 * message in the static cache. 
 * FIXME: Should we add other stuff? like "thingy wrote:\n" first?
 */
void
IndentMsg (Message *msg) {
	gchar *tok;  /* tokenizer we use to find newlines. */
	gint diff=0; /* the number of chars until just after a newline. */

	while (diff < msg->message->len) {

		/* Insert the indent after the newline. */
		/* FIXME: Use a configurable indent from prefs. */
		g_string_insert(msg->message, diff, "> ");

		/* Move the token just past the newline. */
		tok = msg->message->str + diff;

		/* Work out the distance to the next newline + 1 */
		diff += strcspn(tok, "\n")+1; 
	}
	g_string_insert(msg->message, 0, "\n");
}

/* smtp_send - send a message via smtp protocol 
 * 
 * It is up to the calling function to free the msg data.
 *
 * This is very basic. Thanks to Spruce for the reference. It's improved a
 * a bit since then.
 *
 * FIXME/TODO:
 * 
 * - Make the progress bar work according to the contents of the message
 * being uploaded.
 * - Fix awful hack checking for the existance of the timer constantly.
 *
 */
gboolean
smtp_send(Message *msg) 
{
 	struct hostent *ptr_host;
	gchar	*hostname;
	struct sockaddr_in sin;
	port_t 	port = 25;	/* Standard smtp port. FIXME: make configurable */
	gchar ip[17]; /* IPv4 dotted decimal and NULL */
	int sock;
  gchar buffer[256];
  gchar command[256];
	gchar localhost[256];
	GString *sbody;
	gchar *body;
  GtkWidget *button;
  GtkWidget *vbox;
	GtkWidget *child;
	GtkWidget *label;
	gchar fejj[6]; /* Named in honour of Spruce author fejj. */
	DTimer *timer;
	gint i;

	timer = g_malloc(sizeof(DTimer));

	timer->clist = NULL; /* We're not using the clist. */
	timer->value = 0;	
	timer->total = 1;

	timer->dialog = gnome_dialog_new (
                 "Gmail Popup",
                 NULL);
	if (GmailApp->app)
             gnome_dialog_set_parent (GNOME_DIALOG (timer->dialog), GTK_WINDOW (GmailApp->app));

	gtk_widget_set_usize(GTK_WIDGET(timer->dialog), 200, 135);	
  gtk_window_set_modal(GTK_WINDOW (timer->dialog), TRUE);

	gtk_signal_connect(GTK_OBJECT (timer->dialog), "destroy", GTK_SIGNAL_FUNC (destroy_dtimeout), NULL);

	vbox = GNOME_DIALOG(timer->dialog)->vbox;

	label = gtk_label_new("Looking up smtp host...");
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 5);
	gtk_widget_show(label);


	timer->progress = gtk_progress_bar_new ();
	timer->timeout = gtk_timeout_add (10, (GtkFunction) progress_timeout, timer);
	/* gtk_widget_set_usize (timer->progress, 280, 48);*/
	gtk_progress_set_show_text (GTK_PROGRESS (timer->progress), TRUE);
	gtk_box_pack_start (GTK_BOX (vbox), timer->progress, FALSE, FALSE, 5);
  gtk_widget_show (timer->progress);

	button = gtk_button_new_with_label("Abort");
	gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, FALSE, 5);
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC (gnome_dialog_close), GTK_OBJECT (timer->dialog));
  gtk_widget_show(button);

 	gtk_widget_show_all(timer->dialog);
	while (g_main_iteration(FALSE));

	

	hostname = gnome_config_get_string("/gmail/UserInfo/smtp");
	if (hostname == NULL) {
  	gtk_label_set_text((GtkLabel *) label, "No smtp host in config.");
    return(FALSE);
		}
   	if ( (ptr_host = gethostbyname(hostname)) == NULL) {
  	gtk_label_set_text((GtkLabel *) label, "Bad SMTP hostname");
    return(FALSE);
		}

   	memcpy(&sin.sin_addr, ptr_host->h_addr, ptr_host->h_length);
	g_snprintf(ip, sizeof(ip), "%s", inet_ntoa(sin.sin_addr));

	sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sock == -1) {
		set_status_label(label, "SMTP socket failure");
    return(FALSE);
		}

	sin.sin_family = AF_INET;
   	sin.sin_port = htons(port);

	set_status_label(label, "Establishing connection."); 
	if (connect(sock, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
		set_status_label(label, "no connection");
    return(FALSE);
		}

	g_return_val_if_fail(GTK_IS_WIDGET(timer->dialog), FALSE);

	if ( (recvline(sock, buffer, 255) < 0) || (strstr(buffer, ServiceRdy) == NULL) ) {
		g_warning("smtp server didn't send 220.");
		set_status_label(label, "bad smtp server");
		close(sock);
      	return(FALSE);
   	}

	timer->value = 0.1;

	set_status_label(label, "Sending HELO"); 
	/* Send HELO */
	gethostname(localhost, 255);
	g_snprintf(command, 255, "HELO %s\r\n", localhost);
   	if (send (sock, command, strlen(command), 0) < 0) {
		g_warning("smtp didn't like helo");
		set_status_label(label, "untalkative smtp");
      	return(FALSE);
   	}

	/* Check HELO reply. */		
	if ( (recvline(sock, buffer, 255) < 0) 
			|| (strstr(buffer, ReqMlActOK) == NULL) ) {
		g_warning("unfriendly smtp");
		set_status_label(label, "unfriendly smtp");
      	return(FALSE);
   	}

	g_return_val_if_fail(GTK_IS_WIDGET(timer->dialog), FALSE);
	timer->value = 0.2;

	/* set_status("Sending MAIL FROM"); */
	/* Send MAIL FROM */
	g_snprintf(command, 255, "MAIL FROM: %s\r\n", msg->from);
   	if (send (sock, command, strlen(command), 0) < 0) {
		g_warning("smtp didn't like MAIL FROM");
		set_status_label(label, "smtp disliked MAIL FROM");
      	return(FALSE);
   	}

	/* Check MAIL FROM reply. */
	if ( (recvline(sock, buffer, 255) < 0) 
			|| (strstr(buffer, ReqMlActOK) == NULL) ) {
		g_warning("didn't like our mail from");
		set_status_label(label, "smtp rejected from field");
      	return(FALSE);
   	}

	g_return_val_if_fail(GTK_IS_WIDGET(timer->dialog), FALSE);
	timer->value = 0.3;

	smtp_rcpt(msg->to, sock, label); /* Send to ppl in To: field. */
	smtp_rcpt(msg->cc, sock, label); /* Send to ppl in Cc: field. */
	smtp_rcpt(msg->bcc, sock, label); /* Send to ppl in Bcc: field. */

	g_return_val_if_fail(GTK_IS_WIDGET(timer->dialog), FALSE);
	timer->value = 0.4;

	/* set_status("Sending DATA"); */
	/* Send DATA command */
	g_snprintf(command, 255, "DATA\r\n");
   	if (send (sock, command, strlen(command), 0) < 0) {
		g_warning("smtp didn't like DATA");
		set_status_label(label, "smtp disliked DATA");
      	return(FALSE);
   	}

	/* Check reply. */
	if ( (recvline(sock, buffer, 255) < 0) 
			|| (strstr(buffer, StrtMailInpt) == NULL) ) {
		g_warning("didn't like our DATA");
		set_status_label(label, "smtp really disliked DATA");
      	return(FALSE);
   	}

	g_return_val_if_fail(GTK_IS_WIDGET(timer->dialog), FALSE);

	/* FIXME: insert date here */
	sbody = g_string_new(NULL);
	/* g_string_sprintfa(sbody, "X-Mailer: Gmail 0.5.4!\r\nFrom: %s\r\nSubject: %s\r\n%s", msg->from, msg->subject, msg->message->str); */
	g_string_sprintfa(sbody, "%s\r\n%s", msg->headers->str, msg->message->str); 

	
	/* The following paragraph of code is taken from 
	 * fejj, author of the all purposeful linux email program SPRUCE!!!
	 * The variable fejj has been named in his honour!
	 *
	 * The issue here is that we need to make sure \n's become \r\n's
	 * according to the RFC. Qmail in particular is very fussy about this.
	 * 
	 * This code basically sends the data one char at a time, because
	 * we need to be looking to make sure every \n has a \r before it.
	 *
	 */
	body = sbody->str;
  for (i = 0; i < strlen(body); i++) {
		if (body[i] == '.' && body[i-1] == '\n' && body[i+1] == '\n') {
        /* if we have a lone period on a line... */
        strcpy(fejj, "..");
		} else if (body[i] == '\n' && body[i-1] != '\r') {
			strcpy(fejj, "\r\n");
		} else {
			fejj[0] = body[i];
			fejj[1] = '\0';
		}
    send (sock, fejj, strlen(fejj), 0);
  }

	g_string_free(sbody, TRUE);

	/* Send the \r\n. to end the msg. */
	send (sock, "\r\n.\r\n", 5, 0);

   if ( (recvline(sock, buffer, 255) < 0)
       || (strstr(buffer, ReqMlActOK) == NULL) ) {
		g_warning("didn't like our DATA");
		set_status_label(label, "smtp rejected DATA");
      	return(FALSE);
   }

	g_return_val_if_fail(GTK_IS_WIDGET(timer->dialog), FALSE);
	timer->value = 0.8;


	set_status_label(label, "Sending QUIT"); 
	/* Send QUIT */
	g_snprintf(command, 255, "QUIT\r\n");
   	if (send (sock, command, strlen(command), 0) < 0) {
		g_warning("smtp didn't like QUIT");
		set_status_label(label, "smtp disliked QUIT");
      	return(FALSE);
   	}

	/* Check reply. */
	if ( (recvline(sock, buffer, 255) < 0) 
			|| (strstr(buffer, ServClose) == NULL) ) {
		g_warning("didn't like our QUIT");
		set_status_label(label, "smtp server hates quitters");
      	return(FALSE);
   	}


	timer->value = 1;

	/* Send positive feedback to the user. */
  set_status_label(label, "Message sent!"); 

	/* Update the button text. */
	child = GTK_BIN (button)->child;
	if (GTK_IS_LABEL(child))
		gtk_label_set(GTK_LABEL(child), "Close");

	while (g_main_iteration(FALSE));

	gtk_widget_destroy(timer->dialog);
	destroy_dtimeout(NULL, timer);

	gnome_appbar_set_status (GNOME_APPBAR (GmailApp->appbar), "Message Sent!");

	return(TRUE);
} /* SMTP HACK */


/* Strip email.
 *
 * It's ok to have names in the tofield, but we should strip
 * out everything but the bare address for the RCPT.
 * "Barry Smith" <b.smith@hotmail.com>
 * Should become:
 * b.smith@hotmail.com
 * 
 * If they have errors like only one bracket, we let the smtp server deal
 * with it. 
 * This code assumes there is only one email in the 'to' string. Multiple
 * emails should be parsed out before calling this function.
 *
 * The value returned should be freed by the calling function.
 * FIXME: this function alters to the argument string. BAD
 */
gchar *
strip_email(gchar *to)
{
	gchar *tostrip = NULL, *ptr, *ptr2;

	ptr = strstr(to, "<");
	ptr2 = strstr(to, ">");
	if (ptr != NULL && ptr2 != NULL) {
		ptr++;
		strcpy(ptr2, "\0");
		tostrip = g_strdup(ptr);	
		}	

	if (tostrip == NULL) tostrip = g_strdup(to);

	g_strstrip(tostrip); /* Remove any spaces */
	return(tostrip);
}

/* Like strip_email, but returns the name, not the email addy.
 * Calling function should free the memory.
 */
gchar *
strip_name(gchar *to)
{
	gchar *tostrip = NULL, *ptr, *ptr2;

	tostrip = g_strdup(to);	

	ptr = strstr(tostrip, "<");
	ptr2 = strstr(tostrip, ">");
	if (ptr != NULL && ptr2 != NULL) {
		strcpy(ptr, "\0");
		}	

	g_strstrip(tostrip); /* Remove any spaces */
	return(tostrip);
}


/* split_tofield - parse a tofield returning a glist of the addresses.
 *
 */
GList * split_tofield(gchar *tofield)
{ 
	GList *tolist = NULL;
	gchar **split;
	gint i;

	/* Split string into comma seperated chunks. */
	split = g_strsplit(tofield, ",", 255);

	/* Loop through each chunk and add to glist. */
	i = 0;	
	while (split[i] != NULL) {
		gchar *tostrip;

		tostrip = strip_email(split[i]);
    tolist = g_list_append(tolist, tostrip);
		i++;
	}

	g_strfreev(split);
	return(tolist);
}

/* smtp_rcpt - Send a RCPT TO to everyone in the string. 
 *
 */
gboolean 
smtp_rcpt(gchar *to, gint sock, GtkWidget *label)
{
	GList *tolist, *toitem;
  gchar command[256];
  gchar buffer[256];

	/* If the field is empty just return. */
	if (to == NULL) return TRUE;
	
	tolist = split_tofield(to);

	toitem = tolist;

	while (toitem != NULL) {
		gchar *msg;
		GList *olditem;

		msg = g_strdup_printf("RCPT TO: %s", (gchar *) toitem->data);

		/* g_print("%s\n", msg); */

		set_status_label(label, msg); 

		/* RCPT TO for each recipient. */
		g_snprintf(command, 255, "RCPT TO: %s\r\n", (gchar *) toitem->data);
   		if (send (sock, command, strlen(command), 0) < 0) {
			g_warning("smtp didn't like RCPT TO %s", (gchar *) toitem->data);
			set_status_label(label, "SMTP rejected our RCPT TO");
   		return(FALSE);
   	}

		/* Check reply. */
		if ( (recvline(sock, buffer, 255) < 0) 
				|| (strstr(buffer, ReqMlActOK) == NULL) ) {
			set_status_label(label, "SMTP server rejected recipient.");
    	return(FALSE);
   	}

		g_free(toitem->data);
		olditem = toitem;
		toitem = toitem->next;
		g_list_free(olditem);
	}

	return(TRUE);
}

/* This function will launch the users selected address book program. */
void
addressbook_cb (GtkWidget *widget, gpointer *data)
{
	char *argv[1];
	gchar *program = gnome_config_get_string("/gmail/UserInfo/AddressBook=gnomecard"); 

	g_return_if_fail(program != NULL);

	argv[0] = program;
  gnome_execute_async (NULL, 1, argv);
	g_free(program);
}

/* Queue_cb - Queue a composed message for later sending. 
 * Just call send_cb with the TRUE option to tell it to just add to the
 * database and not try and send it immediately.
 */
void
queue_cb (GtkWidget *widget, gpointer *data) 
{
	send_cb(widget, TRUE);
}

/* send_queued_cb - Send all queued messages via SMTP. 
 *
 */
void
send_queued_cb (GtkWidget *widget, gpointer *data) 
{
	MYSQL_RES *res;
	MYSQL_ROW mysqlrow;
	Message *msg;
	gint id, affected_rows;
	gchar *query, *note;

	if (!open_db(&GmailApp->mysql)) { return; };

	/* Get all the messages that are queued. */
	query = g_strdup_printf("SELECT id FROM display WHERE readstatus=\"Queued\"");

	/* Execute query.. */
	if (mysql_query(&GmailApp->mysql, query) != 0) {
		g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		g_print("query was: |%s|\n", query);
		g_free(query);
		mysql_close(&GmailApp->mysql);
		return;
	} 


	/* Check the result. */	
	res = mysql_store_result(&GmailApp->mysql);

	if (res == NULL) {
		g_warning("Query failed: %s\n", mysql_error(&GmailApp->mysql));
		g_warning("vfolder match query returned NULL. :(");
		g_print("query was: |%s|\n", query);
		mysql_close(&GmailApp->mysql);
		return;
		}

	affected_rows = mysql_num_rows(res); 

	while ((mysqlrow = mysql_fetch_row(res))) {
		/* Get the msg id. */
		id = (gint) atoi(mysqlrow[0]);

		/* Load the message.  */
		msg = load_vfolder_msg(id); 

		if (msg != NULL) {

		/* smtp_send the message. */
		smtp_send(msg); 

		/* Free the message. */
		free_message(msg);

		/* Mark the message as read in the database. */	
		mark_readstatus(id, "Sent");
		}
	}	

	g_free(query);
	mysql_free_result(res); 
	mysql_close(&GmailApp->mysql);

	/* If we had the outbox open, update it as it has now changed. */
	/* FIXME: We assume the outbox is the second vfolder. */
	if (GmailApp->left_pos == 1) {
  	vfolder_display_rightpane(GmailApp->left_pos);
	}

	/* Report back to the user. */
	if (affected_rows == 1)
		note = g_strdup_printf("Sent one queued message.");
	else
		note = g_strdup_printf("Sent %i queued messages.", affected_rows);
	gnome_appbar_set_status (GNOME_APPBAR (GmailApp->appbar), note);
	g_free(note);
}
