/*  
 * Gmail. A Gnome email client.
 * Copyright (C) 1999 Wayne Schuller
 *
 * pop3.c - Code for talking to pop3 servers.
 * Alot of this code was done using Spruce as a reference. 
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <sys/param.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>

#include "main.h" 

#define	ERROR	0
#define SUCCESS	1

enum pop3_errnos {P_NO_ERROR, P_ERR_RESOLVE, P_ERR_SOCKET, P_ERR_CONNECT, P_ERR_SERV_ACK, P_ERR_USER_SEND, P_ERR_USER_RECV, P_ERR_PASS_SEND, P_ERR_PASS_RECV, P_ERR_LIST_SEND, P_ERR_LIST_RECV, P_ERR_STAT_SEND, P_ERR_STAT_RECV, P_ERR_RETR_SEND, P_ERR_RETR_RECV, P_ERR_RSET_SEND, P_ERR_RSET_RECV, P_ERR_QUIT_SEND, P_ERR_QUIT_RECV, P_ERR_DELE_SEND, P_ERR_DELE_RECV};
					  
typedef enum pop3_errnos pop3errnos;

gboolean	pop3_check(char *server, int port, char *user, char *pass);
ssize_t recvline(int sock, void *vptr, size_t maxlen);
ssize_t recvch(int sock, char *ptr);
static int p_user(int sock, char *login);
static int p_pass(int sock, char *passwd);
static int p_stat(int sock, gint *num, gint *size);
static int p_list(int sock);
static int p_retr(int sock, int id, gint total, gboolean delete, DTimer *timer);
static int p_dele(int sock, int id, DTimer *timer);
static int p_rset(int sock);
static int p_quit (int sock);
static int Resolve (Server *server);

#define set_status_label(label, text) gtk_label_set_text((GtkLabel*)label, text); while (g_main_iteration(FALSE));

extern int errno;
pop3errnos pop3errno;

extern _GmailApp_	*GmailApp;
extern	gboolean parse_pop3_msg(GString *msg, gchar *uid, DTimer *timer);
extern 	gboolean uid_present(gchar *uid);
extern	DTimer * msg_match_popup (void);
extern  void clear_vfolder_cache(Mailbox *mailbox);
extern  void vfolder_display_rightpane(gint row);


static int p_user(int sock, char *login)
{
   /* login to the server - step 1 (send login name) */
   char buffer[256];
   char command[256];

   /* lets send our user login to the server */
   g_snprintf(command, 255, "USER %s\r\n", login);
	
   if (send (sock, command, strlen(command), 0) < 0)
   {
      pop3errno = P_ERR_USER_SEND;
      return ERROR;
   }

   if ( (recvline(sock, buffer, 255) < 0) 
        || (strstr(buffer, "+OK") == NULL) )
   {
      /* I'm guessing the pop3 server didn't like our USER command... */
      pop3errno = P_ERR_USER_RECV;
      return ERROR;
   }
	
   return SUCCESS;
}

static int p_pass(int sock, char *passwd)
{
   /* login to the server - step 2 (send passwd) */
   char buffer[256];
   char command[256];

   /* lets send our user passwd to the server */
   g_snprintf(command, 255, "PASS %s\r\n", passwd);

   if (send (sock, command, strlen(command), 0) < 0)
   {
      pop3errno = P_ERR_PASS_SEND;
      return ERROR;
   }

   if ( (recvline(sock, buffer, 255) < 0) 
        || (strstr(buffer, "+OK") == NULL) )
   {
      /* I'm guessing the pop3 server didn't like our PASS command... */
      pop3errno = P_ERR_USER_RECV;
      return ERROR;
   }

	/* If there is an error the line is -ERR */
	if (buffer[0] == '-') 
		return ERROR;

   return SUCCESS;
}

static int p_stat(int sock, gint *num, gint *size)
{
	/* returns the number of messages on the server 
	 * or -1 to signify an error */
   char buffer[256];
   char command[256];

   /* lets STAT the pop3 server... */
   strncpy(command, "STAT\r\n", 255);

   if (send (sock, command, strlen(command), 0) < 0)
   {
      pop3errno = P_ERR_STAT_SEND;
      return ERROR;
   }

   if ( (recvline(sock, buffer, 255) < 0) 
        || (strstr(buffer, "+OK") == NULL) )
   {
      /* something went wrong... */
      pop3errno = P_ERR_STAT_RECV;
      return -1;
   }
	
	/* get the number of messages on the server */
	/* spruce had dodgy code here */
	if (sscanf(buffer, "+OK %d %d\r\n", num, size) != 2) 
		return(-1);
	else 	{
		return(TRUE);
		}
}

static int p_list(int sock)
{
	/* list the messages on the server 
	 * returns the number of messages or -1 if an error occurs */
   char buffer[256];
   char command[256];
	int num = 0;

   /* prepare the LIST command */
   strncpy(command, "LIST\r\n", 255);
	
   if (send (sock, command, strlen(command), 0) < 0)
   {
      pop3errno = P_ERR_LIST_SEND;
      return ERROR;
   }

   if ( (recvline(sock, buffer, 255) < 0) 
        || (strstr(buffer, "+OK") == NULL) )
   {
      /* something went wrong */
      pop3errno = P_ERR_LIST_RECV;
      return ERROR;
   }
	
	while (recvline(sock, buffer, 255) > 0)
	{
		num++;
	}
	
	num--;
	
	return num;
}

/* p_retr does 3 things: (we should probably change the name)
 * 1. Check the UID of the message. Stop if we've seen it before.
 * 2. Download the message using RETR, but obey the users d/l threshold limit.
 * 3. Delete the message if the user has that preference set and they have
 * downloaded the whole thing.
 *
 * If delete is FALSE we don't delete the message.
 * If delete is TRUE, but the message is cut off by our threshold limit, then
 * we don't delete the message. This is basically what Eudora does. 
 * Return TRUE if it goes ok.
 * Return FALSE if any problems.
 *
 * FIXME: there is a problem with long lines getting truncated.
 *
 */
static int p_retr(int sock, int id, gint total, gboolean delete, DTimer *timer)
{
 	char buffer[256];
 	char command[256];
	GString	*msg;
	gchar	*uid = NULL;
 	int n;
	gint i, size;
	gchar *tmp;
  gchar *end;
  gint    index;
	gint threshold;

	g_snprintf(buffer, 255, "Looking at message %d of %d.", id, total);
	set_status_label(timer->label, (gchar*) buffer);
	/* timer->value = id*10;  Old CODE */

	/* Get UID of message and check it first. */
 	g_snprintf(command, 255, "UIDL %d\r\n", id);
 	if (send (sock, command, strlen(command), 0) < 0) {
		pop3errno = P_ERR_RETR_SEND;
 		return FALSE;
   	}

   if ( (recvline(sock, buffer, sizeof(buffer)-1) < 0) 
        || (strstr(buffer, "+OK") == NULL) )
   {
      if (strstr(buffer, "+OK") == NULL)
         g_warning("pop3 UIDL: couldn't find \"+OK\" -> %s\n", buffer);
      pop3errno = P_ERR_RETR_RECV;
      return FALSE;
   }


	/* Parse the result of the UIDL command*/
    tmp = g_strdup_printf("+OK %d ", id);
    index = strlen(tmp); g_free(tmp);
    tmp = &buffer[index]; /* Remove intitial +OK text */
	/* Find the /r/n and chuck a \0 where the /r is. */
    for (end = tmp; *end != '\r'; end++) ;
    *end = '\0'; 
    uid = g_strdup(tmp);   /* Record the UIDL */

	/* g_warning("MSG (%d), UID (%s), UIDL (%s)", id, uid, buffer); */
	/* Check if this msg uid has been downloaded before. */
	if (uid_present(uid)) {
		g_warning("UID matched (%s). Skipping msg %d. UIDL returned: %s", uid, id, buffer);
		g_free(uid);
		if (delete) p_dele(sock, id, timer); /* delete if we have to. */
		return TRUE; /* Don't download it if we have this uid msg. */
	}

	memset((void*)&buffer, 0, sizeof(buffer)); /* Clear buffer. */

	g_snprintf(buffer, 255, "Downloading message %d of %d.", id, total);
	set_status_label(timer->label, (gchar*) buffer);

	threshold =  atoi(gnome_config_get_string("/gmail/POP3Servers/dlthreshold"));

   /* lets tell the server which message to retrieve */
	if (threshold < 1) {
   g_snprintf(command, 255, "RETR %d\r\n", id);
		} else {
   g_snprintf(command, 255, "TOP %d %d\r\n", id, threshold);
		}
   if (send (sock, command, strlen(command), 0) < 0)
   {
      pop3errno = P_ERR_RETR_SEND;
      return FALSE;
   }

   if ( (recvline(sock, buffer, sizeof(buffer)-1) < 0) 
        || (strstr(buffer, "+OK") == NULL) )
   {
      pop3errno = P_ERR_RETR_RECV;
      return FALSE;
   }
	if (sscanf(buffer, "+OK %d\r\n", &size) != 1) size = -1;

	i = 0;
	n = recvline(sock, buffer, sizeof(buffer)-1);/* Get line. */
	msg = g_string_new(NULL);

   while ( !(n <= 0) && (strncmp(buffer, ".\r\n", 3) != 0)) {

		g_strstrip(buffer); /* Strip off white space at end. */

		timer->value += strlen(buffer);  /* Increment the timer */
		while (g_main_iteration(FALSE)); /* Wait for screen update. */

		msg = g_string_append(msg, buffer); /* copy to gstring */
		msg = g_string_append_c(msg, '\n'); /* Add newline. */

		memset((void*)&buffer, 0, sizeof(buffer)); /* Clear buffer. */
		n = recvline(sock, buffer, sizeof(buffer)-1);/* Get next line. */
		i++;
   	}


	if (threshold > 0 && i >= threshold) {
		msg = g_string_append(msg, "\n\n---------------------------------------------------------\nThe rest of this message has not been downloaded. Please set a higher threshold to download the full text.\n");
		delete = FALSE; /* Don't delete a half downloaded msg??. */
	}


	/* Parse the message and insert into database. Don't delete the message
	 * off the server if there are any problems.
	 */
	if (!parse_pop3_msg(msg, uid, timer)) delete = FALSE; 
	GmailApp->newmsgs++; /* Should we increment if we didn't parse it? */
	g_free(uid);
	g_string_free(msg, TRUE);

	if (delete) p_dele(sock, id, timer); /* delete if we have to. */

  return TRUE;
}

static int p_dele(int sock, int id, DTimer *timer)
{
   /* delete message #id off the pop3 server */
   char buffer[256];
   char command[256];

		g_snprintf(buffer, 255, "Deleting message %d", id);
		set_status_label(timer->label, (gchar*) buffer);

   /* lets tell the server which message to delete */
   /* g_print("DELE %d...", id); */
   g_snprintf(command, 255, "DELE %d\r\n", id);
	
   if (send (sock, command, strlen(command), 0) < 0)
   {
      pop3errno = P_ERR_DELE_SEND;
      return ERROR;
   }

   if ( (recvline(sock, buffer, sizeof(buffer)-1) < 0) 
        || (strstr(buffer, "+OK") == NULL) )
   {
      pop3errno = P_ERR_DELE_RECV;
      return ERROR;
   }
   return SUCCESS;
}

static int p_rset(int sock)
{
   /* we are going to reset the smtp server (just to be nice) */
   char buffer[256];
   char command[256];

   strncpy(command, "RSET\r\n", 255);
	
   if (send (sock, command, strlen(command), 0) < 0)
   {
      pop3errno = P_ERR_RSET_SEND;
      return ERROR;
   }
	
   if ( (recvline(sock, buffer, 255) < 0)
       || (strstr(buffer, "+OK") == NULL) )
   {
      /* the server should have told us that it's been reset */
      pop3errno = P_ERR_RSET_RECV;
      return ERROR;
   }

   return SUCCESS;
}

static int p_quit(int sock)
{
   /* lets tell the server we want to quit now */
   char buffer[256];
   char command[256];

   strncpy(command, "QUIT\r\n", 255);

   if (send (sock, command, strlen(command), 0) < 0)
   {
      pop3errno = P_ERR_QUIT_SEND;
      return ERROR;
   }
	
   if ( (recvline(sock, buffer, 255) < 0)
       || (strstr(buffer, "+OK") == NULL) )
   {
      /* mr server didn't say bye-bye :( */
      pop3errno = P_ERR_QUIT_RECV;
      return ERROR;
   }
	
   return SUCCESS;
}

/* pop3_check - Connect to POP3 server and download messages. */
gboolean
pop3_check(char *server, int port, char *user, char *pass)
{
	int sock;
	int num, size, i;
	char buffer[256];
	Server *serv;
	gchar tmp[255], *title;
	gboolean delete = FALSE;
	GtkWidget *child;
	DTimer *timer;
	static const char *newmailsound[] = {"gmail", "newmail", NULL};
	static const char *nonewmailsound[] = {"gmail", "nonewmail", NULL};


	/* This global variable is used to see if we need to update a vfolder
	 * on screen when we have finished downloading.
	 * It's value will be -1 or the vfolder we need to update.
	 */
	GmailApp->updated_vfolder = -1;	/* Reset the variable. */

	timer = msg_match_popup();
	timer->value = 0;
	timer->total = 1;

	set_status_label(timer->label, "Looking up hostname.");

	serv = g_malloc(sizeof(Server));
	serv->hostname = server;
   	serv->port = (port_t) port;



   /* make sure we aren't claiming an error */
   pop3errno = P_NO_ERROR;
	
	/* resolve the hostname */
   if ( !Resolve(serv) )
   {
      pop3errno = P_ERR_RESOLVE;
			set_status_label(timer->label, "Bad POP3 hostname.");
      return ERROR;
   }

   /* assign and open a socket */
   sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
   if ( sock < 0 )    /* if socket fails to open */
   {
      pop3errno = P_ERR_SOCKET;
			set_status_label(timer->label, "No Socket Error.");
      return ERROR;
   }

   /* lets set up information about the server we will be connecting to... */
   serv->sin.sin_family = AF_INET;
   serv->sin.sin_port = htons(serv->port);

	set_status_label(timer->label, "Connecting to server...");

	g_return_val_if_fail(GTK_IS_WIDGET(timer->dialog), FALSE);
	
   /* call the server */
   if (connect(sock, (struct sockaddr*)&serv->sin, sizeof(serv->sin)) < 0)
   {
      /* we failed to connect */
      pop3errno = P_ERR_CONNECT;
      close(sock);
			set_status_label(timer->label, "No POP3 response.");
      return ERROR;
   }
	
   if ( (recvline(sock, buffer, 255) < 0)
       || (strstr(buffer, "+OK") == NULL) ) {
      /* the pop3 server failed to send a response (usually version info) */
      pop3errno = P_ERR_SERV_ACK;
      close(sock);
	  set_status_label(timer->label, "No POP3 response.");
      return ERROR;
   }

	set_status_label(timer->label, "Sending Login name...");
	g_return_val_if_fail(GTK_IS_WIDGET(timer->dialog), FALSE);

	/* lets login to the server... */
	if ( !p_user(sock, user) )
	{
		/* bad login name maybe? I dunno... */
		set_status_label(timer->label, "Login Error");
		close(sock);
		return ERROR;
	}

	set_status_label(timer->label, "Sending Password...");
	g_return_val_if_fail(GTK_IS_WIDGET(timer->dialog), FALSE);

	if ( !p_pass(sock, pass) )
	{
		/* bad passwd maybe? I dunno... */
		set_status_label(timer->label, "Username/Password Failed");
		close(sock);
		return ERROR;
	}

	set_status_label(timer->label, "Sending STAT command...");
	g_return_val_if_fail(GTK_IS_WIDGET(timer->dialog), FALSE);

	/* lets find out how many messages we got on the server */
	if ( (p_stat(sock, &num, &size)) < 0 )
	{
		/* something went wrong... */
		g_warning("stat returned %d", num);
		set_status_label(timer->label, "STAT command failed.");
		close(sock);
		return ERROR;
	}

	g_return_val_if_fail(GTK_IS_WIDGET(timer->dialog), FALSE);
	if (strcmp(gnome_config_get_string("/gmail/POP3Servers/leavemailonserver"), "FALSE") == 0) delete = TRUE;

 
	if (num == 0) {
	/* 0 messages is a special case. */
		timer->value = 1;
   	timer->total = 1;
		GmailApp->newmsgs = 0;

		} else {
		/* Set the timer range, based on the size of the maildrop.
	   * FIXME: This isn't 100%
		 */
		timer->value = 0;	
		timer->total = size;
		GmailApp->newmsgs = 0;

		for (i = 1; i <= num; i++) {
	g_return_val_if_fail(GTK_IS_WIDGET(timer->dialog), FALSE);
			if (!p_retr(sock, i, num, delete, timer)) {
				/* eek! an error occured!! */
				set_status_label(timer->label, "RETR Error");
				close(sock);
				return ERROR;
				}
			}
		}

	g_return_val_if_fail(GTK_IS_WIDGET(timer->dialog), FALSE);
	

	if ( !p_quit(sock) )
	{
		set_status_label(timer->label, "QUIT command failed.");
		close(sock);
		return ERROR;
	}
	
	/* close the socket */
	close(sock);

	timer->value = timer->total; 

	if (GmailApp->newmsgs == 1) {
		sprintf(tmp, "You have one new message.");
		} else {
		sprintf(tmp, "You have %i new messages.", GmailApp->newmsgs);
		}

	title = g_strdup_printf("Gmail - %s", tmp);
	gtk_window_set_title(GTK_WINDOW (timer->dialog), title);
	g_free(title);

	if (GmailApp->newmsgs == 0)
		gnome_triggers_vdo("", NULL, nonewmailsound); /* Play sound */

	if (GmailApp->newmsgs > 0)
		gnome_triggers_vdo("", NULL, newmailsound); /* Play sound */

	gnome_appbar_set_status (GNOME_APPBAR (GmailApp->appbar), tmp);

	set_status_label(timer->label, tmp);


	/* Update the button text. */
	child = GTK_BIN (timer->button)->child;
	if (GTK_IS_LABEL(child)) gtk_label_set(GTK_LABEL(child), "Close");

	g_free(serv);


	/* FIXME: See if we downloaded any messages that matched the currently
	 * opened vfolder. If so update that vfolder.
	 */
	if (GmailApp->updated_vfolder >= 0) {
		vfolder_display_rightpane(GmailApp->updated_vfolder);
		}
	/* g_print("updated_vfolder is %i\n", GmailApp->updated_vfolder); */
	GmailApp->updated_vfolder = -1;	/* Reset the variable. */
	
	return SUCCESS;
}


ssize_t recvline(int sock, void *vptr, size_t maxlen)
{
   int n, rc;
   char c, *ptr;

   ptr = vptr;
   for (n = 1; n < maxlen; n++)
   {
      if ( (rc = recvch(sock, &c)) == 1)
      {
         *ptr++ = c;
         if (c == '\n')
         break;                      /* newline is stored, like fgets() */
      }
      else if (rc == 0)
      {
         if (n == 1)
            return 0;                   /* EOF, no data read */
         else
            break;                      /* EOF, some data read */
      }
      else
         return -1;                     /* error, errno set by read() */
   }
   
   *ptr = 0;                            /* null terminate like fgets() */
   return n;
}


static int Resolve (Server *server)
{
   /* Resolves server->hostname to it's corresponding IPv4 IP address.
    * The IP address is stored in server->ip if successful.
    * Returns SUCCESS (1) or ERROR (0).
   */
   struct hostent *ptr_host;

   if ( (ptr_host = gethostbyname(server->hostname)) == NULL)
      return ERROR;

   memcpy(&server->sin.sin_addr, ptr_host->h_addr, ptr_host->h_length);
   g_snprintf(server->ip, sizeof(server->ip), "%s", inet_ntoa(server->sin.sin_addr));

   return SUCCESS;
}



ssize_t recvch(int sock, char *ptr)
{
   static int read_cnt = 0;
   static char *read_ptr;
   static char read_buf[512];

   if (read_cnt <= 0)
   {
      again:
         if ( (read_cnt = read(sock, read_buf, sizeof(read_buf))) < 0)
         {
            if (errno == EINTR)
               goto again;
            return -1;
         }
         else if (read_cnt == 0)
            return 0;
         read_ptr = read_buf;
   }
   read_cnt--;
   *ptr = *read_ptr++;
   return 1;
}
