/*
 * Gmail. A Gnome email client.
 * Copyright (C) 1999 Wayne Schuller
 *
 * print.c - code for printing messages using gnome-print
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "main.h"

#include <libgnomeprint/gnome-printer.h>
#include <libgnomeprint/gnome-print.h>
#include <libgnomeprint/gnome-print-meta.h>
#include <libgnomeprint/gnome-print-preview.h>
#include <libgnomeprint/gnome-print-pixbuf.h>
#include <libgnomeprint/gnome-font.h>
#include <libgnomeprint/gnome-printer-profile.h>
#include <libgnomeprint/gnome-printer-dialog.h>
#include <libgnomeprint/gnome-print-master.h>
#include <libgnomeprint/gnome-print-master-preview.h>
#include <libgnomeprint/gnome-print-dialog.h>

extern 	_GmailApp_	*GmailApp;
extern 	Message * load_vfolder_msg(gint id);
extern void free_message(Message *message);

void print_cb (GtkWidget *widget, void *data);
void print_preview_cb (GtkWidget *widget, void *data);
static void create_printer_page (GnomePrintContext *pc, Message *msg);
static void print_msg (Message *msg);
static gchar * get_line(gchar *pnt, gint line);

/* Gmail Callback for the print button. */
void 
print_cb (GtkWidget *widget, void *data)
{
	Message *msg;

	/* Get id of currently display message. */
	/* Load the message. */
	msg = load_vfolder_msg(GmailApp->msg_id);

	/* Print the message. */
	print_msg(msg);

	free_message(msg);
}

/* Gmail Callback for the print preview button. */
void 
print_preview_cb (GtkWidget *widget, void *data)
{
	Message *msg;
	GnomePrintContext *pc;
	GnomePrintMasterPreview *pmp;
	GnomePrintMaster *gpm;
	gint copies=1, collate=TRUE; 

	/* Get id of currently display message. */
	/* Load the message. */
	msg = load_vfolder_msg(GmailApp->msg_id);

	/* transfer dialog data to output context */
	gpm = gnome_print_master_new();
	gnome_print_master_set_copies(gpm, copies, collate);

	pc = gnome_print_master_get_context(gpm);
	create_printer_page (pc, msg);
	gnome_print_master_close(gpm);
	pmp = gnome_print_master_preview_new(gpm, "Test preview");
	gtk_widget_show(GTK_WIDGET(pmp));

	free_message(msg);
}

/* This is called from print_cb.
 * It opens the print dialog widget, and allows them the choice of printing
 * or viewing the print preview.
 * This is based on one of the gnome-print test programs.
 */
static void
print_msg (Message *msg)
{
	GnomePrintContext *pc;
	GnomePrintMaster *gpm;
	gint copies=1, collate=TRUE; 
	gboolean do_preview = FALSE;
	GnomePrintDialog *gpd;

	gpd = (GnomePrintDialog *) gnome_print_dialog_new("Print test", GNOME_PRINT_DIALOG_COPIES);
	gnome_print_dialog_set_copies(gpd, copies, collate);

	switch (gnome_dialog_run(GNOME_DIALOG(gpd))) {
	case GNOME_PRINT_PRINT:
		do_preview = FALSE;
		break;
	case GNOME_PRINT_PREVIEW:
		do_preview = TRUE;
		break;
	case GNOME_PRINT_CANCEL:
		gnome_dialog_close(GNOME_DIALOG(gpd));
		return;
		break;
	}

	/* printf("printer = %p\n", gnome_print_dialog_get_printer(gpd)); */

	/* transfer dialog data to output context */
	gpm = gnome_print_master_new();
	gnome_print_dialog_get_copies(gpd, &copies, &collate);
	gnome_print_master_set_copies(gpm, copies, collate);
	gnome_print_master_set_printer(gpm, gnome_print_dialog_get_printer(gpd));
	gnome_dialog_close(GNOME_DIALOG(gpd));

	pc = gnome_print_master_get_context(gpm);
	create_printer_page (pc, msg);
	gnome_print_master_close(gpm);

	if (do_preview) {
		GnomePrintMasterPreview *pmp;
		pmp = gnome_print_master_preview_new(gpm, "Test preview");
		gtk_widget_show(GTK_WIDGET(pmp));
	} else {
		gnome_print_master_print(gpm);
	}
}

/* This function actually does the job of rendering the message into
 * a gnome print page.
 * This is the function to modify to improve the look of the gmail
 * printing.
 */
static void
create_printer_page (GnomePrintContext *pc, Message *msg)
{
  GnomeFont *font;
	gint x;	/* The left margin. */
	gint y; /* The top margin (vertical start point). */
	gint inc; /* The amount of spacing between lines. */
	gint n;
	gchar *text;
	gchar *body = g_strdup(msg->message->str);
	gchar *pnt = body;

	/* Starting points. */
	x = 40; y = 800; inc = 30;

  gnome_print_gsave (pc);

  font = gnome_font_new_closest ("Helvetica", GNOME_FONT_BOOK, 0, 16);
  if (font == NULL)
    {
      g_warning ("Font not found\n");
			return;
    }
  gnome_print_setfont (pc, font);

	/* Print the email headers. */
  gnome_print_moveto (pc, x, y);
	text = g_strdup_printf("Subject: %s", msg->subject);
  gnome_print_show (pc, text);
	g_free(text);

	y -= inc; gnome_print_moveto (pc, x, y);
	text = g_strdup_printf("From: %s", msg->from);
  gnome_print_show (pc, text);
	g_free(text);

	y -= inc; gnome_print_moveto (pc, x, y);
	text = g_strdup_printf("To: %s", msg->to);
  gnome_print_show (pc, text);
	g_free(text);

	y -= inc; gnome_print_moveto (pc, x, y);
	text = g_strdup_printf("Message:");
  gnome_print_show (pc, text);
	g_free(text);

	/* Print the body of the email. */

  gtk_object_unref (GTK_OBJECT (font));
  font = gnome_font_new_closest ("Helvetica", GNOME_FONT_BOOK, 0, 14);
  if (font == NULL)
    {
      g_warning ("Font not found\n");
			return;
    }
  gnome_print_setfont (pc, font);

	y -= inc;  
	inc = 20;
	n = 1;

	/* FIXME: This code unfinished. */
	while (TRUE) { 
		gnome_print_moveto (pc, x, y);	/* Increment a newline. */

		text = strtok(body, "\n");	/* Grab the first line. */

		if (text == NULL) break; /* EOF. */

  	gnome_print_show (pc, text); /* Print the line. */

		n++; 
		body = NULL;
		y -= inc;

		/* Check if we need to start a new page. */
		if (y < 40) {
  		gnome_print_showpage (pc);
			gnome_print_beginpage(pc, "new page");
  		gnome_print_setfont (pc, font);
			y = 800;
		}
	}

	g_free(pnt);

	/* Print the footer. */
  gtk_object_unref (GTK_OBJECT (font));
  font = gnome_font_new_closest ("Helvetica", GNOME_FONT_LIGHT, 1, 10);
  gnome_print_setfont (pc, font);
  gnome_print_moveto (pc, 330, 20);

	text = g_strdup_printf("Printed by Gmail %s. (http://gmail.linuxpower.org)", VERSION);
  gnome_print_show (pc, text);
	g_free(text);

  gnome_print_showpage (pc);

  /*  gnome_print_context_close (pc);*/

  gtk_object_unref (GTK_OBJECT (font));
}

/* Return the next line to be printed. 
 * The calling function returns the memory.
 * FIXME: this is borked.
 */
static gchar *
get_line(gchar *body, gint line)
{
	gchar *text, *pnt = body, *start = body;
	gint ln, chars = 0;

	for (ln = 0; ln < line; ln++) {
	for (chars = 0; chars < 20 && *pnt != '\n' && *pnt != '\0'; pnt++, chars++) ;	
	g_print("line %i, chars %i, pnt: %s\n", ln, chars, pnt);
	}
	g_print("we live\n");


	/* Make a copy of the data up to this new line. */
	text = g_malloc(chars + 1 * sizeof(gchar));	
	strncpy(text, start, chars);
	text[chars] = '\0';

	g_print("text is %s\n", text);
	return(text);
}
