/*
 * Gmail. A Gnome email client.
 * Copyright (C) 1999-2000 Wayne Schuller
 *
 * settings.c - Read and parse gmail settings file..etc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "main.h"


extern void	broken_cb (GtkWidget *widget, void *data);
extern gboolean	DestroyRightPane(void);
extern	void 		mboxclicked_cb (GtkWidget *mblist, gint row, gint column, GdkEventButton *event, gpointer data);
extern void columnresize_cb(GtkCList *clist, gint column, gint width, gpointer user_data);
extern gboolean open_db(MYSQL *mysql);
extern void     prepare_app(void);
extern void recreate_matched_index (GtkWidget *widget, void *data);
extern  gboolean				TwoPaneView();
extern  void vfolder_display_rightpane(gint row);

extern _GmailApp_	*GmailApp;

static GnomeUIInfo vfolder_popup_menu [] = {
  { GNOME_APP_UI_ITEM, N_("Vfolder options:"), N_("Message options"), NULL, NULL, NULL, GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_MAIL, 0, 0, NULL },
  GNOMEUIINFO_SEPARATOR,
  { GNOME_APP_UI_ITEM, N_("_Properties"), N_("_Mailbox properties."), broken_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_PROP, 0, 0, NULL },
  { GNOME_APP_UI_ITEM, N_("_Move Up"), N_("_Move mailbox up."), broken_cb, NULL, NULL, 
	GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_UP, 0, 0, NULL },
  { GNOME_APP_UI_ITEM, N_("_Move Down"), N_("_Move mailbox down."), broken_cb, NULL, NULL, 
	GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_DOWN, 0, 0, NULL },
  { GNOME_APP_UI_ITEM, N_("_Delete"), N_("_Delete mailbox."), broken_cb, NULL, NULL, 
	GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_TRASH, 0, 0, NULL },
  GNOMEUIINFO_END
};

void      open_preferences (void);
GtkWidget * read_settings(void);
gboolean 	check_config_version(void);
gchar * 	check_config_prefs(void);
void update_font_styles(GtkStyle *style);
void version_warning_clicked (GnomeDialog * dialog, gint button_number, gpointer data);

static void	apply_prefs (GtkWidget *widget, GtkWidget *window);
static void	cancel_prefs (GtkWidget *widget, GtkWidget *window);
static void	set_prefs (void);
static void help_prefs(GtkWidget *window, gpointer data);
static gchar * check_online_prefs(void);
static void 	add_mbox_cb (GtkWidget * widget, gpointer type);
static void	 	delete_mbox_cb (GtkWidget * widget, gpointer data);
static void add_mbox_clicked(GnomeDialog * dialog, gint button_number, gpointer type);
static void make_copy_cb (GtkWidget * widget, gpointer type);
static void move_prefs_mblist (GtkWidget * widget, gchar *direction);
static void pr_mbox_selected (GtkWidget *clist, gint row, gint column, GdkEventButton *event, gpointer data);
static void	smtp_rb_clicked (GtkWidget * widget, gboolean SMTP);
static gboolean add_vfolder_to_prefs (gchar *name, gchar *query);
static void button_press (GtkWidget *widget, GdkEventButton *event, gpointer data);

/* notebook pages */
static GtkWidget *create_identity_page (void);
static GtkWidget *create_mailboxes_page (void);
static GtkWidget *create_fonts_page (void);
GtkWidget *create_database_page (gboolean full);
static GtkWidget *create_pop_servers_page (void);


/* 
 * Read the settings. 
 * We use the leftpane clist as our datastructure. (This is naughty(TM))
 * Basically the vfolder name is in the clist, and and we attach a Mailbox
 * data structure to each row.
 */
GtkWidget *
read_settings (void) {
	gchar **mboxes; /* Used to get mbox info from gnome_config */
	gint num; 			/* The total number of mbox entries in gnome_config */
	gint i; 				/* dummy counter */
	GtkWidget *clist;
	gchar *titles[2] = { "vfolder name"};
	gint row = 0;
	GtkWidget *gnomepixmap = NULL;
	GtkWidget *pixtitle;

	/* Create default dummy values. */
	GmailApp->left_pos = -1;
	GmailApp->right_pos = -1;
	GmailApp->rightlist = NULL;

	GmailApp->sortfield = gnome_config_get_string("/gmail/UserInfo/sortfield");
	if (GmailApp->sortfield == NULL)
		GmailApp->sortfield = g_strdup("date"); 
	GmailApp->sortorder = gnome_config_get_string("/gmail/UserInfo/sortorder");
	if (GmailApp->sortorder == NULL)
		GmailApp->sortorder = g_strdup("DESC");

	/* Remember database defaults. We could actually just grab these from
	 * gnome_config whenever we use them. But because we call open_db so much
	 * it is more efficient to grab them once at the start.
	 */
	GmailApp->hostname = gnome_config_get_string("/gmail/MysqlServer/hostname");
	GmailApp->username = gnome_config_get_string("/gmail/MysqlServer/username");
	GmailApp->password = gnome_config_private_get_string("/gmail/MysqlServer/password");
	GmailApp->dbname = gnome_config_get_string("/gmail/MysqlServer/dbname");

	/* We use null passwords if it is a blank field. */
	if (GmailApp->password != NULL && strlen(GmailApp->password) < 1) {
		g_free(GmailApp->password);
		GmailApp->password = NULL;
		}

	/* Get mailboxes from gnome_config settings. */
	gnome_config_get_vector("/gmail/mailbox/boxes", &num, &mboxes);

	clist = gtk_clist_new_with_titles (1, titles);
	gtk_clist_set_shadow_type(GTK_CLIST(clist), GTK_SHADOW_NONE); 

	/* Get a stock icon in the clist title.   */
	gnomepixmap = gnome_stock_new_with_icon(GNOME_STOCK_MENU_INDEX); 
	pixtitle = gnome_pixmap_button((GtkWidget *) gnomepixmap, "vfolders"); 
	gtk_button_set_relief(GTK_BUTTON(pixtitle), GTK_RELIEF_NONE);
	gtk_clist_set_column_widget(GTK_CLIST (clist), 0, pixtitle);



	gtk_clist_set_column_auto_resize (GTK_CLIST (clist), 0, TRUE);
	gtk_clist_set_column_resizeable (GTK_CLIST (clist), 0, TRUE);
	gtk_clist_set_selection_mode (GTK_CLIST (clist), GTK_SELECTION_SINGLE);
	gtk_clist_column_titles_passive( GTK_CLIST (clist) );

	/* This call must appear below the above 4 clist commands. Weird. */
	update_font_styles(clist->style);  
	gtk_widget_set_style (clist, GmailApp->style);    

	gtk_signal_connect(GTK_OBJECT(clist), "select_row",
                       GTK_SIGNAL_FUNC(mboxclicked_cb),
                       NULL);

	/* Callback for resizing a column. */
	gtk_signal_connect(GTK_OBJECT(clist), "resize_column",
                       GTK_SIGNAL_FUNC(columnresize_cb),
                       NULL);

  gtk_signal_connect (GTK_OBJECT (clist), "button_press_event",
          GTK_SIGNAL_FUNC (button_press), NULL);

	if (num < 3) {
		return clist;  /* No mailboxes????? */
	}
	


	/* Loop through mailboxes. */
	for (i=0; i < num; i += 3) {
		Mailbox	*mailbox;
		gchar *text[1], *config;

		text[0] = g_strdup(mboxes[i]);

		/* Put it in the clist. */
		gtk_clist_append (GTK_CLIST (clist), text);

		/* Create new mailbox.  */
		mailbox = g_malloc (sizeof(Mailbox)); 

		mailbox->type = g_strdup(mboxes[i+1]);

		/* Get SQL query in it. */
		config = g_strdup_printf("/gmail/vfolders/%s", text[0]);
		mailbox->query = gnome_config_get_string(config);
		g_free(config);

		if (strcmp(mboxes[i+2], "Read") == 0)
			mailbox->unreadmail = FALSE;
		else {
			mailbox->unreadmail = TRUE; 
 			gtk_clist_set_cell_style(GTK_CLIST (clist), row, 0, GmailApp->boldstyle);	
		}

		gtk_clist_set_row_data(GTK_CLIST (clist), row++, mailbox);
		g_free(text[0]);
		}

		/* FIXME: Add stock icon to the clist vfolder name here.
	   * For some reason it doesn't work at position 0 in the list. :(
		gnomepixmap = (GnomePixmap *) gnome_stock_new_with_icon(GNOME_STOCK_MENU_MAIL);
		gtk_clist_set_pixtext(GTK_CLIST (clist), 1, 0, "Test", 3, GNOME_PIXMAP(gnomepixmap)->pixmap, GNOME_PIXMAP(gnomepixmap)->mask);  
	   */

	return clist;
}

/* Open the preferences editor in the right pane. */
void 
open_preferences ()
{
	GtkWidget *window; /* Gnome dialog window */
	GtkWidget *frame;	/* Goes around buttons. */
	GtkWidget *label;	/* Goes on buttons */

	/* If prefs are open don't do anything. */
	if (GmailApp->nbook) return;

	/* Lets make a notebook. */
	GmailApp->nbook = gtk_notebook_new();

	/* Lets make the "identity" frame. */
	frame = create_identity_page();
	label = gtk_label_new("Identity");
	gtk_notebook_append_page(GTK_NOTEBOOK (GmailApp->nbook), frame, label);
	gtk_widget_show_all(frame);

	/* Lets make the mysql settings frame. */
	frame = create_database_page(TRUE);
	label = gtk_label_new("Mysql Server");
	gtk_notebook_append_page(GTK_NOTEBOOK (GmailApp->nbook), frame, label);
	gtk_widget_show_all(frame);

	/* Lets make the vfolders frame. */
	frame = create_mailboxes_page();
	label = gtk_label_new("Vfolders");
	gtk_notebook_append_page(GTK_NOTEBOOK (GmailApp->nbook), frame, label);
	gtk_widget_show_all(frame);

	/* Lets make the POP servers frame. */
	frame = create_pop_servers_page();
	label = gtk_label_new("POP3/SMTP");
	gtk_notebook_append_page(GTK_NOTEBOOK (GmailApp->nbook), frame, label);
	gtk_widget_show_all(frame);

	/* Lets make the fonts frame. */
	frame = create_fonts_page();
	label = gtk_label_new("Appearance");
	gtk_notebook_append_page(GTK_NOTEBOOK (GmailApp->nbook), frame, label);
	gtk_widget_show_all(frame);

	gtk_widget_show(GmailApp->nbook);

	set_prefs();	/* Put the current settings in the widgets. */


	/* Lets put it in a gnome dialog window. */
	window = gnome_dialog_new ("Gmail Preferences Window",
                 GNOME_STOCK_BUTTON_OK,
                 GNOME_STOCK_BUTTON_CANCEL,
                 GNOME_STOCK_BUTTON_HELP,
                 NULL);

 	if (GmailApp->app)
		gnome_dialog_set_parent (GNOME_DIALOG (window), GTK_WINDOW (GmailApp->app)); 
	
	gnome_dialog_button_connect (GNOME_DIALOG (window), 0, GTK_SIGNAL_FUNC (apply_prefs), window);
	gnome_dialog_button_connect (GNOME_DIALOG (window), 1, GTK_SIGNAL_FUNC (cancel_prefs), window);
	gnome_dialog_button_connect (GNOME_DIALOG (window), 2, GTK_SIGNAL_FUNC (help_prefs), window);

	gtk_box_pack_start (GTK_BOX (GNOME_DIALOG (window)->vbox), GmailApp->nbook, TRUE, TRUE, 0);

	gtk_widget_show_all(window);
}


/*
 * Apply the information in current open prefs window to the entire gmail
 * settings. Check for valid data before committing.
 *
 * Go sick at the user if there are essential fields missing.
 *
 */
static void
apply_prefs (GtkWidget *widget, GtkWidget *window)
{
	GtkWidget *dialog;
	GtkWidget *label;
	gchar *result = NULL;
	gchar *tmp;
	gchar *msg;
	gint 	row;
	gboolean vfolders_changed = FALSE;
	gchar **mboxes; /* Used to save vfolders to gnome_config */
	gint cnt = 0;	/* Used for counting positions in mboxes. */

	/* Check if we have all the essential prefs. */

	result = check_online_prefs();

	if (result != NULL) {
	  /* Send them a message saying they need to fill in the prefs */
		msg = g_strdup_printf ("Missing preferences: \n%s\n", result);	
		dialog = gnome_message_box_new(msg, GNOME_MESSAGE_BOX_ERROR, "doh!", 0);
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
    gtk_widget_show(dialog);
		g_free(msg);	
		return; /* Return to event loop leaving them in the prefs window */
		}

	/* If we get to this point we need to actually apply the prefs. */


	/* FIXME: This is open-source so we don't need any error checking here. */
	gnome_config_set_string("/gmail/UserInfo/EmailAddress", gtk_entry_get_text(GTK_ENTRY(GmailApp->email)));
	gnome_config_set_string("/gmail/UserInfo/RealName", gtk_entry_get_text(GTK_ENTRY(GmailApp->real_name)));
	gnome_config_set_string("/gmail/UserInfo/Signature", gtk_entry_get_text(GTK_ENTRY(GmailApp->signature)));
	gnome_config_set_string("/gmail/fonts/mainfont", gnome_font_picker_get_font_name(GNOME_FONT_PICKER(GmailApp->defaultfont)));
	gnome_config_set_string("/gmail/fonts/boldfont", gnome_font_picker_get_font_name(GNOME_FONT_PICKER(GmailApp->boldfont)));

	gnome_config_set_string("/gmail/MysqlServer/hostname", gtk_entry_get_text(GTK_ENTRY(GmailApp->db_host_entry)));
	gnome_config_set_string("/gmail/MysqlServer/username", gtk_entry_get_text(GTK_ENTRY(GmailApp->db_user_entry)));
	gnome_config_private_set_string("/gmail/MysqlServer/password", gtk_entry_get_text(GTK_ENTRY(GmailApp->db_pass_entry)));
	gnome_config_set_string("/gmail/MysqlServer/dbname", gtk_entry_get_text(GTK_ENTRY(GmailApp->db_name_entry)));

	/* Set outgoing mail method. */
	if (GTK_TOGGLE_BUTTON (GmailApp->rb_smtp_server)->active)  {
	gnome_config_set_string("/gmail/UserInfo/out_method", "smtp");
	gnome_config_set_string("/gmail/UserInfo/smtp", gtk_entry_get_text(GTK_ENTRY(GmailApp->smtp_server)));
		}
	else gnome_config_set_string("/gmail/UserInfo/out_method", "lmua");


	/* Apply the POP3 settings. */
	gnome_config_set_string("/gmail/POP3Servers/username", gtk_entry_get_text(GTK_ENTRY(GmailApp->username_entry)));
	gnome_config_set_string("/gmail/POP3Servers/servername", gtk_entry_get_text(GTK_ENTRY(GmailApp->pop_server_entry)));
	gnome_config_private_set_string("/gmail/POP3Servers/password", gtk_entry_get_text(GTK_ENTRY(GmailApp->pop_password_entry)));

	/* Apply pop3 leavemailonserver button. */
	if (GTK_TOGGLE_BUTTON (GmailApp->pop_leave_mail)->active) {
		gnome_config_set_string("/gmail/POP3Servers/leavemailonserver", "TRUE");
		} else {
		gnome_config_set_string("/gmail/POP3Servers/leavemailonserver", "FALSE");
		}

	/* Apply open first message button. */
	if (GTK_TOGGLE_BUTTON (GmailApp->open_first_msg)->active) {
		gnome_config_set_bool("/gmail/UserInfo/display_first_message", TRUE);
		} else {
		gnome_config_set_bool("/gmail/UserInfo/display_first_message", FALSE);
		}

	gnome_config_set_string("/gmail/UserInfo/AddressBook", gtk_entry_get_text(GTK_ENTRY(GmailApp->addressbook_entry)));

	/* Apply pop3 download threshold. */
	/* FIXME: do error checking on this */
	tmp = g_strdup_printf("%d", gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON (GmailApp->pop_threshold)));
	gnome_config_set_string("/gmail/POP3Servers/dlthreshold", tmp);

	/* Apply cache_vfolders toggle. */
	if (GTK_TOGGLE_BUTTON (GmailApp->cache_vf_toggle)->active) {
		gnome_config_set_string("/gmail/UserInfo/CacheVfolders", "TRUE");
	} else {
		gnome_config_set_string("/gmail/UserInfo/CacheVfolders", "FALSE");
		}

	/* Apply two or three pane view. 
	 * Check if they are changing the view type. If so we have some extra
	 * work to do.
	 */
	if (GTK_TOGGLE_BUTTON (GmailApp->rb_twopane)->active)   {
		if (!TwoPaneView()) {
			/* Need to change from three pane to two pane view. */

			/* Just destroy the rightpane. */
			gtk_widget_destroy(GmailApp->rightpane);
			GmailApp->rightpane = NULL;			
			GmailApp->msgpane = NULL;			
			GmailApp->msglistpane = NULL;

			/* FIXME: for some reason we coredump without destroying righlist. */
			gtk_widget_destroy(GmailApp->rightlist);
			GmailApp->rightlist = NULL;

			}
		gnome_config_set_string("/gmail/UserInfo/TwoPaneView", "TRUE");
		} else {
		if (TwoPaneView()) {
			/* Need to change from two pane to three pane view. */
			DestroyRightPane();

			GmailApp->rightpane = gtk_vpaned_new();
  		gtk_paned_set_handle_size (GTK_PANED(GmailApp->rightpane), 10);
  		gtk_paned_set_gutter_size (GTK_PANED(GmailApp->rightpane), 15);
			gtk_paned_set_position(GTK_PANED(GmailApp->rightpane), 150);

			/* What happens if you add2 over something else? */
			gtk_paned_add2 (GTK_PANED(GmailApp->hpaned), GmailApp->rightpane);
			gtk_widget_show_all (GmailApp->hpaned);

			}
		gnome_config_set_string("/gmail/UserInfo/TwoPaneView", "FALSE");
		}

	gnome_config_sync();	/* Write info to config file */

	/* Update our global variables of the mysql settings. */
	if (GmailApp->hostname) g_free(GmailApp->hostname);
	GmailApp->hostname = gnome_config_get_string("/gmail/MysqlServer/hostname");
	if (GmailApp->username) g_free(GmailApp->username);
	GmailApp->username = gnome_config_get_string("/gmail/MysqlServer/username");
	if (GmailApp->password) g_free(GmailApp->password);
	GmailApp->password = gnome_config_private_get_string("/gmail/MysqlServer/password");
	if (GmailApp->dbname) g_free(GmailApp->dbname);
	GmailApp->dbname = gnome_config_get_string("/gmail/MysqlServer/dbname");

	/* We use null passwords if it is a blank field. */
	if (GmailApp->password != NULL && strlen(GmailApp->password) < 1) {
		g_free(GmailApp->password);
		GmailApp->password = NULL;
		}
	
	/* We update the font style on top of the current default. This should already
	 * be created when the first left pane clist was drawn. */
	update_font_styles(GmailApp->style);  

	/* Have they changed any vfolders? */
	/* FIXME: We need to do a better check than this. Use a global var to mark
	 * any vfolder modifications.
	 */
	if (GTK_CLIST (GmailApp->pr_mboxes)->rows != GTK_CLIST (GmailApp->mblist)->rows)
		vfolders_changed = TRUE;

	/* Remove everything in the leftpane. */
	gtk_clist_freeze(GTK_CLIST(GmailApp->mblist));
	gtk_clist_clear(GTK_CLIST(GmailApp->mblist));
	/* Why is the below line commented out? */
	/*gtk_widget_set_style (GmailApp->mblist, GmailApp->style);  */

	/* Add everything from the prefs clist to the leftpane clist. */
	/* Save everything to the config file along the way. */
	mboxes = g_malloc(sizeof(gchar *) * ((GTK_CLIST(GmailApp->pr_mboxes)->rows)*3));
	for (row = 0; row != GTK_CLIST (GmailApp->pr_mboxes)->rows; row++) {
		Mailbox	*mailbox;
		gchar *name;
		gchar *newtext[1];
		gchar *config;

		/* Copy over the row values. */
		gtk_clist_get_text(GTK_CLIST(GmailApp->pr_mboxes), row, 0, &name);
		newtext[0] = g_strdup(name);
		gtk_clist_append (GTK_CLIST (GmailApp->mblist), newtext);

		/* Copy over the row data. */
		mailbox = gtk_clist_get_row_data(GTK_CLIST(GmailApp->pr_mboxes), row);	

		if (mailbox == NULL) g_warning("Pref mailbox list had a NULL row data (row %i)\n", row);

		gtk_clist_set_row_data(GTK_CLIST (GmailApp->mblist), row, mailbox);

		if (mailbox->unreadmail) {
				gtk_clist_set_cell_style(GTK_CLIST (GmailApp->mblist), row, 0, GmailApp->boldstyle);
			} else {
				gtk_clist_set_cell_style(GTK_CLIST (GmailApp->mblist), row, 0, GmailApp->style);
			}

		mboxes[cnt++] = g_strdup(name);
		mboxes[cnt++] = g_strdup("Vfolder"); /* Currently fixed */

		if (mailbox->unreadmail)  
			mboxes[cnt++] = g_strdup("Unread");
		else 
			mboxes[cnt++] = g_strdup("Read");

		/* Save the query. */
		config = g_strdup_printf("/gmail/vfolders/%s", name);
		gnome_config_set_string(config, mailbox->query);
		g_free(config);
		}

	gnome_config_set_vector("/gmail/mailbox/boxes", (GTK_CLIST(GmailApp->pr_mboxes)->rows)*3, (const char **) mboxes); 
	gnome_config_sync();
	g_free(mboxes); /* FIXME: This leaks. */

	gtk_clist_thaw(GTK_CLIST(GmailApp->mblist));

	/* Close the window and report to the user. */
	gtk_widget_destroy(GTK_WIDGET (window));
	GmailApp->nbook = NULL; /* Used for testing when prefs window is open. */

	/* Always display the first vfolder. */
	GmailApp->left_pos = 0;
 	vfolder_display_rightpane(0); 

	/* Report to the user. */
	if (vfolders_changed) {
		dialog = gnome_dialog_new (
                 "Gmail Popup",
                 GNOME_STOCK_PIXMAP_INDEX,
                 GNOME_STOCK_BUTTON_OK,
                 NULL);
		label = gtk_label_new ("Preferences Updated\nIf you have added, deleted or modified a vfolder (either the name or query),\nyou will have to run Settings->Advanced->Recreate Matched Index.\nTo do this immediately, click on 'Index', otherwise click 'OK' to continue."); 
	} else {
		dialog = gnome_dialog_new (
                 "Gmail Popup",
                 GNOME_STOCK_BUTTON_OK,
                 NULL);
		label = gtk_label_new ("Preferences Updated");
	}

 	if (GmailApp->app)
             gnome_dialog_set_parent (GNOME_DIALOG (dialog), GTK_WINDOW (GmailApp->app)); 
	gtk_box_pack_start (GTK_BOX (GNOME_DIALOG (dialog)->vbox), label, TRUE, TRUE, 0);


	gtk_widget_show_all(GNOME_DIALOG(dialog)->vbox);
	switch(gnome_dialog_run_and_close(GNOME_DIALOG (dialog))) {
		case 0: if (vfolders_changed) recreate_matched_index(NULL, NULL);
						break;

		case 1: break;

		default: break;
	}

}


/*
 * refresh data in the preferences window
 * Get the data straight from the config files.
 * FIXME: should this be the main place that we fix up their prefs if there
 * are incorrect or no defaults?
 */
static void
set_prefs (void)
{
	gchar *tmp;

	/* FIXME - do error checking here, is the prefs window open? */

	/* Identity Tab */
	gtk_entry_set_text (GTK_ENTRY (GmailApp->real_name), gnome_config_get_string("/gmail/UserInfo/RealName"));
	gtk_entry_set_text (GTK_ENTRY (GmailApp->email), gnome_config_get_string("/gmail/UserInfo/EmailAddress"));
	gtk_entry_set_text (GTK_ENTRY (GmailApp->signature), gnome_config_get_string("/gmail/UserInfo/Signature"));

	/* Database Tab */
	gtk_entry_set_text (GTK_ENTRY (GmailApp->db_host_entry), gnome_config_get_string("/gmail/MysqlServer/hostname"));
	gtk_entry_set_text (GTK_ENTRY (GmailApp->db_user_entry), gnome_config_get_string("/gmail/MysqlServer/username"));
	gtk_entry_set_text (GTK_ENTRY (GmailApp->db_pass_entry), gnome_config_private_get_string("/gmail/MysqlServer/password"));
	gtk_entry_set_text (GTK_ENTRY (GmailApp->db_name_entry), gnome_config_get_string("/gmail/MysqlServer/dbname"));

	/* Mailservers Tab. */
	tmp = gnome_config_get_string("/gmail/UserInfo/out_method");

	/* Set whether whether mail agent default is smtp or lmua  */
	if ((tmp == NULL) || (strcmp(tmp, "lmua") != 0))  {
  		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (GmailApp->rb_smtp_server), TRUE);
		gtk_entry_set_editable(GTK_ENTRY (GmailApp->smtp_server), TRUE);
		gtk_entry_set_text (GTK_ENTRY (GmailApp->smtp_server), gnome_config_get_string("/gmail/UserInfo/smtp"));
	} else  {
  		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (GmailApp->rb_local_mua), TRUE); 
		gtk_entry_set_editable(GTK_ENTRY (GmailApp->smtp_server), FALSE);
	}

	/* POP Servers Tab. */
	gtk_entry_set_text (GTK_ENTRY (GmailApp->username_entry), gnome_config_get_string("/gmail/POP3Servers/username"));
	gtk_entry_set_text (GTK_ENTRY (GmailApp->pop_server_entry), gnome_config_get_string("/gmail/POP3Servers/servername"));
	gtk_entry_set_text (GTK_ENTRY (GmailApp->pop_password_entry), gnome_config_private_get_string("/gmail/POP3Servers/password"));

	
	tmp = gnome_config_get_string("/gmail/POP3Servers/leavemailonserver");
	/* If they don't have a default for this, we set it to TRUE. */
	if (tmp == NULL) {
		gnome_config_set_string("/gmail/POP3Servers/leavemailonserver", "TRUE");
		gnome_config_sync();
		}

	/* Set pop3 leave mail on server toggle. */
	if (strcmp(gnome_config_get_string("/gmail/POP3Servers/leavemailonserver"), "TRUE") == 0) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (GmailApp->pop_leave_mail), TRUE);
	} else {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (GmailApp->pop_leave_mail), FALSE);
	}
	gtk_entry_set_text (GTK_ENTRY (GmailApp->addressbook_entry), gnome_config_get_string("/gmail/UserInfo/AddressBook=gnomecard"));

	/* Set open first msg toggle. */
	if (gnome_config_get_bool("/gmail/UserInfo/display_first_message") == TRUE) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (GmailApp->open_first_msg), TRUE);
	} else {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (GmailApp->open_first_msg), FALSE);
	}
	
	tmp = gnome_config_get_string("/gmail/POP3Servers/dlthreshold");
	/* If they don't have a default for this, we set it to 50 lines. */
	if (tmp == NULL) {
		gnome_config_set_string("/gmail/POP3Servers/dlthreshold", "50");
		gnome_config_sync();
		}

 gtk_spin_button_set_value(GTK_SPIN_BUTTON (GmailApp->pop_threshold), atoi(gnome_config_get_string("/gmail/POP3Servers/dlthreshold")));

	/* Misc Tab */

	/* FIXME: This is borked. 
	if(!gnome_font_picker_set_font_name(GNOME_FONT_PICKER(GmailApp->defaultfont), gnome_config_get_string("/gmail/fonts/mainfont")))
		g_print("FIXME: set default font failed. :(\n");
	if (!gnome_font_picker_set_font_name(GNOME_FONT_PICKER(GmailApp->boldfont), gnome_config_get_string("/gmail/fonts/boldfont")))
		g_print("FIXME: set bold font failed. :(\n"); */

	gnome_font_picker_set_font_name(GNOME_FONT_PICKER(GmailApp->boldfont), gnome_config_get_string("/gmail/fonts/boldfont"));
	gnome_font_picker_fi_set_use_font_in_label(GNOME_FONT_PICKER (GmailApp->defaultfont), TRUE, 15);
	gnome_font_picker_set_font_name(GNOME_FONT_PICKER(GmailApp->defaultfont), gnome_config_get_string("/gmail/fonts/mainfont"));
	gnome_font_picker_fi_set_use_font_in_label(GNOME_FONT_PICKER (GmailApp->boldfont), TRUE, 15); 

	/* Set vfolder cache toggle. (Default is yes) */
	if (gnome_config_get_string("/gmail/UserInfo/CacheVfolders") != NULL &&
		  (strcmp(gnome_config_get_string("/gmail/UserInfo/CacheVfolders"), 
										"FALSE") == 0)) 
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (GmailApp->cache_vf_toggle), FALSE);
	else 
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (GmailApp->cache_vf_toggle), TRUE);

	
	/* Set twopane or three pane mode. (default is two pane). */
	tmp = gnome_config_get_string("/gmail/UserInfo/TwoPaneView");
	if (tmp == NULL || strcmp(tmp, "TRUE")==0) 
  	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (GmailApp->rb_twopane), TRUE); 
	else
  	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (GmailApp->rb_threepane), TRUE); 
	g_free(tmp);

}


/*
 * identity notebook page
 */
static GtkWidget *
create_identity_page (void)
{
  GtkWidget *vbox;
  GtkWidget *table;
  GtkWidget *label;
  GtkWidget *signature;

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox), 10);

  table = gtk_table_new (5, 2, FALSE);
  gtk_box_pack_start (GTK_BOX (vbox), table, FALSE, FALSE, 5);

  /* your name */
  label = gtk_label_new ("Your name:");
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1,
		    GTK_FILL, GTK_FILL, 10, 10);

  GmailApp->real_name = gtk_entry_new ();
  gtk_table_attach (GTK_TABLE (table), GmailApp->real_name, 1, 2, 0, 1,
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 10);


  /* reply-to address */
  label = gtk_label_new ("Your email:");
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 2, 3,
		    GTK_FILL, GTK_FILL, 10, 10);

  GmailApp->email = gtk_entry_new ();
  gtk_table_attach (GTK_TABLE (table), GmailApp->email, 1, 2, 2, 3,
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 10);


  /* signature */
  label = gtk_label_new ("Signature:");
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 3, 4,
		    GTK_FILL, GTK_FILL, 10, 10);

  signature = gnome_file_entry_new ("Signature", "Signature");
  gtk_table_attach (GTK_TABLE (table), signature, 1, 2, 3, 4,
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 10);

  GmailApp->signature = gnome_file_entry_gtk_entry (GNOME_FILE_ENTRY (signature));
	

  label = gtk_label_new ("Your 'From' address can be modified when composing an email message.\nThe above email address only appears as the default.");
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 1);
  gtk_table_attach (GTK_TABLE (table), label, 0, 2, 4, 5,
		    GTK_FILL, GTK_FILL, 10, 10);

  return vbox;
}

/* Create the database stuff. We also call this from the gnome druid, so we
 * have a special 'full' option to decide whether we want to see all the 
 * settings or just the basics. In the druid we full=FALSE so we just get
 * the basics.
 */
GtkWidget *
create_database_page (gboolean full) {
  	GtkWidget *vbox;
  	GtkWidget *frame;
  	GtkWidget *label;
  	GtkWidget *table;

  	vbox = gtk_vbox_new (FALSE, 0);
  	gtk_container_set_border_width (GTK_CONTAINER (vbox), 10);

  	frame = gtk_frame_new ("Mysql Database Info");
  	gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 5);

  	table = gtk_table_new (4, 7, FALSE);
  	gtk_container_add (GTK_CONTAINER (frame), table);

	/* FIXME: this dumb label doesn't use the width of the table. :( 
	labelintro = gtk_label_new("Vfolders require a mysql server. It's good. These basic settings should have been done for you by the gnome-druid. See the README");

  	gtk_table_attach_defaults (GTK_TABLE (table), labelintro, 0, 3, 0, 1);
	gtk_label_set_justify(GTK_LABEL (labelintro), GTK_JUSTIFY_LEFT);
	gtk_label_set_line_wrap (GTK_LABEL (labelintro), TRUE); */

  	label = gtk_label_new ("Mysql server hostname:");
  	gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2,
		    GTK_FILL, GTK_FILL, 10, 10);
	gtk_label_set_justify(GTK_LABEL (label), GTK_JUSTIFY_RIGHT);

  	GmailApp->db_host_entry = gtk_entry_new ();
  	gtk_table_attach (GTK_TABLE (table), GmailApp->db_host_entry, 1, 3, 1, 2,
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 10);

  label = gtk_label_new ("Mysql user name:");
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 2, 3,
		    GTK_FILL, GTK_FILL, 10, 10);
  GmailApp->db_user_entry = gtk_entry_new ();
  gtk_table_attach (GTK_TABLE (table), GmailApp->db_user_entry, 1, 3, 2, 3,
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 10);

  label = gtk_label_new ("Mysql user password:");
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 3, 4,
		    GTK_FILL, GTK_FILL, 10, 10);
  GmailApp->db_pass_entry = gtk_entry_new ();
	gtk_entry_set_visibility(GTK_ENTRY(GmailApp->db_pass_entry), FALSE);
  gtk_table_attach (GTK_TABLE (table), GmailApp->db_pass_entry, 1, 3, 3, 4,
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 10);

  label = gtk_label_new ("Mysql database name:");
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 4, 5,
		    GTK_FILL, GTK_FILL, 10, 10);
  GmailApp->db_name_entry = gtk_entry_new ();
  gtk_table_attach (GTK_TABLE (table), GmailApp->db_name_entry, 1, 3, 4, 5,
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 10);

	/* Performance frame. */
	if (full) {
  frame = gtk_frame_new ("Performance");
  gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 5); 

  table = gtk_table_new (3, 3, FALSE);
  gtk_container_add (GTK_CONTAINER (frame), GTK_WIDGET (table)); 

	GmailApp->cache_vf_toggle = gtk_toggle_button_new_with_label("Cache vfolders");
  gtk_table_attach (GTK_TABLE (table), GmailApp->cache_vf_toggle, 0, 1, 0, 1,
		    GTK_FILL, GTK_FILL, 10, 10);

  label = gtk_label_new ("Recommended, but read the online help first.");
  /* gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5); */
  gtk_table_attach (GTK_TABLE (table), label, 1, 2, 0, 1,
		    GTK_FILL, GTK_FILL, 10, 10);
	}

  return vbox;
}

/*
 * mailboxes notebook page
 */
static GtkWidget *
create_mailboxes_page (void)
{
  GtkWidget *sw;
  GtkWidget *vbox;
  GtkWidget *frame;
  GtkWidget *hbox;
  GtkWidget *bbox;
  GtkWidget *button;
	gint row = 0;
	gchar *titles[2] = { "vfolder name"};


  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox), 10);

  frame = gtk_frame_new ("Vfolders");
  gtk_box_pack_start (GTK_BOX (vbox), frame, TRUE, TRUE, 5);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), 5);
  gtk_container_add (GTK_CONTAINER (frame), hbox);

  sw = gtk_scrolled_window_new (NULL, NULL);

	/* The mailboxes clist is an exact mirror of the one in the leftpane.
	 * The user fiddles with it and if they choose to keep the changes
	 * we move this one into the leftpane.
	 */
	/* Setup the basics. */
  GmailApp->pr_mboxes = gtk_clist_new_with_titles (1, titles);
  gtk_clist_set_column_title(GTK_CLIST(GmailApp->pr_mboxes), 0, _("vfolder name"));
  gtk_clist_column_titles_show(GTK_CLIST(GmailApp->pr_mboxes));
  gtk_clist_column_titles_passive(GTK_CLIST(GmailApp->pr_mboxes));
  gtk_container_add (GTK_CONTAINER (sw), GmailApp->pr_mboxes);

	/* Copy over info from the leftpane. */
	for (row = 0; row != GTK_CLIST (GmailApp->mblist)->rows; row++) {
		Mailbox	*mailbox;
		gchar *text;
		gchar *newtext[1];

		/* Copy over the row values. */
		gtk_clist_get_text(GTK_CLIST(GmailApp->mblist), row, 0, &text);
		newtext[0] = g_strdup(text);
		gtk_clist_append (GTK_CLIST (GmailApp->pr_mboxes), newtext);
		g_free(newtext[0]);

		/* Copy over the row data. */
		mailbox = gtk_clist_get_row_data(GTK_CLIST(GmailApp->mblist), row);	
		gtk_clist_set_row_data(GTK_CLIST (GmailApp->pr_mboxes), row, mailbox);
		}

	/* When a selection is made, we will remember which it is in our global
	 * variable area.
   */
	gtk_signal_connect(GTK_OBJECT(GmailApp->pr_mboxes), "select_row", GTK_SIGNAL_FUNC(pr_mbox_selected), NULL);
	GmailApp->pr_mbox_row = -1; /* this means nothing is selected */

  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
				  GTK_POLICY_AUTOMATIC,
				  GTK_POLICY_AUTOMATIC);

  gtk_box_pack_start (GTK_BOX (hbox), sw, TRUE, TRUE, 2);


  bbox = gtk_vbutton_box_new ();
  gtk_box_pack_start (GTK_BOX (hbox), bbox, FALSE, TRUE, 2);
  gtk_button_box_set_spacing (GTK_BUTTON_BOX (bbox), 2);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (bbox), GTK_BUTTONBOX_SPREAD);
  gtk_button_box_set_child_size (GTK_BUTTON_BOX (bbox), 25, 15);

  button = gtk_button_new_with_label ("add");
  gtk_container_add (GTK_CONTAINER (bbox), button);
	gtk_signal_connect (GTK_OBJECT (button), "clicked", 
			GTK_SIGNAL_FUNC (add_mbox_cb), (gpointer) "Add Mailbox");

  button = gtk_button_new_with_label ("modify");
  gtk_container_add (GTK_CONTAINER (bbox), button);
	gtk_signal_connect (GTK_OBJECT (button), "clicked", 
			GTK_SIGNAL_FUNC (add_mbox_cb), (gpointer) "Modify Mailbox");

  button = gtk_button_new_with_label ("delete");
  gtk_container_add (GTK_CONTAINER (bbox), button);
	gtk_signal_connect_object (GTK_OBJECT (button), "clicked", 
			GTK_SIGNAL_FUNC (delete_mbox_cb), NULL);

  button = gtk_button_new_with_label ("move up");
  gtk_container_add (GTK_CONTAINER (bbox), button);
	gtk_signal_connect (GTK_OBJECT (button), "clicked", 
			GTK_SIGNAL_FUNC (move_prefs_mblist), "up");

  button = gtk_button_new_with_label ("move down");
  gtk_container_add (GTK_CONTAINER (bbox), button);
	gtk_signal_connect (GTK_OBJECT (button), "clicked", 
			GTK_SIGNAL_FUNC (move_prefs_mblist), "down");

  button = gtk_button_new_with_label ("make copy");
  gtk_container_add (GTK_CONTAINER (bbox), button);
	gtk_signal_connect_object (GTK_OBJECT (button), "clicked", 
			GTK_SIGNAL_FUNC (make_copy_cb), NULL);

	return vbox;
}


/* create_fonts_page - Creates the fonts/display tabs in the prefs.
 * FIXME: should add stuff here so they can customise the message clist.
 * They should be able to choose whatever fields they want in whatever order.
 */
static GtkWidget *
create_fonts_page ()
{
  GtkWidget *vbox;
  GtkWidget *frame;
  GtkWidget *table;
  GtkWidget *label;
  GSList *rbgroup;


  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox), 10);

	/* standard font frame. */
  frame = gtk_frame_new ("Fonts");
  gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 5); 

  table = gtk_table_new (3, 3, FALSE);
  gtk_container_add (GTK_CONTAINER (frame), GTK_WIDGET (table)); 


	/* Add a font picker */
  label = gtk_label_new ("standard:");
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1,
		    GTK_FILL, GTK_FILL, 10, 10);

	GmailApp->defaultfont = gnome_font_picker_new();
	/* FIXME: The below mode always gets the font name wrong. */
	/* gnome_font_picker_set_mode(GNOME_FONT_PICKER(GmailApp->defaultfont), GNOME_FONT_PICKER_MODE_FONT_INFO); */
	gtk_widget_show(GmailApp->defaultfont);

	gtk_table_attach (GTK_TABLE (table), GTK_WIDGET(GmailApp->defaultfont), 1, 2, 0, 1, GTK_EXPAND | GTK_FILL, GTK_EXPAND, 10, 10);

  label = gtk_label_new ("highlighted:");
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2,
		    GTK_FILL, GTK_FILL, 10, 10);

	GmailApp->boldfont = gnome_font_picker_new();
	/* gnome_font_picker_set_mode(GNOME_FONT_PICKER(GmailApp->boldfont), GNOME_FONT_PICKER_MODE_FONT_INFO);*/
	gtk_widget_show(GmailApp->boldfont);
	gtk_table_attach (GTK_TABLE (table), GTK_WIDGET(GmailApp->boldfont), 1, 2, 1, 2, GTK_EXPAND | GTK_FILL, GTK_EXPAND, 10, 10);


  /* Panes */
  frame = gtk_frame_new ("Pane View");
  gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 5);

  table = gtk_table_new (2, 2, FALSE);
  gtk_container_add (GTK_CONTAINER (frame), table);
	GmailApp->rb_twopane = gtk_radio_button_new_with_label (NULL, _("Two Pane View"));
 	rbgroup = gtk_radio_button_group (GTK_RADIO_BUTTON (GmailApp->rb_twopane));

  gtk_table_attach (GTK_TABLE (table), GmailApp->rb_twopane, 0, 1, 0, 1, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  	gtk_widget_set_sensitive (GmailApp->rb_twopane, TRUE);

  	GmailApp->rb_threepane = gtk_radio_button_new_with_label (rbgroup, _("Three Pane View"));
  	gtk_table_attach (GTK_TABLE (table), GmailApp->rb_threepane, 0, 1, 1, 2, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);

  /* Panes */
  frame = gtk_frame_new ("Misc");
  gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 5);
  table = gtk_table_new (2, 3, FALSE);
  gtk_container_add (GTK_CONTAINER (frame), table);
	GmailApp->open_first_msg = gtk_toggle_button_new_with_label("Automatically open first message in 3 pane mode");
  gtk_table_attach (GTK_TABLE (table), GmailApp->open_first_msg, 0, 1, 0, 1,
		    GTK_FILL, GTK_FILL, 10, 10);

	/* Setup Addressbook program name. */
  label = gtk_label_new ("Address Book program:");
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2,
		    GTK_FILL, GTK_FILL, 10, 10);
  GmailApp->addressbook_entry = gtk_entry_new ();
  gtk_table_attach (GTK_TABLE (table), GmailApp->addressbook_entry, 1, 2, 1, 2,
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 10, 10);

  return vbox;
}


/* Cancel preferences window.
 * Don't let them get away with cancelling if we need essential data from them.
 */
static void
cancel_prefs (GtkWidget *widget, GtkWidget *window)
{
	GtkWidget *dialog;
	gchar			*result;
	gchar			*msg = NULL;

	/* Don't let them escape without sufficient config prefs to fall back on. */
	result = check_config_prefs();

	if (result == NULL)	{
		/* Close the prefs window. */
		gtk_widget_destroy(GTK_WIDGET (window));
		GmailApp->nbook = NULL; /* Used for testing when prefs window is open. */

		} else {
		/* Send dialog telling them to fill in the prefs they need. */
		msg = g_strdup_printf ("Missing preferences: \n%s\n", result);	
		dialog = gnome_message_box_new(msg, GNOME_MESSAGE_BOX_ERROR, "doh!", 0);
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
    gtk_widget_show(dialog);
		}
	g_free(msg);
	g_free(result);
}

/* check_online_prefs
 * This function will check the prefs in the current preferences window.
 * It is used by the ok and cancel buttons to make sure that it is safe
 * to leave the prefs window.
 * It returns NULL if everything is ok, otherwise it returns a string of
 * what is missing (seperated by newlines).
 */
static gchar *
check_online_prefs () {
	gchar *result = NULL;
	gchar *tmp;

	/* Is there at least one mailbox in the mailbox clist? */
	/* FIXME: Is there a better way to check this? */
	if (GTK_CLIST (GmailApp->pr_mboxes)->rows < 1) {
			result = g_strdup("No Mailboxes\n");
			}

	/* Have they entered an email address longer than 5 characters? */
	tmp = gtk_entry_get_text(GTK_ENTRY(GmailApp->email));
	if (tmp == NULL || (strlen(tmp) < 5)) {
			if (result == NULL) {
				result = g_strdup("No Email Address");
				} else {
				result = g_strdup_printf("No Email Address\n%s", result);	
				}
			}

	/* FIXME: Check that they have entered an smtp server if they have
     * checked the radio button to use outgoing smtp. 
     */

	return(result);
}

/* check_config_prefs
 * This will check the gmail config file to see that we have the essential
 * prefs items we need to do email.
 * If they don't we force them to set the prefs.
 * Returns NULL if prefs are ok.
 * Returns a string describing what is missing if prefs aren't ok.
 * This function is called at the start of running gmail, and when we
 * are processing their prefs editing, to make sure they don't remove
 * something essential.
 * WARNING: Don't get this confused with check_online_prefs, which is used
 * to check the prefs window for completeness. check_config_prefs just checks the
 * gnome_config settings file. They should be kept in sync for whatever are the
 * essential items for gmail.
 * 
 * FIXME: check for mysql settings.
 */
gchar *
check_config_prefs () {
	gchar *item;
	gchar **mboxes;
	gint 	num;
	gchar *result = NULL;

	/* g_print("Check config prefs\n"); */
	
	item = gnome_config_get_string("/gmail/UserInfo/EmailAddress");
	
	if (item == NULL) {
			result = g_strdup("No Email Address");		
			}

	gnome_config_get_vector("/gmail/mailbox/boxes", &num, &mboxes);
	if (num == 0) {
			gchar *msg;
			if (result != NULL) {
				msg = g_strdup_printf("%s\nNo Mailboxes", result);
				g_free(result);
				result = msg;	
				} else {
				result = g_strdup("No Mailboxes");		
				}	
			}

	return(result);
}

gboolean
check_config_version () {
	gint version;

	version = gnome_config_get_int("/gmail/UserInfo/DatabaseVersion");
	if (version == DATABASE_VERSION) {
		return(TRUE);
	} else {
		GtkWidget *label;
		gchar *msg;

		GmailApp->mbox_popup = gnome_dialog_new("Gmail Warning", GNOME_STOCK_BUTTON_OK, GNOME_STOCK_BUTTON_CANCEL, NULL);
 	if (GmailApp->app)
		gnome_dialog_set_parent (GNOME_DIALOG (GmailApp->app), GTK_WINDOW (GmailApp->app)); 
		gtk_window_set_modal(GTK_WINDOW (GmailApp->mbox_popup), TRUE);

		msg = g_strdup_printf("Warning: Your database version number is %i, but this version of Gmail requires version %i.\nYou should probably delete your ~/.gnome/gmail and setup a new database.\nIf you want to attempt to transfer your database, read README.export_import\nYou can click ok here to continue running gmail, but things will probably go wrong.\nOtherwise click on cancel to give up.", version, DATABASE_VERSION);
		label = gtk_label_new (msg);
		gtk_box_pack_start (GTK_BOX (GNOME_DIALOG(GmailApp->mbox_popup)->vbox), label, FALSE, FALSE, 5);
  gtk_signal_connect(GTK_OBJECT(GmailApp->mbox_popup), "clicked",
         GTK_SIGNAL_FUNC(version_warning_clicked), NULL);
		gtk_widget_show_all(GmailApp->mbox_popup);
	return(FALSE);
	}
	
	return(FALSE);
}

void
version_warning_clicked (GnomeDialog * dialog, gint button_number, gpointer data)
{
	gtk_widget_destroy(GTK_WIDGET(dialog));

	switch (button_number) {
		case 0: /* OK button */
		prepare_app(); /* Normal Gmail start */
		break;
		case 1: /* Cancel button */
		gtk_main_quit ();
		return;
		break;
		}
}

/* delete_mbox_cb
 * This function is called when the user clicks on the 'delete' mailbox button
 * in the preferences window.
 * We should only delete it from the prefs clist, we only change
 * global settings when they click on ok. 
 * 
 */
static void
delete_mbox_cb (GtkWidget * widget, gpointer data)
{

	/* g_print("lets delete %d\n", GmailApp->pr_mbox_row); */

	/* Delete from preferences clist. */
	gtk_clist_remove(GTK_CLIST(GmailApp->pr_mboxes), GmailApp->pr_mbox_row);

	/* FIXME: Send dialog to user? */

	/* Nothing is selected now. */	
	GmailApp->pr_mbox_row = -1;
}

/* add_mbox_cb
 * This function is called when the user clicks on the 'add' mailbox button
 * in the preferences window. It adds a popup window for the user to 
 * add a vfolder. 
 * This function is also now called when they click on the modify mailbox
 * button. Behaviour is mostly simliar.
 * 
 */
static void
add_mbox_cb (GtkWidget * widget, gpointer type)
{
  GtkWidget *label;
  GtkWidget *table;
  GtkWidget *frame;

	if ((strcmp((gchar *) type, "Modify Mailbox") == 0) && 
				(GmailApp->pr_mbox_row == -1)) {
		/* They want to modify a mailbox but they haven't selected one.
     * So we will just make them add a mailbox. 
     */
		type = g_strdup("Add Mailbox");
		}
		
	GmailApp->mbox_popup = gnome_dialog_new((char *) type, GNOME_STOCK_BUTTON_OK, GNOME_STOCK_BUTTON_CANCEL, GNOME_STOCK_BUTTON_HELP, NULL);
	gtk_window_set_modal(GTK_WINDOW (GmailApp->mbox_popup), TRUE);
	if (GmailApp->app)
    	gnome_dialog_set_parent (GNOME_DIALOG (GmailApp->mbox_popup), GTK_WINDOW (GmailApp->app));
  gtk_signal_connect(GTK_OBJECT(GmailApp->mbox_popup), "clicked",
         GTK_SIGNAL_FUNC(add_mbox_clicked), type);
	gtk_widget_set_usize(GTK_WIDGET(GmailApp->mbox_popup), 550, 230);	

	/* Vfolder Section */
  frame = gtk_frame_new ("Vfolder (SQL Query) Mailbox");
  gtk_box_pack_start (GTK_BOX (GNOME_DIALOG(GmailApp->mbox_popup)->vbox), frame, FALSE, FALSE, 5);
  table = gtk_table_new (3, 5, FALSE);
  gtk_container_add (GTK_CONTAINER (frame), table);


  label = gtk_label_new ("Vfolder Name:");
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1,
		    GTK_SHRINK, GTK_FILL, 10, 10);
  GmailApp->vfolder_name_entry = gtk_entry_new ();
  gtk_table_attach (GTK_TABLE (table), GmailApp->vfolder_name_entry, 1, 3, 0, 1,
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 10, 10);
  label = gtk_label_new ("WHERE (match conditions):");
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2,
		    GTK_SHRINK, GTK_FILL, 10, 10);
  GmailApp->sql_entry = gtk_entry_new ();
  gtk_table_attach (GTK_TABLE (table), GmailApp->sql_entry, 1, 3, 1, 2,
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 10, 10);
  label = gtk_label_new ("Click on the Help button for specific help on writing vfolder queries.");

  gtk_table_attach (GTK_TABLE (table), label, 0, 3, 2, 4,
		    GTK_FILL | GTK_SHRINK, GTK_FILL, 10, 10);

	/* Put in the current values if we are modifying a mailbox. */
	if (strcmp((gchar *) type, "Modify Mailbox") == 0) {
			gchar *name;
			Mailbox	*mailbox;

			gtk_clist_get_text(GTK_CLIST(GmailApp->pr_mboxes), GmailApp->pr_mbox_row, 0, &name);
			gtk_entry_set_text (GTK_ENTRY (GmailApp->vfolder_name_entry), name);
			mailbox = gtk_clist_get_row_data(GTK_CLIST (GmailApp->pr_mboxes), GmailApp->pr_mbox_row);

			gtk_entry_set_text (GTK_ENTRY (GmailApp->sql_entry), mailbox->query);
			
		}

	gtk_widget_show_all(GmailApp->mbox_popup);
}


/* This callback is called when the user has given a mailbox to add to the
 * list, or when they have modified an existing mailbox.
 */
static void
add_mbox_clicked(GnomeDialog * dialog, gint button_number, gpointer type)
{
	static GnomeHelpMenuEntry help_ref = {PACKAGE, "vfolder.html" };
	Mailbox	*mailbox;

	switch (button_number) {
		case 0: /* OK button */
			/* Assume they want a vfolder. */
			/* Redraw mblist in prefs window. */
			if (strcmp((gchar *) type, "Modify Mailbox") == 0) {

				/* Modify existing clist entry. */

				gtk_clist_set_text(GTK_CLIST (GmailApp->pr_mboxes),
					GmailApp->pr_mbox_row, 0, 
					gtk_entry_get_text(GTK_ENTRY(GmailApp->vfolder_name_entry)));

				gtk_clist_set_text(GTK_CLIST (GmailApp->pr_mboxes),
					GmailApp->pr_mbox_row, 1, "vfolder");

				/* Modify query in the mailbox. */
				mailbox = gtk_clist_get_row_data(GTK_CLIST (GmailApp->pr_mboxes), GmailApp->pr_mbox_row);
				g_free(mailbox->query);
				mailbox->query = g_strdup(gtk_entry_get_text(GTK_ENTRY(GmailApp->sql_entry)));
				g_strstrip(mailbox->query);	/* To avoid newlines etc. */

				/* FIXME: rematch the mailbox here now that the sql has changed? */

				} else {
				/* Brand new vfolder then. */
				add_vfolder_to_prefs (
					gtk_entry_get_text(GTK_ENTRY(GmailApp->vfolder_name_entry)), 
					gtk_entry_get_text(GTK_ENTRY(GmailApp->sql_entry)));
					}
			gnome_dialog_close(dialog);
			break;

		case 1: /* Cancel Button */
			/* Not much work to do in this case. */
			gnome_dialog_close(dialog);
			break;
		case 2: /* Help Button */
			gnome_help_display(0, &help_ref);
			break;
		};
}


/* Callback for the prefs help button. 
 * FIXME: write a specific help section for the prefs area then redirect
 * this callback to that section.
 */
static void
help_prefs(GtkWidget *window, gpointer data)
{
	static GnomeHelpMenuEntry help_ref = {PACKAGE, "index.html" };
	gnome_help_display(0, &help_ref);
}

/* FIXME: this function is redundant. Lose it. */
static void
pr_mbox_selected (GtkWidget *clist, gint row, gint column, GdkEventButton *event, gpointer data)
{
	GmailApp->pr_mbox_row = row;
}


/* create_pop_servers_page
 * This is the pop servers tab in the preferences notebook.
 */
static GtkWidget *
create_pop_servers_page () {
  GtkWidget *vbox;
  GtkWidget *frame;
  GtkWidget *label;
  GtkWidget *table;
  GtkAdjustment *adj;
  GSList *rbgroup;
  GtkWidget *table1;

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox), 10);

	/* POP Section */
  frame = gtk_frame_new ("POP3 (Post Office Protocol) Server");
  gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 5);
  table = gtk_table_new (3, 5, FALSE);
  gtk_container_add (GTK_CONTAINER (frame), table);

  label = gtk_label_new ("POP3 username:");
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1,
		    GTK_FILL, GTK_FILL, 10, 10);
  GmailApp->username_entry = gtk_entry_new ();
  gtk_table_attach (GTK_TABLE (table), GmailApp->username_entry, 1, 3, 0, 1,
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 10);
  label = gtk_label_new ("POP3 Password:");
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2,
		    GTK_FILL, GTK_FILL, 10, 10);
  GmailApp->pop_password_entry = gtk_entry_new ();
	gtk_entry_set_visibility(GTK_ENTRY(GmailApp->pop_password_entry), FALSE);
  gtk_table_attach (GTK_TABLE (table), GmailApp->pop_password_entry, 1, 3, 1, 2,
		    GTK_FILL, GTK_FILL, 0, 10);
  label = gtk_label_new ("POP3 server:");
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 2, 3,
		    GTK_FILL, GTK_FILL, 10, 10);
  GmailApp->pop_server_entry = gtk_entry_new ();
  gtk_table_attach (GTK_TABLE (table), GmailApp->pop_server_entry, 1, 3, 2, 3,
		    GTK_FILL, GTK_FILL, 0, 10);

	GmailApp->pop_leave_mail = gtk_toggle_button_new_with_label("Leave mail on server");
  gtk_table_attach (GTK_TABLE (table), GmailApp->pop_leave_mail, 0, 1, 3, 4,
		    GTK_FILL, GTK_FILL, 10, 10);

  label = gtk_label_new ("Max lines to download (0 means all):");
  gtk_misc_set_alignment (GTK_MISC (label), 1, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 1, 2, 3, 4,
		    GTK_FILL, GTK_FILL, 10, 10);
	adj = (GtkAdjustment *) gtk_adjustment_new (0.0, 0.0, 1000.0, 1.0, 5.0, 0.0);
	GmailApp->pop_threshold = gtk_spin_button_new((GtkAdjustment *) adj, 0, 0);
  gtk_table_attach (GTK_TABLE (table), GmailApp->pop_threshold, 2, 3, 3, 4,
		    GTK_FILL, GTK_FILL, 10, 10);

  /* smtp server */
  frame = gtk_frame_new ("Outgoing Mail");
  gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 5);

  table1 = gtk_table_new (2, 2, FALSE);
  gtk_container_add (GTK_CONTAINER (frame), table1);
	GmailApp->rb_smtp_server = gtk_radio_button_new_with_label (NULL, _("Remote SMTP Server"));
	if (GmailApp->rb_smtp_server == NULL) g_warning("rb_smtp_server is NULL.");

  	rbgroup = gtk_radio_button_group (GTK_RADIO_BUTTON (GmailApp->rb_smtp_server));
	if (rbgroup == NULL) g_warning("rbgroup is NULL.");

  	gtk_table_attach (GTK_TABLE (table1), GmailApp->rb_smtp_server, 0, 1, 0, 1,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
		    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  	gtk_widget_set_sensitive (GmailApp->rb_smtp_server, TRUE);

  	GmailApp->rb_local_mua = gtk_radio_button_new_with_label (rbgroup,
  	_("Local mail user agent (very buggy)"));
  	gtk_table_attach (GTK_TABLE (table1), GmailApp->rb_local_mua, 0, 1, 1, 2,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
		    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);

  	GmailApp->smtp_server = gtk_entry_new ();
  	gtk_table_attach (GTK_TABLE (table1), GmailApp->smtp_server, 1, 2, 0, 1,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
		    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  	gtk_widget_set_sensitive (GmailApp->smtp_server, TRUE);

	gtk_signal_connect (GTK_OBJECT (GmailApp->rb_smtp_server), "toggled", 
				GTK_SIGNAL_FUNC (smtp_rb_clicked), (gpointer) TRUE);
	gtk_signal_connect (GTK_OBJECT (GmailApp->rb_local_mua), "toggled", 
				GTK_SIGNAL_FUNC (smtp_rb_clicked), (gpointer) FALSE);

	gtk_widget_show_all(frame);

	return vbox;
}

/* This callback is called whenever they click on one of the radio buttons
 * to choose outgoing mail method. 
 * This actually gets called twice everytime a radio button is clicked,
 * because one is toggled on and one off. I should have found a better signal.
 * It works though.
 */
static void	
smtp_rb_clicked (GtkWidget *widget, gboolean SMTP) {

	if (SMTP && GTK_TOGGLE_BUTTON (widget)->active) {
		gtk_entry_set_editable(GTK_ENTRY (GmailApp->smtp_server), TRUE);
		gtk_entry_set_visibility(GTK_ENTRY (GmailApp->smtp_server), TRUE);
	} else if (!SMTP && GTK_TOGGLE_BUTTON (widget)->active) {
		gtk_entry_set_editable(GTK_ENTRY (GmailApp->smtp_server), FALSE);
		gtk_entry_set_visibility(GTK_ENTRY (GmailApp->smtp_server), FALSE);
	}
	/* Force screen update. (taken from GTK FAQ)  */
	while (g_main_iteration(FALSE));
}

static gboolean
add_vfolder_to_prefs (gchar *name, gchar *query) {
	gchar *clist[2];
	Mailbox	*mailbox;
	gint row;

	g_strstrip(name); 
	g_strstrip(query); 

	/* Create new mailbox.  */
	mailbox = g_malloc (sizeof(Mailbox)); 
	mailbox->unreadmail = FALSE;	/* Assume readmail. */
	mailbox->query = g_strdup(query);

	/* g_print ("add_vfolder_to_prefs passed name: %s\n", name); */
	clist[0] = g_strdup(name);
	clist[1] = g_strdup("vfolder");	


	/* Put it in preferences mailboxes clist. */
	row = gtk_clist_append(GTK_CLIST (GmailApp->pr_mboxes), clist);
	gtk_clist_set_row_data(GTK_CLIST (GmailApp->pr_mboxes), row, mailbox);

	/* Move view to end of the clist and select it. */
	gtk_clist_moveto(GTK_CLIST (GmailApp->pr_mboxes), GTK_CLIST (GmailApp->pr_mboxes)->rows - 1 , 0, 0.5, 0);
	gtk_clist_select_row(GTK_CLIST (GmailApp->pr_mboxes), GTK_CLIST (GmailApp->pr_mboxes)->rows - 1 , 0);

	/* free stuff (not as in speech or beer) */ 
	g_free(clist[0]);
	g_free(clist[1]);
	return TRUE;
}

/* make_copy_cb makes a copy of the mailbox they have clicked on.
 * This button is useful for making a copy of a mailbox, then changing
 * it slightly to be a new one.
 */
static void
make_copy_cb (GtkWidget * widget, gpointer type) {
	gchar *newname;
	gchar *text;
	Mailbox	*mailbox;

	/* If they haven't selected a mailbox, just abort. */
	if 	(GmailApp->pr_mbox_row == -1) return;

	gtk_clist_get_text(GTK_CLIST(GmailApp->pr_mboxes), GmailApp->pr_mbox_row, 0, &text);
	mailbox = gtk_clist_get_row_data(GTK_CLIST (GmailApp->pr_mboxes), GmailApp->pr_mbox_row);

	newname = g_strdup_printf("copy of %s", text);	
	add_vfolder_to_prefs (newname, mailbox->query); 

	g_free(newname);
}

/* move_prefs_mblist - move the select row up or down.
 *
 *
 */
static void
move_prefs_mblist (GtkWidget * widget, gchar *direction) {
	gchar *name;
	gchar *type;
	gchar *text[2];
	Mailbox	*mailbox;
	gint row = GmailApp->pr_mbox_row;
	gint newrow;
	gint size;

	if (row < 0) return;
	
	size = GTK_CLIST (GmailApp->pr_mboxes)->rows;

	/* Get values. */
	gtk_clist_get_text(GTK_CLIST(GmailApp->pr_mboxes), row, 0, &name);	
	gtk_clist_get_text(GTK_CLIST(GmailApp->pr_mboxes), row, 1, &type);	

	mailbox = gtk_clist_get_row_data(GTK_CLIST (GmailApp->pr_mboxes), row);

	text[0] = g_strdup(name);
	text[1] = g_strdup(type);

	/* g_print("move %i %s (%i)\n", row, direction, size); */

	if (strcmp(direction, "up") == 0) {
		newrow = row - 1; if (newrow < 0) newrow = 0;
		} else {
		newrow = row + 1; if (newrow >= size) newrow = size - 1;
		}

	if (newrow >= 0 && newrow < size) {
		/* Remove it. */
		gtk_clist_remove(GTK_CLIST(GmailApp->pr_mboxes), row);	

		/* Add it again! */
		gtk_clist_insert(GTK_CLIST(GmailApp->pr_mboxes), newrow, text);	
		gtk_clist_set_row_data(GTK_CLIST (GmailApp->pr_mboxes), newrow, mailbox);

		/* select it. */
		gtk_clist_select_row(GTK_CLIST (GmailApp->pr_mboxes), newrow, 0);

		/* centre it. */
		gtk_clist_moveto(GTK_CLIST (GmailApp->pr_mboxes), newrow, 0, 0.5, 0);

		} else {
		g_print("can't move clist past borders.\n");
		}

	g_free(text[0]);
	g_free(text[1]);
}

/* Handle rightclick popup menus. */
static void
button_press (GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	GtkWidget *menu;
	gint result;
	gint row, col;

	if (event->button != 3) return;

	gtk_clist_get_selection_info(GTK_CLIST(widget), event->x, event->y, &row, &col);
	gtk_clist_select_row(GTK_CLIST(widget), row, col);

  menu = gnome_popup_menu_new (vfolder_popup_menu);
	result = gnome_popup_menu_do_popup_modal (menu, NULL, NULL, event, widget);
  gtk_widget_destroy (menu);
	/* g_print("User chose %i\n", result); */
}

/* update_font_styles - set the global font variables based on gnome config
 * settings.
 * Called in read_settings at start of program and also when the fonts
 * are updated.
 * FIXME: It is passed the clist in use in order to be able to copy the
 * style. I need to do this to obey gtk themes. There must be a better way
 * to just get a copy of the current gtk theme style.
 */
void
update_font_styles(GtkStyle *style)
{
	/* copy the current style. */
	GmailApp->style = gtk_style_copy(style);  
	if (gnome_config_get_string("/gmail/fonts/mainfont") != NULL) {
	gdk_font_unref(GmailApp->style->font);
	GmailApp->style->font = gdk_font_load (gnome_config_get_string("/gmail/fonts/mainfont")); 
		}

	/* copy the current style. */
	GmailApp->boldstyle = gtk_style_copy(style); 
	if (gnome_config_get_string("/gmail/fonts/boldfont") != NULL) {
		gdk_font_unref(GmailApp->boldstyle->font);
		GmailApp->boldstyle->font = gdk_font_load (gnome_config_get_string("/gmail/fonts/boldfont"));
		}
}
